#include <stdio.h>
#define __USE_LARGEFILE64
#define __USE_FILE_OFFSET64
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <signal.h>
#include <pthread.h>
#include "dtvstream.h"

extern int errno;

struct buffer_ref {
	struct buffer_ref *next;
	unsigned char *addr;
	ssize_t length;			// set for refs in full list
	int file_seqnum;		// set for refs in full list
	};
typedef struct buffer_ref buffer_ref;

// variable from main program
extern char output_file_name[];
extern char output_file_ext[];	// does not include "."
extern char write_replace;
extern unsigned long long packet_limit;
extern char debug_output;
extern char verbose_output;
extern int buffer_k_bytes;
extern int input_is_char_dev;
extern int max_file_segments;
extern int max_file_count;

// variables exclusive to writer while it is running
extern char output_file[];
extern unsigned long long write_byte_count;
buffer_ref *current_write_ref;
int out_fd;
char out_open;
mode_t write_open_mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;
mode_t open_flags;
int out_blocksize;
int writer_error;
char writer_done;

unsigned long long write_limit_counter;
int output_file_seqnum = 1;
pthread_mutex_t full_mutex;
pthread_cond_t full_cond;
pthread_t writer_thread;
pthread_mutex_t free_mutex;
pthread_cond_t free_cond;
pthread_mutex_t error_mutex;

buffer_ref *free_list;
buffer_ref *full_list;
buffer_ref *last_full;
buffer_ref *current_fill_ref;
unsigned char *output_buffer;
unsigned char *output_ptr;
int alloc_buffer_count = 0;
int alloc_buffer_max = 0;

int init_writer(void);
int output_packet(unsigned char *packet);
int had_writer_error(void);
void end_writer(void);
int queue_output(char end);
int get_new_buffer(void);
int add_new_buffer(void);
void free_buffer_list(buffer_ref *ref);
void free_buffer_ref(buffer_ref *ref);
void *writer(void *args);
void close_output(void);

extern char *safe_strcat(char *dest, char *src, int src_copylen, int maxlen);
void end_status(void);

int init_writer(void)
{
	struct stat out_stat;
	int result;

	// initialize control variables
	write_byte_count = 0;
	write_limit_counter = 0;
	out_open = 0;
	output_buffer = 0;
	open_flags = O_WRONLY | O_CREAT;
	if (write_replace == 0) open_flags |= O_EXCL;
	else open_flags |= O_TRUNC;
#ifdef O_LARGEFILE
	open_flags |= O_LARGEFILE;
#endif

	// open output file
	out_fd = open(output_file, open_flags, write_open_mode);
	if (out_fd == -1) {
		end_status();
		fprintf(stderr, "Unable to create \"%s\": %s\n", output_file, strerror(errno));
		return errno;
		}
	out_open = 1;

	// get optimal blocksize
	result = fstat(out_fd, &out_stat);
	if (result != 0) {
		end_status();
		fprintf(stderr, "Unable to stat \"%s\": %s\n", output_file, strerror(errno));
		return(errno);
		}
	out_blocksize = out_stat.st_blksize;
	if (debug_output) {
		end_status();
		printf("optimal output blocksize = %d\n", out_blocksize);
		}
	if (out_blocksize < TRANSPORT_PKT_SIZE * 3) {
		out_blocksize = REPLACEMENT_OUTPUT_BLKSIZE;
		if (debug_output) {
			end_status();
			printf("changed output blocksize to %d\n", out_blocksize);
			}
		}

	if (buffer_k_bytes > 0) {
		alloc_buffer_max = (((long)buffer_k_bytes) * 1024) / out_blocksize;
		if (debug_output) {
			end_status();
			printf("Maximum buffer count = %d\n", alloc_buffer_max);
			}
		}

	// initialize buffers
	free_list = 0;
	full_list = 0;
	last_full = 0;
	current_fill_ref = 0;
	current_write_ref = 0;

	// we need at least two buffers
	result = add_new_buffer();
	if (result != 0) return result;
	result = add_new_buffer();
	if (result != 0) return result;

	// use one as the initial fill buffer
	result = get_new_buffer();
	if (result != 0) return result;
	output_ptr = output_buffer = current_fill_ref->addr;
	
	// initialize pthread objects
	result = pthread_mutex_init(&full_mutex, NULL);
	if (result != 0) {
		end_status();
		fprintf(stderr, "Unable to initialize full list mutex: %s\n", strerror(result));
		return result;
		}
	result = pthread_mutex_init(&free_mutex, NULL);
	if (result != 0) {
		end_status();
		fprintf(stderr, "Unable to initialize free list mutex: %s\n", strerror(result));
		return result;
		}
	result = pthread_mutex_init(&error_mutex, NULL);
	if (result != 0) {
		end_status();
		fprintf(stderr, "Unable to initialize error mutex: %s\n", strerror(result));
		return result;
		}
	result = pthread_cond_init(&full_cond, NULL);
	if (result != 0) {
		end_status();
		fprintf(stderr, "Unable to initialize full list condition variable: %s\n", strerror(result));
		return result;
		}
	result = pthread_cond_init(&free_cond, NULL);
	if (result != 0) {
		end_status();
		fprintf(stderr, "Unable to initialize free list condition variable: %s\n", strerror(result));
		return result;
		}

	// start the writer thread
	writer_done = 0;
	writer_error = 0;
	result = pthread_create(&writer_thread, NULL, writer, (void *)0);
	if (result != 0) {
		end_status();
		fprintf(stderr, "Unable to create new thread: %s\n", strerror(result));
		return result;
		}

	return 0;
}

int output_packet(unsigned char *packet)
{
	ssize_t remain, copylen;
	int result;
	
	// check if this packet needs to start a new file
	if ((packet_limit > 0) && 
		((output_file_seqnum < max_file_segments) || (max_file_segments == 0))) {
		if (write_limit_counter == packet_limit) {
			write_limit_counter = 0;
			// finish current file
			result = queue_output(0);
			if (result != 0) {
				return result;
				}
			// increment file
			if (output_file_seqnum == max_file_count) {
				end_status();
				return REACHED_FILE_LIMIT;
				}
			output_file_seqnum++;
			if (output_file_seqnum > MAX_OUTPUT_FILES) {
				end_status();
				fprintf(stderr, "Limit of %d output files reached.\n", MAX_OUTPUT_FILES);
				return 1;
				}
			}
		}
		
	// copy as much of packet as fits
	// copylen is the amount the did fit
	remain = out_blocksize - (output_ptr - output_buffer);
	copylen = 0;
	if (remain > 0) {
		copylen = TRANSPORT_PKT_SIZE;
		if (copylen > remain) copylen = remain;
		memcpy(output_ptr, packet, copylen);
		output_ptr += copylen;
		if (copylen == TRANSPORT_PKT_SIZE) {	// entire packet fit
			write_limit_counter++;
			return 0;
			}
		}

	// need to queue the buffer and get a new one
	result = queue_output(0);
	if (result != 0) return result;

	// now can store remainder in the buffer
	memcpy(output_buffer, packet + copylen, TRANSPORT_PKT_SIZE - copylen);
	output_ptr = output_buffer + (TRANSPORT_PKT_SIZE - copylen);
	write_limit_counter++;

	return 0;
}

int had_writer_error(void)
{
	int result;
	
	pthread_mutex_lock(&error_mutex);
	result = writer_error;
	pthread_mutex_unlock(&error_mutex);
	
	return result;
}

void end_writer(void)
{
	// write any remaining data
	queue_output(1);

	// signal writer to end
	pthread_mutex_lock(&full_mutex);
	writer_done = 1;
	pthread_mutex_unlock(&full_mutex);
	pthread_cond_signal(&full_cond);

	// wait for it to end
	pthread_join(writer_thread, NULL);

	// free storage
	free_buffer_list(free_list);
	free_buffer_list(full_list);
	free_buffer_ref(current_fill_ref);
	free_buffer_ref(current_write_ref);
	free_list = 0;
	full_list = 0;
	last_full = 0;
	current_fill_ref = 0;
	current_write_ref = 0;
}


int queue_output(char end)
{
	int result;
	ssize_t used;
	
	used  = output_ptr - output_buffer;
	// if buffer not used, can just use again
	if (used == 0) return 0;
	current_fill_ref->length = used;
	current_fill_ref->file_seqnum = output_file_seqnum;
	
	// get exclusive access to the full list
	pthread_mutex_lock(&full_mutex);

	// add current ref to full list
	if (last_full == 0) {
		full_list = current_fill_ref;
		}
	else {
		last_full->next = current_fill_ref;		
		}
	last_full = current_fill_ref;
	current_fill_ref->next = 0;

	// relinquish exclusive access
	pthread_mutex_unlock(&full_mutex);
	
	// signal buffer ready
	pthread_cond_signal(&full_cond);

	current_fill_ref = 0;

	if (end) {
		return 0;
		}
		
	// get a new output buffer (with exclusive access)
	pthread_mutex_lock(&free_mutex);
	result = get_new_buffer();
	pthread_mutex_unlock(&free_mutex);
	if (result != 0) return result;
	output_ptr = output_buffer = current_fill_ref->addr;
	
	return 0;
}

int get_new_buffer(void)
{
	int result;

	// get a new buffer if there isn't one
	if (free_list == 0) {
		if ((alloc_buffer_max > 0) && (alloc_buffer_count >= alloc_buffer_max)) {
			// block until a buffer becomes available
			if (input_is_char_dev) {
				end_status();
				fprintf(stderr, "Waiting for a free buffer (may cause buffer overrun.)\n");
				}
			else if (debug_output) {
				end_status();
				printf("Waiting for a free buffer.\n");
				}
			while (free_list == 0) {
				pthread_cond_wait(&free_cond, &free_mutex);
				}
			}
		else {
			result = add_new_buffer();
			if (result != 0) {
				return result;
				}
			}
		}

	// make a buffer from the free list the current buffer
	current_fill_ref = free_list;
	free_list = current_fill_ref->next;
	current_fill_ref->next = 0;
	
	return 0;
}


int add_new_buffer(void)
{
	buffer_ref *new_ref;
	buffer_ref *current_list;

	new_ref = (buffer_ref *)malloc(sizeof(buffer_ref));
	if (new_ref == 0) {
		end_status();
		fprintf(stderr, "Unable to allocate buffer ref of size %d\n", sizeof(buffer_ref));
		return ENOMEM;
		}
	
	new_ref->addr = (unsigned char *)malloc(out_blocksize);
	if (new_ref->addr == 0) {
		end_status();
		fprintf(stderr, "Unable to allocate buffer of size %d\n", out_blocksize);
		free(new_ref);
		return ENOMEM;
		}

	current_list = free_list;
	free_list = new_ref;
	new_ref->next = current_list;
	alloc_buffer_count++;

	if (debug_output) {
		end_status();
		printf("Added new output buffer: %d buffer(s), %dK\n", alloc_buffer_count,
				(alloc_buffer_count * out_blocksize)/1024);
		}

	return 0;
}

void free_buffer_list(buffer_ref *ref)
{
	buffer_ref *current;
	buffer_ref *next;
	
	current = ref;
	while (current != 0) {
		next = current->next;
		free_buffer_ref(current);
		current = next;
		}
}

void free_buffer_ref(buffer_ref *ref)
{
	if (ref == 0) return;
	if (ref->addr != 0) free(ref->addr);
	free(ref);
}

void *writer(void *args) {
	static int current_seqnum = 1;
	static char had_error = 0;
	char new_extension[MAX_EXTENSION_LENGTH];
	ssize_t remaining, write_actual;
	unsigned char *write_next;
	sigset_t mask;
	
	// mask interrupts we want the main thread to handle
	sigemptyset(&mask);
	sigaddset(&mask, SIGINT);
	sigaddset(&mask, SIGALRM);
	pthread_sigmask(SIG_SETMASK, &mask, NULL);

	while (1) {
		// wait for a new buffer to write
		pthread_mutex_lock(&full_mutex);
		while ((full_list == 0) && (writer_done == 0)) {
			pthread_cond_wait(&full_cond, &full_mutex);
			}

		// exit when end set and no buffers left
		if (full_list == 0) {	// must be writer_done
			close_output();
			return 0;
			}

		// make the buffer the current write buffer
		current_write_ref = full_list;
		full_list = full_list->next;
		if (full_list == 0) last_full = 0;

		// relinquish exclusive access
		pthread_mutex_unlock(&full_mutex);

		// skip to end if previous error
		if (had_error) goto free_buffer;

		// check if this buffer starts a new file
		if (current_write_ref->file_seqnum != current_seqnum) {
			current_seqnum = current_write_ref->file_seqnum;

			// close current file
			close_output();
	
			// open new output file
			sprintf(new_extension, "%d", output_file_seqnum);
			if (output_file_ext[0] != 0) {
				safe_strcat(new_extension, ".", -1, MAX_EXTENSION_LENGTH);
				safe_strcat(new_extension, output_file_ext, -1, MAX_EXTENSION_LENGTH);
				}
			strcpy(output_file, output_file_name);
			strcat(output_file, ".");
			strcat(output_file, new_extension);
			out_fd = open(output_file, open_flags, write_open_mode);
			if (out_fd == -1) {
				end_status();
				fprintf(stderr, "Unable to create \"%s\": %s\n", output_file, strerror(errno));
				had_error = 1;
				pthread_mutex_lock(&error_mutex);
				writer_error = 1;
				pthread_mutex_unlock(&error_mutex);
				goto free_buffer;
				}
			out_open = 1;
			}

		// write this buffer
		write_next = current_write_ref->addr;
		remaining = current_write_ref->length;
		while (remaining > 0) {	
			write_actual = write(out_fd, write_next, remaining);
			if (write_actual == 0) {
				errno = ENOSPC;
				}
			if (write_actual <= 0) {
				#ifdef EINTR
				if (errno == EINTR) {
					continue;
					}
				#endif
				end_status();
				fprintf(stderr, "Error writing output file %s: %s\n", output_file, strerror(errno));
				had_error = 1;
				pthread_mutex_lock(&error_mutex);
				writer_error = 1;
				pthread_mutex_unlock(&error_mutex);
				goto free_buffer;
				}	
			write_byte_count += write_actual;
			write_next += write_actual;
			remaining -= write_actual;
			}	

		// add this buffer to the free list

free_buffer:
		// get exclusive access
		pthread_mutex_lock(&free_mutex);

		current_write_ref->next = free_list;
		free_list = current_write_ref;
		current_write_ref = 0;

		// relinquish exclusive access
		pthread_mutex_unlock(&free_mutex);

		// signal buffer is available
		pthread_cond_signal(&free_cond);
		}
}

void close_output(void)
{
// *** called by writer thread ***
if (out_open) {
	end_status();
	printf("Finished file \"%s\"\n", output_file);
	close(out_fd);
	out_open = 0;
	}
}
