/*
 * NIST/DASE API Reference Implementation
 * parser.c
 *     Alan Mink
 *
 * Last changed on $Date$
 * Last changed on 2 May 2001.
 *  by Alan Mink
 */

/*
 *       NOTICE OF SOFTWARE ACKNOWLEDGMENT AND REDISTRIBUTION
 *
 * The software (named NDRI, for NIST/DASE API RI) provided herein is
 * released by the National Institute of Standards and Technology (NIST),
 * an agency of the U.S. Department of Commerce, Gaithersburg MD 20899, USA.
 * The software presented here is intended to be utilized for research
 * purposes only and bear no warranty, responsibility for a USER's use of
 * a NIST-derived software product or the results of such use.
 *
 * Please note that within the United States, copyright protection, under
 * Section 105 of the United States Code, Title 17, is not available for any
 * work of the United States Government and/or for any works created by United
 * States Government employees. USER acknowledges that this software contains
 * work which was created by NIST employees and is therefore in the public
 * domain and is not subject to copyright. The USER may use, distribute, or
 * incorporate this code or any part of it provided the USER acknowledges
 * this via an explicit acknowledgment of NIST-related contributions to the
 * USER's work. USER also agrees to acknowledge, via an explicit
 * acknowledgment, that modifications or alterations have been made
 * to this software by USER before redistribution.
 */

/*
 * parse ATSC (subset of MPEG-2) transport stream
 *
 *    - get a full Transport Stream pkt at a time
 *    - strip off Transport Stream header
 *    - IF ( PES pkt )
 *        IF ( PES with PID of "interest" ) forward PES pkt to handler
 *        ELSE  discard pkt
 *    - ELSE IF ( table of "interest" (i.e., ATSC/MPEG) )
 *        IF ( dup of existing table ) discard dups
 *        ELSE process table
 *          keep table in compressed form
 *          accumulate entire table, then "pipe" table to "API process"
 *
 */

/*
 * added input params via "argc" & "argv" for:
 *
 *      data source (i.e., tuner/dev or file) - default: std input
 *
 *      "pes" output dest (i.e., MPEG-2 A/V decoder) - default: discard
 *
 *      API output pipe (i.e., pipe for PSIP tables & data)
 *                                      - default:  discard
 *
 *      API input pipe (i.e., pipe for API to send cmds to the parser)
 *                                      - default:  none (ignore)
 */

/*
 * added "CMDs" from API simulation
 *
 *
 * Each back channel cmd is a single string of ASCII char terminated
 * by either the '\n' or '\0' char. Each field in a cmd is separated
 * by white space. Most cmd have 2 fields, the CMD (a numerical value)
 * and a qualifying field (qual #1 - mostly a numerical value, but can
 * contain ASCII symbols '+' & '-'), the open cmd has 3 (qual #2, which
 * is a string).
 *
 *         Summary:
 *                   CMD          ||     | qual | qual
 *                 Description    || CMD |  #1  |  #2
 *                 ---------------||-----|------|---------
 *                 Change_Channel ||  1  |   #  |  -
 *                 Open           ||  2  |  PID | dev_name
 *                 Start          ||  3  |  PID |  -
 *                 Stop           ||  4  |  PID |  -
 *                 Close          ||  5  |  PID |  -
 *
 *
 *       - Change_Channel
 *
 *                 1   #
 *
 *           where # is the new channel number or the '+' char for "next
 *           channel up" or '-' char for "next channel down"
 *
 *           Change_Channel function which does:
 *
 *              - delete all the dynamic PSIP tables & free their storage
 *              - clear out input buffers
 *              - set new "channel #"
 *              - simulate new channel via recycle input stream file
 *                      (go to beginning of current input file)
 *
 *       - Open dev and send PID stream to it
 *
 *                 2   PID   dev_name
 *
 *           where PID is the PID_number of the stream to be sent to "dev_name"
 *
 *       - Start sending PID stream to its assigned dev
 *
 *                 3   PID
 *
 *           where PID is the PID_number of the stream to be sent
 *           (the dev was originally assigned via Open)
 *           (Note: Open does both a dev assignment & a start)
 *
 *       - Stop sending PID stream to its assigned dev
 *
 *                 4   PID
 *
 *           where PID is the PID_number of the stream to be stopped
 *           (the dev was originally assigned via Open)
 *
 *       - Close PID stream & remove it from its assigned dev
 *
 *                 5   PID
 *
 *           where PID is the PID_number of the stream to be closed
 *           (the dev was originally assigned via Open). The PID to dev
 *           assignment is terminated and if no other PID is assigned to
 *           that device, it is closed.
 *
 */

/*

		     ______________________        ______________________
		     |                    |        |                    |
 dii_ddb_lst_head--->|  nxt_carousel_ptr  |------->|  nxt_carousel_ptr  |------>                      |
		<----|  bck_carousel_ptr  |<-------|  bck_carousel_ptr  |<-----                       |
		     |                    |        |                    |
		     |   dii_q_hdr_ptr    |---->   |   dii_q_hdr_ptr    |---->
   ----------------->|  dsi_buf_hdr_ptr   |---->   |  dsi_buf_hdr_ptr   |---->
   (from ts_control) |  dc_buf_hdr_ptr    |---->   |  dc_buf_hdr_ptr    |---->
		     |                    |        |                    |
		     |   Carousel_ptr     |--\     |    Carousel_ptr    |------>
		     |____________________|  |     |____________________|
					     |
					     |
				 /-----------/
				 |
				 |
		       _________\|/__________       ____________________
		       |                    |       |                    |
    carousel_head_ptr->|  nxt_carousel_ptr  |------>|  nxt_carousel_ptr  |------>                      |
		  <----|  bck_carousel_ptr  |<------|  bck_carousel_ptr  |<-----                       |
		       |                    |       |                    |
		       |   Download_id      |       |   Download_id      |
		       |   (Carousel_id)    |       |   (Carousel_id)    |
		       |                    |       |                    |
		       | # of DIIs (groups) |       | # of DIIs (groups) |
		       |                    |       |                    |
		       |       DII_ptr      |--\    |       DII_ptr      |--->
		       |____________________|  |    |____________________|
					       |
		    ___________________________/
		   /
		   |
		   |
		   |  DII hdrs (if a group, more than one,
		   |            then contiguously allocated)
		   |
		   |   _________________________________        _________________
		   \-->|     DII       |     DII       |        |     DII       |
		       |  Trans_id     |  Trans_id     |        |  Trans_id     |
		       |     ver       |     ver       |  ...   |     ver       |
		       | # of Modules  | # of Modules  |        | # of Modules  |
		       |   Module_ptr  |  Module_ptr   |        |  Module_ptr   |
		       |_______________|_______________|        |_______________|
			      |               |
			      |               |
		    __________/
		   /
		   |   _________________
		   \-->|  Module_id    |
		       |     size      |
		       |     ver       |
		       |  # of blks    |   data buffer (contiguously allocated)
		       |  not_done_flg |       ______________________________
		       |               |       |                            |
		       |   Buffer      |------>|                            |
		       |               |       |____________________________|
		       |               |
		       |               |   blk_hdr for blks (should be allocated
		       |               |        contiguously for easy indexing)
		       |               |       _______________       ________
		       |               |       |      |      |       |      |
		       |   Blk_ptr     |------>|      |      | ...   |      |
		       |    PTS        |       |______|______|       |______|
		       |_______________|
		       |   Module_id   |
		       |     size      |
		       |     ver       |
		       |  # of blks    |
		       |  not_done_flg |
		       |               |
		       |  Buffer       |------>
		       |               |
		       |   Blk_ptr     |------>
		       |    PTS        |
		       |_______________|

			      .
			      .
			      .
		       _________________
		       |   Module_id   |
		       |     size      |
		       |     ver       |
		       |  # of blks    |
		       |  not_done_flg |
		       |               |
		       |  Buffer       |------>
		       |               |
		       |   Blk_ptr     |------>
		       |    PTS        |
		       |_______________|

 */

#include <stdlib.h>
#include <sys/time.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>


#include "rt_parser.h"

extern int write_msg(char msgsource, char msgtype, char *format, ...);
extern void start_msg(char *format, ...);
extern void append_msg(char *format, ...);
extern int end_msg(char msgsource, char msgtype, char *format, ...);

//#define DEBUG_SUMMARY   1
/*
#define DEBUG_RCV       1       /* */

/*
#define DEBUG_RCV_PRT   1       /* */

/*
#define DEBUG_ERR_MSG   1       /* */
/*
#define DEBUG           1       /* */

/* ***
#define DEEP_DEBUG      1       /* */


#define ERROR                                 -1

#define SYNC_BYTE                           0x47
#define TRANSPORT_PKT_FIXED_HEADER_SIZE        4
#define TRANSPORT_PKT_SIZE                   188
#define PES_PKT_START_CODE_PREFIX       0x000001
#define PES_PKT_FIXED_HEADER_SIZE              6
#define MAX_SECTION_SIZE                    4096

#define GEN_TBL_HDR_SIZE                       9
#define STT_DATA_SIZE                          7
#define STT_MAX_SECTION_SIZE                1021
#define STT_BUF_SIZE                        1024
#define STT_MAX_SECTION_NUMBER                 1
#define MGT_MAX_SECTION_SIZE                4093
#define MGT_BUF_SIZE                        4096
#define MGT_MAX_SECTION_NUMBER                 1
#define MGT_ITEM_SIZE                         11
#define MGT_TBL_HDR_SIZE                      11
#define VCT_MAX_SECTION_SIZE                1021
#define VCT_MAX_SECTION_NUMBER               256
#define VCT_HDR_OFFSET                        10
#define VCT_ITEM_SIZE                         18
/*
#define CH_OFFSET                             32
#define CVCT_MAX_SECTION_SIZE               1021
#define CVCT_MAX_SECTION_NUMBER              256
#define RRT_MAX_SECTION_SIZE                1021
#define RRT_MAX_SECTION_NUMBER                 1
*/
#define PAT_BUF_SIZE                        4096
#define PAT_HEADER_BYTES                       8
#define PAT_OVERHEAD_BYTES                    12
#define PAT_ENTRY_BYTES                        4
#define CAT_BUF_SIZE                        4096
#define DCCT_BUF_SIZE                       4096
#define DCCSCT_BUF_SIZE                     4096
#define CRC_BYTES                              4

struct   timeval   time_val_struct;
struct   timeval   next_time;
struct   timeval   start_time;

/*  declare fct types */

struct          mgt_tbl_entry   *search_mgt_for_type();
struct          buf_hdr         *dequeue_buf();
struct          queue_hdr       *create_queue();


#define ACK             0x21
#define NACK            0x22
#define PCR             0x23

  /*
   * char buffer used to send ACK msgs to the API Simulation
   * via "api_input_stream"
   */
unsigned char ack_bytes[16]= {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

/* header for a Queue */

struct queue_hdr {
	struct   buf_hdr     *q_head_ptr   ; /* proc Q head             */
	struct   buf_hdr     *q_tail_ptr   ; /* proc Q tail             */
	int                  q_tot_bytes   ; /* total table size        */
	short                q_valid       ; /* 1 - table valid         */
					     /* 0 - table NOT valid     */
	short                q_ver         ; /* 1 - table valid         */
};

extern struct queue_hdr *eit_queue;
  /*
   * this control struct is indexed by PID & provides the steering,
   * references to parent structures & buffers, status & usage info
   * for each PID stream
   */
struct ts_control {
	unsigned short       ts_ctrl_type     ; /* used to steer processing */
	unsigned short       ts_ctrl_status   ; /* 0=inactive, 1=active */
						/* 2=ready */
	unsigned short       ts_orig_type     ; /* Orig type, prior to any DST */
	unsigned char        ts_orig_stream   ; /* Orig Stream Type */
	unsigned char        ts_cont_cntr     ; /* next expected Cont Cntr value */
	struct   buf_hdr     *ts_buf_hdr_ptr  ; /* accumulate Tbl Sect buf */
	struct   queue_hdr   ts_queue_hdr    ; /* Q of Tbl Sect buf to proc */
	union {
	   unsigned char        *ts_ctrl_ptr     ; /* general use ptr */
						   /* ptr back to MGT for*/
						   /*         std tables */
						   /* ptr to buf lst for */
						   /*     DST_NRT tables */
	   unsigned int         ts_ctrl_fd       ; /* fd for streams */
	} ts_ctrl_gen;
	unsigned int         ts_pkt_cnt       ; /* accumulates PID activity;*/
						/* info only -- useful to */
						/* find unexpected activity */
} ts_control[0x2000];     /* allocate one for each potenial PID */

/* define codes for "ts_ctrl_status" */

#define INACTIVE        0
#define ACTIVE          1
#define READY           2

/* define codes for "ts_ctrl_type " */

#define UNDEFINED       0
#define AUDIO_AC3       1
#define VIDEO           2
#define PAT             3
#define CAT             4
#define ATSC            5       /* 1FFB */
#define EIT             6
#define ETT             7
#define DET             8
#define LTST            9
#define PMT             10
#define DST_NRT         11
#define DII_DDB         12
#define ADDR_SECT       13
#define PES_IZED        14
#define PES_OUS         15
#define DSM_CC_ASYNC    16
#define PIPING_RAW      17
#define PIPING_COOKED   18
#define EMPTY           20      /* 1FFF */


char *ctrl_type_names[] = {
       "UNDEFINED",
       "AUDIO_AC3",
       "VIDEO",
       "PAT",
       "CAT",
       "ATSC",
       "EIT",
       "ETT",
       "DET",
       "LTST",
       "PMT",
       "DST_NRT",
       "DII_DDB",
       "ADDR_SECT",
       "PES_IZED",
       "PES_OUS",
       "DSM_CC_ASYNC",
       "PIPING_RAW",
       "PIPING_COOKED",
       "?",
       "NULL"
};

struct mgt_tbl_hdr        current_mgt_tbl = {0, 0, NULL, 0, NULL};

struct vct_hdr            current_tvct_hdr = { 0, 0, 0, 0, NULL, NULL };
struct vct_hdr            current_cvct_hdr = { 0, 0, 0, 0, NULL, NULL };

struct stt_data           stt_data;

unsigned int              stt_system_time = 0;  /* internal time that corresponds
						     to the lasest SST time */
char                      *sst_buf = NULL;      /* ptr to SST table buf */

struct pat_hdr            pat_hdr = {0, 0, 0, NULL};


/* struct buf_hdr          *buf_hdr_ptr;
*/

struct queue_hdr        stt_q_hdr         =  { NULL, NULL, 0, 0, 0};
struct queue_hdr        mgt_q_hdr         =  { NULL, NULL, 0, 0, 0};
struct queue_hdr        mgt_nxt_q_hdr     =  { NULL, NULL, 0, 0, 0};
struct queue_hdr        tvct_q_hdr        =  { NULL, NULL, 0, 0, 0};
struct queue_hdr        tvct_nxt_q_hdr    =  { NULL, NULL, 0, 0, 0};
struct queue_hdr        cvct_q_hdr        =  { NULL, NULL, 0, 0, 0};
struct queue_hdr        cvct_nxt_q_hdr    =  { NULL, NULL, 0, 0, 0};
struct queue_hdr        pat_q_hdr         =  { NULL, NULL, 0, 0, 0};
struct queue_hdr        cat_q_hdr         =  { NULL, NULL, 0, 0, 0};
struct queue_hdr        dcct_q_hdr        =  { NULL, NULL, 0, 0, 0};
struct queue_hdr        dccsct_q_hdr      =  { NULL, NULL, 0, 0, 0};


/*
 * declare a list for DII/DSI/DC/DDB table set for Data Carousels
 * the dii_ddb_lst items are dynamically created & deleted as needed
 * therefore one doesn't need to know how many to allocate in advance
 */

struct carousel *carousel_hdr = {NULL}; /* init a linked list of data carousels */

/* allocate a list for DII/DDB (carousel) items */

struct dii_ddb_lst    *dii_ddb_lst_head = NULL;

struct dii_ddb_lst {
	struct dii_ddb_lst  *nxt_dii_ddb_lst            ;
	struct dii_ddb_lst  *bck_dii_ddb_lst            ;
	unsigned short   dii_pid                        ;
	unsigned short   assoc_tag                      ;
	struct queue_hdr *dii_q_hdr                     ;
	struct buf_hdr   *dsi_buf_hdr                   ;
	struct buf_hdr   *dc_buf_hdr                    ;
	struct carousel  *carousel                      ;
};     

/*
 * declare a list for DST/NRT table pairs
 * the dst_nrt_q_lst items are dynamically created & deleted as needed
 * therefore one doesn't need to know how many to allocate in advance
 */

struct dst_nrt_q_lst  *dst_nrt_q_lst_head = {NULL}    ; /* head of list ptr */

struct dst_nrt_q_lst {
	struct dst_nrt_q_lst *nxt_q_lst_ptr           ;
	struct dst_nrt_q_lst *bck_q_lst_ptr           ;
	unsigned short       pid                      ;
	unsigned short       assoc_tag                ;
	struct queue_hdr     *dst_q_hdr               ;
	struct queue_hdr     *nrt_q_hdr               ;
};

/*
 * declare a list for ADDR_SECT info
 * the addr_sect_lst items are dynamically created & deleted as needed
 * therefore one doesn't need to know how many to allocate in advance
 */

struct addr_sect_lst  *addr_sect_lst_head = {NULL}    ; /* head of list ptr */

struct addr_sect_lst {
	struct addr_sect_lst *nxt_addr_sect_ptr       ;
	struct addr_sect_lst *bck_addr_sect_ptr       ;
	unsigned short       pid                      ;
	unsigned short       assoc_tag                ;
};


/* statistical data measurement/performance variables */

int     num_pes_pkt = 0;

int     num_stt       = 0;
int     num_rrt       = 0;
int     num_mgt       = 0;
int     num_vct       = 0;
int     num_dcct      = 0;
int     num_dccsct    = 0;
int     num_stt_comp  = 0;
int     num_rrt_comp  = 0;
int     num_mgt_comp  = 0;
int     num_vct_comp  = 0;
int     num_dcct_comp = 0;
int     num_dccsct_comp = 0;
int     cont_cntr_errors = 0;


/* --------------------------------------------------------------------- */

#define INPUT_BUF_SIZE           7*TRANSPORT_PKT_SIZE    /* size of input buf
					- 7 transport pkts (an arbitrarily
					chosen size), ea 188 bytes long */

#define  DATA_STREAM_EPOCH      50000000        /* freq (in terms of input
				       data bytes) to print debug stat msgs */

unsigned int  input_megs;          /* debug count of input data in MBytes */
unsigned int  input_epoch;         /* debug count of input data in Bytes
					  since last DATA_STREAM_EPOCH */

char *input_buf_ptr;               /* ptr to current data byte in input_buf */
char *input_buf_start;             /* ptr to start of input_buf */
char *input_buf_next_pkt;          /* ptr to next pkt in input_buf */
char *input_buf_end;               /* ptr to end of input_buf */
int   input_buf_bit;               /* current bit position within byte */

int   input_buf_cnt;               /* amount of data in input_buf */



int           current_chan;

/* stream ptr for output to API Simulation */

FILE  *api_pipe_stream;

/* input stream fd for input from API Simulation */

int   api_input_stream;

int   input_buf_stream;            /* handle for input_buf data stream */
int   output_audio_stream;         /* handle for output Audio stream */
int   output_video_stream;         /* handle for output Video stream */
FILE  *output_time_stream;         /* handle for output PCR time stream */

short first_audio_stream = 1;      /* 1 => no PID yet allocated to stream */
short first_video_stream = 1;      /* 1 => no PID yet allocated to stream */

short pes_header_video_only = 0;      /* 1 => YES, 0 => NO */
short pes_header_audio_only = 0;      /* 1 => YES, 0 => NO */
short rate_limit_input_stream = 1;    /* 1 => YES, 0 => NO */

char  input_buf[INPUT_BUF_SIZE];   /* input_buf */

char  *input_stream_dev_ptr = NULL;      /* -is */
char  *input_fifo_dev_ptr   = NULL;      /* -if */
char  *output_fifo_dev_ptr  = NULL;      /* -of */
char  *output_audio_dev_ptr = NULL;      /* -oa */
char  *output_video_dev_ptr = NULL;      /* -ov */
char  *output_time_dev_ptr  = NULL;      /* -ot */

char  input_stream_dev[128] = "";      /* -is */
char  input_fifo_dev[128]   = "";      /* -if */
char  output_fifo_dev[128]  = "";      /* -of */
char  output_audio_dev[128] = "";      /* -oa */
char  output_video_dev[128] = "";      /* -ov */
char  output_time_dev[128]  = "";      /* -ot */

struct dev_tbl {
	int             dev_fd;
	unsigned short  type_of_stream;
	unsigned short  pid_cnt;
	unsigned short  pid_lst[10];
	char            dev_name[128];
} dev_tbl[50];

unsigned int    dev_tbl_cnt = 0;  /* entries in dev_tbl */

/* --------------------------------------------------------------------- */

/*
 * this routine re-initials input buffers & their associated variables
 *
 */

input_buf_reinit()
{
     input_buf_start = &input_buf[0];   /* init buf ptr */

     input_buf_ptr = input_buf_start;           /* init current buf ptr */
     input_buf_next_pkt = input_buf_start;      /* init next buf ptr */
     input_buf_cnt = 0;                         /* init byte offset from
							beginning of pkt */
     input_buf_end = input_buf_start + input_buf_cnt-1; /* init end of input */
							/*   buf ptr */
}

/* --------------------------------------------------------------------- */

/*
 * this routine initials input buffers & their associated variables
 *
 * opening an input device should be done here
 */

input_buf_init()
{
     input_megs = 0;    /* init debug variables */
     input_epoch= 0;

     input_buf_start = &input_buf[0];   /* init buf ptr */

     input_buf_ptr = input_buf_start;           /* init current buf ptr */
     input_buf_next_pkt = input_buf_start;      /* init next buf ptr */
     input_buf_cnt = 0;                         /* init byte offset from
							beginning of pkt */
     input_buf_end = input_buf_start + input_buf_cnt-1; /* init end of input */
							/*  buf ptr */
/* source "dev" for input bitstream
 */
     if ( input_stream_dev_ptr == NULL )
     {
	 input_buf_stream = 0;                  /* init input stream to the */
     }                                          /*  default - std input */
     else
     {                  /* open specified file as READ_ONLY */
	 if ( (input_buf_stream = open(input_stream_dev_ptr,O_RDONLY)) == -1 )
	 {
		write_msg('P', 'E', "**ERROR in openning input_stream_dev <%s>, Terminating",
			input_stream_dev_ptr);
		exit(0);
	 }
     }
/*
 * output pipe to API simulator
 *   api_pipe_stream = FILE *fopen(const char *filename, const char *type);
 */
     if ( output_fifo_dev_ptr == NULL )
     {
	 //write_msg('P', 'I', "***NO  Output_FIFO_to_API specified, API output will be DISCARDED");
     }
     else
     {
	 if ( (api_pipe_stream = fopen(output_fifo_dev_ptr,  "w")) == NULL )
	 {
	    write_msg('P', 'E', "***ERROR opening Stream <dase_api_simul> in input_buf_init()");
	 }
	 else
	    write_msg('P', 'I', "Stream <dase_api_simul> opened successfully");
     }

/*
 * output pipe to time preocess
 */
     if ( output_time_dev_ptr == NULL )
     {
	 //write_msg('P', 'I', "***NO  Output_TIME_to_Proc specified, PCR time output will be DISCARDED");
     }
     else
     {
	 if ( (output_time_stream = fopen(output_time_dev_ptr,  "w")) == NULL )
	 {
	    write_msg('P', 'E', "***ERROR opening Stream <output_time_stream> in input_buf_init()");
	 }
	 else
	    write_msg('P', 'I', "Stream <output_time_stream> opened successfully");
     }

/*
 * input pipe from API simulator
 *   api_pipe_stream = FILE *fopen(const char *filename, const char *type);
 */
     if ( input_fifo_dev_ptr == NULL )
     {
	// write_msg('P', 'I', "***NO  Input_FIFO_from_API specified, no API interaction will occur");
	 api_input_stream = -1;
     }
     else
     {
	 if ( strcasecmp(input_fifo_dev_ptr,"std_in") == 0 )
	 {
		api_input_stream = 0;  /* recognize std_in, as a special case */
/* don't need to make fd=0 NON_BLOCKING, since one should assume
/* that any input would be a from a redirected file which shouldn't
/* block & on EOF would act similar to the NON_BLOCKING option, */
/* returning an error code, just a different one from NON_BLOCKING */
/* Input from the cmd line, would be an obvious problem !! */
/*              if ( fcntl(0, F_SETFL, O_NONBLOCK) < 0 )
		   write_msg('P', 'E', "**ERROR in setting STD_IN to NON_BLOCKING");
 */      }
	 else if ( (api_input_stream = open(input_fifo_dev_ptr,O_RDONLY | O_NONBLOCK)) == -1 )
	 {              /* open specified file as READ_ONLY & Non_BLOCKING */
		write_msg('P', 'E', "**ERROR in openning api_input_stream_dev <%s>, No API Input",
			input_fifo_dev_ptr);
		input_fifo_dev_ptr = NULL;
	 }
     }

/* output "dev" for Audio bitstream
 */
     if ( output_audio_dev_ptr == NULL )
     {
	 output_audio_stream = -1;              /* init Audio dev */
     }                                          /*  default - none */
     else
     {                  /* open specified file as WRITE_ONLY */
	 if ( (output_audio_stream = open(output_audio_dev_ptr,
			 O_WRONLY | O_NONBLOCK | O_CREAT, 0666)) == -1 )
	 {
	      write_msg('P', 'E', "**ERROR in openning output_Audio_dev <%s>, Terminating",
		      output_audio_dev_ptr );
	      exit(0);
	 }
	 else
	 {
	      write_msg('P', 'I', "Output_Audio_Dev <%s> opened successfully", output_audio_dev_ptr );
	      strcpy(dev_tbl[dev_tbl_cnt].dev_name, output_audio_dev_ptr);
	      dev_tbl[dev_tbl_cnt].dev_fd = output_audio_stream;
	      dev_tbl[dev_tbl_cnt].pid_cnt = 0;
	      dev_tbl[dev_tbl_cnt].type_of_stream = AUDIO_AC3;
	      dev_tbl_cnt++;
	 }
     }

/* output "dev" for Video bitstream
 */
     if ( output_video_dev_ptr == NULL )
     {
	 output_video_stream = -1;              /* init Video dev */
     }                                          /*  default - none */
     else
     {                  /* open specified file as WRITE_ONLY */
	 if ( (output_video_stream = open(output_video_dev_ptr,
			O_WRONLY | O_NONBLOCK | O_CREAT, 0666)) == -1 )
	 {
	      write_msg('P', 'E', "**ERROR in openning output_Video_dev <%s>, Terminating",
		      output_video_dev_ptr );
	      exit(0);
	 }
	 else
	 {
	      write_msg('P', 'I', "Output_Video_Dev <%s> opened successfully", output_video_dev_ptr );
	      strcpy(dev_tbl[dev_tbl_cnt].dev_name, output_video_dev_ptr);
	      dev_tbl[dev_tbl_cnt].dev_fd = output_video_stream;
	      dev_tbl[dev_tbl_cnt].pid_cnt = 0;
	      dev_tbl[dev_tbl_cnt].type_of_stream = VIDEO;
	      dev_tbl_cnt++;
	 }
     }

}

/* --------------------------------------------------------------------- */

char  cmd_input[256];
char  *cmd_input_ptr = cmd_input;

proc_api_input_cmd()
{
    char          *c_ptr;
    char          cmd_strg1[256], cmd_strg2[256];
    unsigned int  num1, num2;
    int           request_code, chan, new_chan;
    struct  dst_nrt_q_lst  *q_lst_ptr, *next_q_lst_ptr;
    struct  ts_control     *ts_control_ptr;
    struct  dii_ddb_lst     *dii_ddb_lst_ptr, *next_dii_ddb_lst_ptr;
    struct  addr_sect_lst   *addr_sect_lst, *next_addr_sect_ptr;

    if ( read(api_input_stream,cmd_input_ptr, 1) == 1 )
    {
	c_ptr = cmd_input_ptr;
	if ( *c_ptr == '\0' || *c_ptr == '\n' )
	{
	    write_msg('P', 'I', "EMPTY API_CMD, IGNORED");
	    return;
	}
	while ( read(api_input_stream,++c_ptr, 1) == 1 )
	{
	    if ( *c_ptr == '\0' || *c_ptr == '\n' )
	    {
		*c_ptr = '\0';
		break;
	    }
	}
	num1 = -1;
	sscanf(cmd_input_ptr,"%d %s %s", &num1, cmd_strg1, cmd_strg2);

/*      write_msg('P', 'D', "API_CMD_DECODED=%d, %s, %s", num1, cmd_strg1, cmd_strg2);
*/
	/* process requested action */

       if ( num1 == 1 )
       {     /* change channel */
	   if ( *cmd_strg1 == '+' )
	   {
	      new_chan = -10;
	      write_msg('P', 'I', "Change channel UP"); /* up to next channel */
	   }
	   else if ( *cmd_strg1 == '-' )
	   {
	      new_chan = -20;
	      write_msg('P', 'I', "Change channel DOWN"); /* down to next channel */
	   }
	   else
	   {
	      sscanf(cmd_strg1,"%d", &num2); /* select channel */
	      write_msg('P', 'I', "Change channel to %d", num2);
	      new_chan = num2;
	   }
	   if ( current_chan == new_chan )
	   {
	       ack_bytes[0] = ACK;
	       ack_bytes[1] = 0x1;
	       ack_bytes[2] = (new_chan>>8)&0xFF;
	       ack_bytes[3] = (new_chan   )&0xFF;
	   }
	   else if ( new_chan < 0 )
	   {
	       ack_bytes[0] = NACK;
	       ack_bytes[1] = 0x1;
	       ack_bytes[2] = (current_chan>>8)&0xFF;
	       ack_bytes[3] = (current_chan   )&0xFF;
	   }
	   else if ( (chan=change_channel(new_chan)) > 0 )
	   {
		/* delete root tables */

	       delete_mgt_table();
	       delete_pat();

		/* delete DST/NRT tables */

		q_lst_ptr = dst_nrt_q_lst_head;
		while ( q_lst_ptr != NULL )
		{
		    next_q_lst_ptr = q_lst_ptr->nxt_q_lst_ptr;
			/* clear assoc "ts_control" entry */
		    ts_control_ptr = &ts_control[q_lst_ptr->pid];
		    ts_control_ptr->ts_ctrl_gen.ts_ctrl_ptr = NULL;
		    ts_control_ptr->ts_ctrl_type = UNDEFINED;
		    ts_control_ptr->ts_ctrl_status = 0;
		    if ( ts_control_ptr->ts_buf_hdr_ptr != NULL )
		    {
			if ( ts_control_ptr->ts_buf_hdr_ptr->buf_ptr != NULL )
			   free( ts_control_ptr->ts_buf_hdr_ptr->buf_ptr );
			free( ts_control_ptr->ts_buf_hdr_ptr );
		    }
		    clear_queue( ts_control_ptr->ts_queue_hdr);

		    delete_dst_nrt_q_item( q_lst_ptr );
		    q_lst_ptr = next_q_lst_ptr;
		}

		/* delete carousels & associated signalling tables   */
		/* since deallocating each item in list sequentially */
		/* no need to delink the items                       */

		dii_ddb_lst_ptr = dii_ddb_lst_head;
		while ( dii_ddb_lst_ptr != NULL )
		{
		    next_dii_ddb_lst_ptr = dii_ddb_lst_ptr->nxt_dii_ddb_lst;
		    if ( dii_ddb_lst_ptr->dc_buf_hdr != NULL )
		    {
			if ( dii_ddb_lst_ptr->dc_buf_hdr->buf_ptr != NULL )
			   free( dii_ddb_lst_ptr->dc_buf_hdr->buf_ptr );
			free( dii_ddb_lst_ptr->dc_buf_hdr );
		    }
		    if ( dii_ddb_lst_ptr->dsi_buf_hdr != NULL )
		    {
			if ( dii_ddb_lst_ptr->dsi_buf_hdr->buf_ptr != NULL )
			   free( dii_ddb_lst_ptr->dsi_buf_hdr->buf_ptr );
			free( dii_ddb_lst_ptr->dsi_buf_hdr );
		    }
		    if ( dii_ddb_lst_ptr->dii_q_hdr != NULL )
		    {
			clear_queue( dii_ddb_lst_ptr->dii_q_hdr );
			free( dii_ddb_lst_ptr->dii_q_hdr );
		    }
		    delete_carousel( dii_ddb_lst_ptr->carousel );
			/* clear assoc "ts_control" entry */
		    ts_control_ptr = &ts_control[dii_ddb_lst_ptr->dii_pid];
		    ts_control_ptr->ts_ctrl_gen.ts_ctrl_ptr = NULL;
		    ts_control_ptr->ts_ctrl_type = UNDEFINED;
		    ts_control_ptr->ts_ctrl_status = 0;
		    if ( ts_control_ptr->ts_buf_hdr_ptr != NULL )
		    {
			if ( ts_control_ptr->ts_buf_hdr_ptr->buf_ptr != NULL )
			   free( ts_control_ptr->ts_buf_hdr_ptr->buf_ptr );
			free( ts_control_ptr->ts_buf_hdr_ptr );
		    }
		    clear_queue( ts_control_ptr->ts_queue_hdr);

		    free( dii_ddb_lst_ptr ); /* deallocate list item */
		    dii_ddb_lst_ptr = next_dii_ddb_lst_ptr;
		}
		dii_ddb_lst_head = NULL; /* init list prt now that list is empty */

		/* delete addr_sect list                             */
		/* since deallocating each item in list sequentially */
		/* no need to delink the items                       */

		addr_sect_lst = addr_sect_lst_head;
		while ( addr_sect_lst != NULL )
		{
		    next_addr_sect_ptr = addr_sect_lst->nxt_addr_sect_ptr;
			/* clear assoc "ts_control" entry */
		    ts_control_ptr = &ts_control[addr_sect_lst->pid];
		    ts_control_ptr->ts_ctrl_gen.ts_ctrl_ptr = NULL;
		    ts_control_ptr->ts_ctrl_type = UNDEFINED;
		    ts_control_ptr->ts_ctrl_status = 0;
		    if ( ts_control_ptr->ts_buf_hdr_ptr != NULL )
		    {
			if ( ts_control_ptr->ts_buf_hdr_ptr->buf_ptr != NULL )
			   free( ts_control_ptr->ts_buf_hdr_ptr->buf_ptr );
			free( ts_control_ptr->ts_buf_hdr_ptr );
		    }
		    clear_queue( ts_control_ptr->ts_queue_hdr);

		    free( addr_sect_lst ); /* deallocate list item */
		    addr_sect_lst = next_addr_sect_ptr;
		}
		addr_sect_lst_head = NULL; /* init list prt now that list is empty */



	       first_video_stream = 1;
	       first_audio_stream = 1;
	       input_buf_reinit();
	       ctrl_init();
	       current_chan = chan;
	       ack_bytes[0] = ACK;
	       ack_bytes[1] = 0x1;
	       ack_bytes[2] = (chan>>8)&0xFF;
	       ack_bytes[3] = (chan   )&0xFF;
	   }
	   else
	   {
	       ack_bytes[0] = NACK;
	       ack_bytes[1] = 0x1;
	       ack_bytes[2] = (current_chan>>8)&0xFF;
	       ack_bytes[3] = (current_chan   )&0xFF;
	   }
	   pipe_ack2api_proc(ack_bytes);
       }
       else if ( (num1>1) && (num1<6) )
       {
	    sscanf(cmd_strg1,"%d", &num2); /* convert PID */
	   if ( num1 == 2 )
	   {     /* open dev for PID */
	       request_code = open_pid_output_dev(num2,cmd_strg2);
	      write_msg('P', 'I', "Open PID %d on %s", num2, cmd_strg2);
	   }
	   else if ( num1 == 3 )
	   {     /* start PID stream */
	       request_code = start_dev_player(num2);
	      write_msg('P', 'I', "Start PID %d", num2);
	   }
	   else if ( num1 == 4 )
	   {     /* stop PID stream */
	       request_code = stop_dev_player(num2);
	      write_msg('P', 'I', "Stop PID %d", num2);
	   }
	   else if ( num1 == 5 )
	   {     /* close PID stream */
	       request_code = close_pid_output_dev(num2);
	      write_msg('P', 'I', "Close PID %d", num2);
	   }
	   if ( request_code != -1 )
	   {
	       ack_bytes[0] = ACK;
	       ack_bytes[1] = num1;
	       ack_bytes[2] = (num2>>8)&0xFF;
	       ack_bytes[3] = (num2   )&0xFF;
	   }
	   else
	   {
	       ack_bytes[0] = NACK;
	       ack_bytes[1] = num1;
	       ack_bytes[2] = (num2>>8)&0xFF;
	       ack_bytes[3] = (num2   )&0xFF;
	   }
	   pipe_ack2api_proc(ack_bytes);
       }
       else
       {     /* unknown CMD */
	   write_msg('P', 'I', "UNKNOWN command=(%d,%s) from API, IGNORED",
			     num1, cmd_strg1);
       }
    }
}

/* --------------------------------------------------------------------- */

/* close an output stream "dev"  for a PID
 */

close_pid_output_dev(pid)
unsigned int pid;
{
     int loc, i, j, k;

     stop_dev_player(pid);
     for ( loc=0; loc<dev_tbl_cnt; loc++)
     {
	for ( i=0; i<dev_tbl[loc].pid_cnt; i++)
	{
	    if ( dev_tbl[loc].pid_lst[i] == pid )
	    {
		ts_control[pid].ts_ctrl_gen.ts_ctrl_fd = -1;
		if ( dev_tbl[loc].pid_cnt > 1 )
		{
		  /* delete PID from dev list, other PID still assoc */
		    for ( j=i; j<dev_tbl[loc].pid_cnt-1; j++)
			dev_tbl[loc].pid_lst[j] = dev_tbl[loc].pid_lst[j+1];
		    dev_tbl[loc].pid_cnt--;
		    return (loc);
		}
		else
		{
		  /* delete dev, PID was the only stream using dev */
		    dev_tbl[loc].pid_cnt = 0;
		    close (dev_tbl[loc].dev_fd);
		    dev_tbl[loc].dev_fd = -1;
		    if ( dev_tbl_cnt == 1 )
		       dev_tbl_cnt = 0;  /* only dev in tbl, tbl now empty */
		    else
		    {   /* remove dev from tbl */
		       for ( k=loc; k<dev_tbl_cnt-1; k++)
		       {
			   dev_tbl[k].dev_fd = dev_tbl[k+1].dev_fd;
			   dev_tbl[k].type_of_stream = dev_tbl[k+1].type_of_stream;
			   dev_tbl[k].pid_cnt = dev_tbl[k+1].pid_cnt;
			   strcpy(dev_tbl[k].dev_name, dev_tbl[k+1].dev_name);
			   for ( j=0; j<dev_tbl[k+1].pid_cnt; j++)
			       dev_tbl[k].pid_lst[j] = dev_tbl[k+1].pid_lst[j];
		       }
		    }
		    return (0);
		}
	    }
	}
     }
     return (-1);
}

/* --------------------------------------------------------------------- */

/* open an output stream "dev"  for a PID
 */

open_pid_output_dev(pid,dev_name)
unsigned int pid;
char         *dev_name;
{
     int loc, i, j, fd;

     for ( loc=0; loc<dev_tbl_cnt; loc++)
     {
	for ( i=0; i<dev_tbl[loc].pid_cnt; i++)
	{
	    if ( dev_tbl[loc].pid_lst[i] == pid )
	    {    /* PID already allocated to a dev */
		if ( strcmp(dev_name,dev_tbl[loc].dev_name) == 0 )
		{       /* the allocation is a dup of the requested one */
			/* do nothing */
		    write_msg('P', 'I', "PID 0x%x & dev %s pair already allocated",
			 pid, dev_name );
		    return (loc);
		}
		else
		{       /* the allocation is for a different dev */
			/* delete PID assoc with old dev */
		    write_msg('P', 'I', "PID 0x%x allocated to dev %s, Changing to dev %s",
			 pid, dev_tbl[loc].dev_name, dev_name );
		    for ( j=i; j<dev_tbl[loc].pid_cnt-1; j++)
			dev_tbl[loc].pid_lst[j] = dev_tbl[loc].pid_lst[j+1];
		    dev_tbl[loc].pid_cnt--;
		    ts_control[pid].ts_ctrl_gen.ts_ctrl_fd = -1;
		    loc = dev_tbl_cnt;
		    break;
		}
	    }
	}
	   /*
	    * add dev to dev_tbl, if it its not there already,
	    * & then add PID to its list & set fd in pid ctrl struct
	    */
	if ( (loc=open_output_dev(dev_name)) > -1 )
	{
	     dev_tbl[loc].pid_lst[dev_tbl[loc].pid_cnt++] = pid;
	     ts_control[pid].ts_ctrl_gen.ts_ctrl_fd = dev_tbl[loc].dev_fd;
	}
     }
     return (loc);
}

/* --------------------------------------------------------------------- */

/* open an output stream "dev"
 */

open_output_dev(dev_name)
char *dev_name;
{
     int loc, fd;

     if (  (loc=search_dev_tbl(dev_name)) == -1 )
     {
	/* open & add to table, set file not to block on I/O  */
	  if ( (fd=open(dev_name, O_WRONLY | O_NONBLOCK | O_CREAT )) == -1 )
	  {
		write_msg('P', 'E', "**ERROR in opening output_Stream_dev <%s>", dev_name);
	  }
	  else
	  {
	      loc = dev_tbl_cnt++;
	      strcpy(dev_tbl[loc].dev_name, dev_name);
	      dev_tbl[loc].dev_fd = fd;
	      dev_tbl[loc].pid_cnt = 0;
	      dev_tbl[loc].type_of_stream = 0; /* unknown */
	  }
     }
	/* already in table */
     return( loc );
}

/* --------------------------------------------------------------------- */

/* search dev table for dev name
 */

search_dev_tbl(dev_name)
char *dev_name;
{
     int loc, i;

     loc = -1;
     for ( i=0; i<dev_tbl_cnt; i++)
     {
	if (  strcmp(dev_name,dev_tbl[i].dev_name) == 0 )
	     return( i );
     }
     return( loc );
}

/* --------------------------------------------------------------------- */

/*
 * this routine sets the PID stream status to READY
 */

start_dev_player(pid)
unsigned int pid;
{
	short type;

	type =  ts_control[pid].ts_ctrl_type;
	if ( (type==VIDEO) || (type==AUDIO_AC3) || (type==PES_IZED)
			    || (type==PES_OUS) || (type==DII_DDB) )
	{
	    if ( ts_control[pid].ts_ctrl_gen.ts_ctrl_fd > 0 )
	    {
		ts_control[pid].ts_ctrl_status = READY;
		return (1);
	    }
	}
	return (-1);
}

/* --------------------------------------------------------------------- */

/*
 * this routine sets the PID stream status to READY
 */

stop_dev_player(pid)
unsigned int pid;
{
	short type;

	type =  ts_control[pid].ts_ctrl_type;
	if ( (type==VIDEO) || (type==AUDIO_AC3) || (type==PES_IZED)
			    || (type==PES_OUS) || (type==DII_DDB) )
	{
	    if ( ts_control[pid].ts_ctrl_gen.ts_ctrl_fd > 0 )
	    {
		ts_control[pid].ts_ctrl_status = INACTIVE;
		return (1);
	    }
	}
	return (-1);
}

/* ---------------------------------------------------------------------- */

/*
 * This routine simulates changing a channel by repositioning the
 * input_stream to its beginning
 *
 * The next simul step would be to setup a list of bitstream files
 * and associate a "channel number" with each of them. Then to change
 * to a channel simply close the current file & open the one corresponding
 * to that channel.
 */

change_channel(new_chan)
int  new_chan;
{
	int chan;

	if ( new_chan == -10 )  /* UP */
	{
	   chan = current_chan + 1;
	   if ( chan > 99 )
		chan = 2;
	}
	else if ( new_chan == -20 )  /* DOWN */
	{
	   chan = current_chan - 1;
	   if ( chan < 2 )
		chan = 2;
	}
	else   /* change to new channel */
	{
	   if ( (new_chan<2) || (new_chan>99) )
		return (-1);            /* ERROR */
	   chan = new_chan;
	}
	if ( lseek(input_buf_stream, 0, SEEK_SET) == -1 )
		write_msg('P', 'E', "ERROR on setting <lseek> on Input_Stream, Ignored");
	prt_summary();
	return (chan);
}

/* --------------------------------------------------------------------- */

/*
 *    Format of "pipe" info sent from parser to API Simulation
 *
 * 16 bytes FIFO header
 *
 * fifoHeader {
 *   headerType             8 bits         // data or table
 *
 *   if (headerType == TABLE) {
 *
 *     table_id             8 bits
 *     reserved             3 bits = 0
 *     pid                 13 bits
 *     table_id_extension  16 bits
 *     table_type          16 bits
 *     reserved            12 bits = 0
 *     tableLength         20 bits
 *     reserved            32 bits = 0
 *
 *   } else if (headerType == DATA) {
 *
 *     dataId               8 bits         // DSM-CC Carousel only for now
 *     reserved             3 bits = 0
 *     pid                 13 bits
 *     downloadId          32 bits
 *     reserved             8 bits = 0
 *     moduleVersion        8 bits
 *     moduleId            16 bits
 *     moduleSize          32 bits
 *
 *   } else if (headerType == ACK || headerType == NACK) {
 *
 *     command              8 bits
 *     if (command == CHANGE_CHANNEL) {
 *       newChannel        16 bits
 *     } else {
 *       reserved           3 bits = 0
 *       targetPid         13 bits
 *     }
 *     reserved            96 bits = 0
 *
 *   } else if (headerType == PCR) {
 *
 *     adaptationFlags      8 bits
 *     reserved             3 bits = 0
 *     pid                 13 bits
 *     reserved            22 bits = 0
 *     if (adaptationFlags & 0x10 == '1') {
 *       pcr               42 bits
 *     } else {
 *       reserved          42 bits = 0
 *     }
 *     reserved            32 bits = 0
 *
 *   }
 * }
 *
 * if (fifoHeader.headerType = DATA && fifoHeader.dataId = DATA_CAROUSEL) {
 *   reserved              23 bits = 0
 *   pts_is_valid           1 bit
 *   reserved               7 bits = 0
 *   pts                   33 bits
 * }
 *
 * fifoData {
 *   // tableLength or moduleSize bytes of data
 * }
 *
 *
 * CONSTANTS:
 * ----------
 *
 * headerType: 0x01 -> TABLE
 *             0x02 -> DATA
 *             0x21 -> ACK
 *             0x22 -> NACK
 *             0x23 -> PCR
 *             other values are reserved for future use
 *
 * dataId:  1 -> DSM-CC Data Carousel
 *          other values are reserved for future use
 *
 * command: 0x31 ("1") -> CHANGE_CHANNEL
 *          0x32 ("2") -> OPEN
 *          0x33 ("3") -> START
 *          0x34 ("4") -> STOP
 *          0x35 ("5") -> CLOSE
 */

/* --------------------------------------------------------------------- */

/*
 * pipe PCR in our own 16 byte header, to the
 * a separate procees, NOT the API simulator process via a named pipe.
 *
 * The format of the PCR pkt is:
 *
 *              8 bits     8 bits     8 bits   8 bits
 *             __________|_________|________|_________
 *            | <PCR:=23> < flags > <      PID      > |
 *            |---------------------------------------|
 *            | <         N/A          >< PCR[42:33] >|
 *            |---------------------------------------|
 *            | <           PCR[32:1]               > |
 *            |---------------------------------------|
 *            | <               N/A                 > |
 *            -----------------------------------------
 *
 */

unsigned char pcr_bytes[16]= {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

pipe_pcr2api_proc(flags,pcr_hi,pcr_lo,pid)
unsigned int  flags;
unsigned int  pcr_hi;
unsigned int  pcr_lo;
unsigned int  pid;
{       /* pipe PCR buffer via FIFO */
/*
	int pkt[4];

	pkt[0] = (PCR<<24) | ((flags&0xFF)<<16) | (pid&0x1FFFF);
	pkt[1] = pcr_hi;
	pkt[2] = pcr_lo;
	pkt[3] = 0;
*/
	pcr_bytes[0] =  PCR;
	pcr_bytes[1] =  flags&0xFF;
	pcr_bytes[2] =  (pid>>8)&0x1F;
	pcr_bytes[3] =   pid&0xFF;
	pcr_bytes[4] =  0x0;
	pcr_bytes[5] =  0x0;
	pcr_bytes[6] =  (pcr_hi>>8)&0xFF;
	pcr_bytes[7] =  (pcr_hi   )&0xFF;
	pcr_bytes[8] =  (pcr_lo>>24)&0xFF;
	pcr_bytes[9] =  (pcr_lo>>16)&0xFF;
	pcr_bytes[10] = (pcr_lo>> 8)&0xFF;
	pcr_bytes[11] = (pcr_lo    )&0xFF;

	if ( output_time_dev_ptr == NULL )
		return;         /* no output fifo pipe specified
				   therefore, don't write anything to it */
/*
 * pipe our header
 */
	fwrite(pcr_bytes,1,16, output_time_stream);
/*      write_msg('P', 'D', "pipe PCR = 0x%x", pkt[0]);
*/

/*
 * flush the buffer contents into the pipe, now
 */
	fflush(output_time_stream);
	return;
}

/* --------------------------------------------------------------------- */

/*
 * pipe ACK/NACK in our own 16 byte header, to the
 * API simulator process via a named pipe.
 *
 * The format of the ACK/NACK pkt is:
 *
 *              8 bits     8 bits     8 bits   8 bits
 *             __________|_________|________|_________
 *            | <ACK:=21> <  CMD  > <  new_channel  > |
 *            |<NACK:=22>           <      PID      > |
 *            |---------------------------------------|
 *            | <               N/A                 > |
 *            |---------------------------------------|
 *            | <               N/A                 > |
 *            |---------------------------------------|
 *            | <               N/A                 > |
 *            -----------------------------------------
 *
 * CMD  = 1 --> change channel, response is new current channel
 *                              if request failed new current channel is
 *                              the unchanged current channel
 * CMD  = 2 --> Open DEV for PID, response is PID sent
 * CMD  = 3 --> Start PID       , response is PID sent
 * CMD  = 4 --> Stop  PID       , response is PID sent
 * CMD  = 5 --> Close PID       , response is PID sent
 */

pipe_ack2api_proc(ack_buf)
unsigned char  ack_buf[];
{       /* pipe ACKO/NACK buffer via FIFO */

	if ( output_fifo_dev_ptr == NULL )
		return;         /* no output fifo pipe specified
				   therefore, don't write anything to it */
/*
 * pipe our header
 */
	fwrite(ack_buf,1,16, api_pipe_stream);
/*      write_msg('P', 'D', "pipe ACK/NACK = 0x%x", ack_buf[0]);
*/

/*
 * flush the buffer contents into the pipe, now
 */
	fflush(api_pipe_stream);
	return;
}

/* --------------------------------------------------------------------- */

/*
 * pipe a table Queue, preceded by our own 16 byte header, to the
 * API simulator process via a named pipe.
 *
 * The table buffers are completely specified by the q_hdr struct passed to this
 * fct. Our own 16 byte header is:
 *
 *              8 bits     8 bits     8 bits   8 bits
 *             __________|_________|________|_________
 *            |<table:=1> <tableID> <      PID      > |
 *            |---------------------------------------|
 *            | <     ID_ext     >  <   table_type   >|
 *            |---------------------------------------|
 *            |<   N/A     ><    table_length       > |
 *            -----------------------------------------
 *            | <               N/A                 > |
 *            -----------------------------------------
 *
 * The "table_type" is from the MGT table_type field & is only
 * meaningful for those types of table; for any other tables
 * the value passed is zero.
 */

pipe_q2api_proc(q_hdr,pid,tbl_type)
struct  queue_hdr *q_hdr;
unsigned int      pid;
unsigned int      tbl_type;
{       /* pipe table buffer via FIFO */

	char          *c_ptr;
	unsigned int  tbl_len, tbl_id_ext;
	unsigned char pipe_bytes[16];
	struct  buf_hdr  *cur_link;

	cur_link = q_hdr->q_head_ptr;
	c_ptr    =   cur_link->buf_ptr;
	tbl_len  =  q_hdr->q_tot_bytes;

	if ( output_fifo_dev_ptr == NULL )
		return;         /* no output fifo pipe specified
				   therefore, don't write anything to it */
 /*
  * pack data into our 16 byte header
  */
	pipe_bytes[0] =  0x1;
	pipe_bytes[1] =  (*c_ptr)&0xFF;
	pipe_bytes[2] =  (pid>>8)&0x1F;
	pipe_bytes[3] =   pid&0xFF;
	pipe_bytes[4] =  c_ptr[3]&0xFF;
	pipe_bytes[5] =  c_ptr[4]&0xFF;
	pipe_bytes[6] =  (tbl_type>>8)&0xFF;
	pipe_bytes[7] =  (tbl_type&0xFF);
	pipe_bytes[8] =  0x0;
	pipe_bytes[9] =  (tbl_len>>16)&0xF;
	pipe_bytes[10] =  (tbl_len>>8)&0xFF;
	pipe_bytes[11] =  (tbl_len)&0xFF;
	pipe_bytes[12] =  0x0;
	pipe_bytes[13] =  0x0;
	pipe_bytes[14] =  0x0;
	pipe_bytes[15] =  0x0;

/*
 * pipe our header
 */
	fwrite(pipe_bytes,1,16, api_pipe_stream);
/*        write_msg('P', 'D', "pipe header = 0x%x %x %x", pipe_data[0], pipe_data[1],
						pipe_data[2] );
 */
/*
 * pipe the table, section by section
 */
/*        write(api_pipe_stream,c_ptr,tbl_len);
*/

	while ( cur_link != NULL )
	{
		c_ptr    =   cur_link->buf_ptr;
		tbl_len = (((c_ptr[1]&0xF)<<8) | (c_ptr[2]&0xFF)) + 3;
		fwrite(c_ptr,1,tbl_len, api_pipe_stream);
		cur_link = cur_link->q_ptr;
	}
/*
 * flush the buffer contents into the pipe, now
 */
	fflush(api_pipe_stream);
	return;
}

/* --------------------------------------------------------------------- */

/*
 * pipe a table buf, preceded by our own 16 byte header, to the
 * API simulator process via a named pipe.
 *
 * The buf is completely specified by the buf_hdr struct passed to this
 * fct. Our own 16 byte header is:
 *
 *              8 bits     8 bits     8 bits   8 bits
 *             __________|_________|________|_________
 *            |<table:=1> <tableID> <      PID      > |
 *            |---------------------------------------|
 *            | <     ID_ext     >  <   table_type   >|
 *            |---------------------------------------|
 *            |<   N/A     ><    table_length       > |
 *            -----------------------------------------
 *            | <               N/A                 > |
 *            -----------------------------------------
 *
 * The "table_type" is from the MGT table_type field & is only
 * meaningful for those types of table; for any other tables
 * the value passed is zero.
 */

pipe_tab2api_proc(buf_hdr_ptr,pid,tbl_type)
struct  buf_hdr  *buf_hdr_ptr;
unsigned int     pid;
unsigned int     tbl_type;
{       /* pipe table buffer via FIFO */

	char          *c_ptr;
	unsigned int  tbl_len, tbl_id_ext;
	unsigned char pipe_bytes[16];

	c_ptr    =   buf_hdr_ptr->buf_ptr;
	tbl_len  =  buf_hdr_ptr->bytes_checked;

	if ( output_fifo_dev_ptr == NULL )
		return;         /* no output fifo pipe specified
				   therefore, don't write anything to it */
 /*
  * pack data into our 16 byte header
  */
	pipe_bytes[0] =  0x1;
	pipe_bytes[1] =  (*c_ptr)&0xFF;
	pipe_bytes[2] =  (pid>>8)&0x1F;
	pipe_bytes[3] =   pid&0xFF;
	pipe_bytes[4] =  c_ptr[3]&0xFF;
	pipe_bytes[5] =  c_ptr[4]&0xFF;
	pipe_bytes[6] =  (tbl_type>>8)&0xFF;
	pipe_bytes[7] =  (tbl_type&0xFF);
	pipe_bytes[8] =  0x0;
	pipe_bytes[9] =  (tbl_len>>16)&0xF;
	pipe_bytes[10] =  (tbl_len>>8)&0xFF;
	pipe_bytes[11] =  (tbl_len)&0xFF;
	pipe_bytes[12] =  0x0;
	pipe_bytes[13] =  0x0;
	pipe_bytes[14] =  0x0;
	pipe_bytes[15] =  0x0;

/*
 * pipe our header
 */
	fwrite(pipe_bytes,1,16, api_pipe_stream);
/*        write_msg('P', 'D', "pipe header = 0x%x %x %x", pipe_data[0], pipe_data[1],
						pipe_data[2] );
/*
 * pipe the table
 */
/*        write(api_pipe_stream,c_ptr,tbl_len);
*/      fwrite(c_ptr,1,tbl_len, api_pipe_stream);

/*
 * flush the buffer contents into the pipe, now
 */
	fflush(api_pipe_stream);
	return;
}

/* --------------------------------------------------------------------- */

/*
 * pipe a Raw Data Piping string to the API simulator process via a named pipe.
 *
 * A pointer to the string (a full 188 byte MPEG-2 TS pkt) is passed to this
 * fct. Our own 16 byte header is:
 *
 *              8 bits     8 bits     8 bits   8 bits
 *             __________|_________|________|_________
 *            |<data:=2> <dataID:=3> <     PID      > |
 *            |---------------------------------------|
 *            | <            N/A                     >|
 *            |---------------------------------------|
 *            | <            N/A                     >|
 *            |---------------------------------------|
 *            | <        length = 188               > |
 *            -----------------------------------------
 *
 * The "data = 2" indicates that info to follow is data (vs a table).
 * The "dataID = 1" indicates that the data is from a Data Carousel Module.
 *                      or an Async Non-Flow Controlled Data Module
 * The "dataID = 2" indicates that the data is from an Async Streaming
 *                      Data Module (i.e., module size = 0)
 * The "dataID = 3" indicates that the data is from Raw Data Piping
 *                      (a full 188 byte MPEG-2 TS pkt)
 * The "dataID = 4" indicates that the data is from Cooked Data Piping
 *                      (only the payload portion of an MPEG-2 TS pkt)
 */

pipe_raw2api_proc(pkt_ptr,pid,len)
unsigned char   *pkt_ptr;
unsigned int    pid;
unsigned int    len;
{
  /* pipe Raw Data Piping (a full 188 byte MPEG-2 TS pkt) */
  /* or Cooked Data Piping (only the payload portion of the MPEG-2 TS pkt) */

	unsigned char pipe_bytes[16];

/*
 * pack data into our 16 byte header
 */
	pipe_bytes[0] =  0x2;
	if ( len == 188 )
	{
	    pipe_bytes[1] =  0x3;   /* Raw Data Piping (a full 188 byte MPEG-2 TS pkt) */
	}
	else if ( len < 188 )
	{
	    pipe_bytes[1] =  0x4;   /* Cooked Data Piping (less than 188 bytes
				     - only the payload portion of the MPEG-2 TS pkt) */
	}
	else
	{
	   write_msg('P', 'E', "ERROR Raw Piping more than 188 Bytes(%d)", len);
	   return;
	}
	pipe_bytes[2] =  (pid>>8)&0x1F;
	pipe_bytes[3] =   pid&0xFF;
	pipe_bytes[4] =  0x0;
	pipe_bytes[5] =  0x0;
	pipe_bytes[6] =  0x0;
	pipe_bytes[7] =  0x0;
	pipe_bytes[8] =  0x0;
	pipe_bytes[9] =  0x0;
	pipe_bytes[10] = 0x0;
	pipe_bytes[11] = 0x0;
	pipe_bytes[12] = 0x0;
	pipe_bytes[13] = 0x0;
	pipe_bytes[14] = 0x0;
	pipe_bytes[15] = len & 0xFF;

	if ( output_fifo_dev_ptr == NULL )
		return;         /* no output fifo pipe specified */
				/* therefore, don't write anything to it */
/*
 * pipe our header
 */
	fwrite(pipe_bytes,1,16, api_pipe_stream);

/*
 * pipe the module data
 */
	fwrite(pkt_ptr,1,len,api_pipe_stream);
/*
 * flush the buffer contents into the pipe, now
 */
	fflush(api_pipe_stream);
	return;
}

/* --------------------------------------------------------------------- */

/*
 * pipe a carousel module to the API simulator process via a named pipe.
 *
 * The buf is completely specified by the buf_hdr struct passed to this
 * fct. Our own 16 byte header is:
 *
 *              8 bits     8 bits     8 bits   8 bits
 *             __________|_________|________|_________
 *            |<data:=2> <dataID:=1> <     PID      > |
 *            |---------------------------------------|
 *            | <     Download_ID (Carousel_ID)      >|
 *            |---------------------------------------|
 *            | < N/A  >  <mod_ver> <   module_ID    >|
 *            |---------------------------------------|
 *            | <           module_length           > |
 *            -----------------------------------------
 *
 * The "data = 2" indicates that info to follow is data (vs a table).
 * The "dataID = 1" indicates that the data is from a Data Carousel Module.
 *                      or an Async Non-Flow Controlled Data Module
 * The "dataID = 2" indicates that the data is from an Async Streaming
 *                      Data Module (i.e., module size = 0)
 * The "dataID = 3" indicates that the data is from Raw Data Piping
 *                      (a full 188 byte MPEG-2 TS pkt)
 * The "dataID = 4" indicates that the data is from Cooked Data Piping
 *                      (only the payload portion of an MPEG-2 TS pkt)
 *
 * If "dataID = 1" (Data Carousel Module) Then the next 8 bytes after
 * the above header is PTS (Presentation Time Stamp) info as follows:
 *
 *
 *             __________|_________|________|__________
 *            |<      N/A         > <PTS_flg><PTS[33]> |
 *            |----------------------------------------|
 *            | <         PTS[32..1]                >  |
 *            ------------------------------------------
 *
 * PTS_flag is a 16 bit quantity whose value 1=yes, 0=no
 * PTS[33] is the 33rd bit of the PTS and is located in the LSB of this
 * 16 bit field.
 *
 */

pipe_car2api_proc(carousel,pid,carousel_mod)
struct carousel *carousel;
unsigned int    pid;
struct data_mod *carousel_mod;
{
  /* pipe Data Carousel Module */


	unsigned int  mod_len, i, num, blk_size;
	unsigned char pipe_bytes[16];

/*
 * pack data into our 16 byte header
 */
	pipe_bytes[0] =  0x2;
	if ( carousel_mod->mod_size == 0 )
	{
	    mod_len  =  carousel_mod->buf_size;
	    pipe_bytes[1] =  0x2;   /* Async Streaming Data, blk-by-blk */
	}
	else
	{
	    mod_len  =  carousel_mod->mod_size;
	    pipe_bytes[1] =  0x1;   /* Async Non-Flow Ctrl & Carousel data module */
	}
	pipe_bytes[2] =  (pid>>8)&0x1F;
	pipe_bytes[3] =   pid&0xFF;
	pipe_bytes[4] =  ((carousel->carousel_id>>24)&0xFF);
	pipe_bytes[5] =  ((carousel->carousel_id>>16)&0xFF);
	pipe_bytes[6] =  ((carousel->carousel_id>> 8)&0xFF);
	pipe_bytes[7] =  ((carousel->carousel_id    )&0xFF);
	pipe_bytes[8] =  0x0;
	pipe_bytes[9] =  ((carousel_mod->mod_ver)&0xFF);
	pipe_bytes[10] = ((carousel_mod->module_id>>8)&0xFF);
	pipe_bytes[11] = ((carousel_mod->module_id)&0xFF);
	pipe_bytes[12] = ((mod_len>>24)&0xFF);
	pipe_bytes[13] = ((mod_len>>16)&0xFF);
	pipe_bytes[14] = ((mod_len>> 8)&0xFF);
	pipe_bytes[15] = ((mod_len    )&0xFF);

	if ( output_fifo_dev_ptr == NULL )
		return;         /* no output fifo pipe specified */
				/* therefore, don't write anything to it */
/*
 * pipe our header
 */
	fwrite(pipe_bytes,1,16, api_pipe_stream);

/*
 * pack data into our 8 byte PTS info
 */
	pipe_bytes[0] =  0x0;
	pipe_bytes[1] =  0x0;
	pipe_bytes[2] =  carousel_mod->blk_hdr_ptr[0].PTS_flag;
	pipe_bytes[3] =  carousel_mod->blk_hdr_ptr[0].PTS_hi;
	pipe_bytes[4] =  ((carousel_mod->blk_hdr_ptr[0].PTS_low>>24)&0xFF);
	pipe_bytes[5] =  ((carousel_mod->blk_hdr_ptr[0].PTS_low>>16)&0xFF);
	pipe_bytes[6] =  ((carousel_mod->blk_hdr_ptr[0].PTS_low>> 8)&0xFF);
	pipe_bytes[7] =  ((carousel_mod->blk_hdr_ptr[0].PTS_low    )&0xFF);
/*
 * pipe the PTS info for the 1st data blk
 */
	fwrite(pipe_bytes,1,8, api_pipe_stream);

/*
 * pipe the module data
 */
	fwrite(carousel_mod->buffer,1,mod_len,api_pipe_stream);
/*
 * flush the buffer contents into the pipe, now
 */
	fflush(api_pipe_stream);
	return;
}

/* ---------------------------------------------------------------------- */

/*
 * get current system time, set next MPEG pkt time and use that as
 * start time (for effective bandwidth calc)
 */

init_time_throttle()
{
   gettimeofday(&next_time, NULL );

    if ( (next_time.tv_usec+=70) >= 1000000 )
    {
	next_time.tv_sec++;
	next_time.tv_usec -= 1000000;
    }
    start_time.tv_sec =  next_time.tv_sec;
    start_time.tv_usec =  next_time.tv_usec;
    return;
}

/* ---------------------------------------------------------------------- */

/*
 * wait for current system time to catch up with time of next
 * expected MPEG pkt rcv time. This will simulate a correct
 * bandwidth for the MPEG input stream to ~20Mb/s
 */

#define TIME_PER_PKT    70 /* time in us between MPEG pkts */
			   /* 188 Bytes @ 20 Mbits/s ==> 75.2 us per pkt */
			   /* use 70 us & allow ~6 us overhead */
			   /* for throttle code */
throttle_stream()
{
   struct   timeval   current_time;

   while ( gettimeofday(&current_time, NULL) != -1 )
   {
	if ( next_time.tv_sec > current_time.tv_sec )
	   continue;
	if ( next_time.tv_sec == current_time.tv_sec )
	{
	    if ( next_time.tv_usec > current_time.tv_usec )
	       continue;
	}
	if ( (next_time.tv_usec+=TIME_PER_PKT) >= 1000000 )
	{
	    next_time.tv_sec++;
	    next_time.tv_usec -= 1000000;
	}
	return;
   }
   return;
}

/* ---------------------------------------------------------------------- */

/*
 * This fct gets the next FULL input Transport Pkt and returns a ptr to it.
 *
 * It advances the ptr to the next sequential pkt in the input buf,
 *   if there are no more pkts in the buf, read more in;
 *     if the "dev" returns no pkts then terminate
 */

/*
char *get_next_transport_pkt()
{
     int        i;

     input_buf_ptr = input_buf_next_pkt;        // set current pkt ptr
						    					// to the next pkt
     input_buf_next_pkt += TRANSPORT_PKT_SIZE;  // advance next pkt ptr

 // throttle input bit stream to ~20 Mb/s

     if ( rate_limit_input_stream == 1 )
     {
	    throttle_stream();
     }
 // check if current ptr is beyond end of buf

     if ( input_buf_ptr > input_buf_end )
     {
      // buf exhausted get more data from input stream
      // termination on error is applicable only for simulated/emulated ver

	  if ( (input_buf_cnt=read(input_buf_stream,input_buf_start,INPUT_BUF_SIZE)) < TRANSPORT_PKT_SIZE )
	  {
	       write_msg('P', 'I', "Transport Stream empty - terminating, no more pkts in get_next_transport_pkt()");
	       prt_summary();
	       write_msg('P', 'I', "Program Terminating");
	       exit(-1);
	  }
	  if ( (input_buf_cnt % TRANSPORT_PKT_SIZE) != 0 )
	  {
	       write_msg('P', 'E', "error retrieving full transport pkts in get_next_transport_pkt()");
	       prt_summary();
	       write_msg('P', 'I', "Program Terminating");
	       exit(-1);
	  }
      // init buf ptrs to the new data in the buf

	  input_buf_ptr = input_buf_start;
	  input_buf_next_pkt = input_buf_start + TRANSPORT_PKT_SIZE;
	  input_buf_end = input_buf_start + input_buf_cnt-1;
     }

#ifdef DEBUG
     start_msg("new pkt - ");
	append_msg(" %2x", input_buf_ptr[0]&0xFF );
	append_msg(" %x", ((input_buf_ptr[1]>>6)&0x01) );
	append_msg(" %4x", ((input_buf_ptr[1]&0x1F)<<8) | (input_buf_ptr[2]&0xFF));
	i = ((input_buf_ptr[3]>>4)&0x03);
	append_msg(" adopt=%x%s", i, ((i&0x2) ? "(Y)" : "(N)") );
	append_msg(" cntr=%x", ((input_buf_ptr[3]   )&0x0F) );
     append_msg(" -");
     for ( i=4; i<10; i++)
	append_msg(" %2x", input_buf_ptr[i]&0xFF );
     end_msg('P', 'D', "");
#endif


#ifdef DEBUG_SUMMARY
     input_epoch += TRANSPORT_PKT_SIZE;
     if ( input_epoch > DATA_STREAM_EPOCH )
     {
	input_megs++;
	input_epoch -= DATA_STREAM_EPOCH;
	prt_summary();
     }
#endif
     return( input_buf_ptr );
} */

/* --------------------------------------------------------------------- */

prt_summary()
{
    int     i, Mbytes_proc, elapsed_time, bit_rate;
    struct  ts_control *ts_control_ptr;
    struct  timeval   current_time;

	gettimeofday(&current_time, NULL );

	Mbytes_proc = input_megs*(DATA_STREAM_EPOCH/1000000) + (input_epoch/1000000);
	elapsed_time = (current_time.tv_sec - start_time.tv_sec)*1000000
			     + (current_time.tv_usec - start_time.tv_usec);
	elapsed_time /= 1000; /* change us to ms */
	bit_rate = (Mbytes_proc*8000) / elapsed_time;
/*
	write_msg('P', 'S', "*** %d elasped time us (%d,%d) (%d,%d) ***",
			elapsed_time, current_time.tv_sec, current_time.tv_usec,
			start_time.tv_sec, start_time.tv_usec);
*/
	write_msg('P', 'S', "    %d MBytes processed @ %d Mb/s", Mbytes_proc, bit_rate );
	write_msg('P', 'S', "    %d pes_pkts processed", num_pes_pkt );
	write_msg('P', 'S', "    %d atsc_pkts, w/ distribution:", ts_control[0x1FFB].ts_pkt_cnt );
	write_msg('P', 'S', "        sst=%d/%d, mgt=%d/%d, vct=%d/%d, rrt=%d/%d, dcct=%d/%d, dccsct=%d/%d",
			num_stt, num_stt_comp, num_mgt, num_mgt_comp,
			num_vct, num_vct_comp, num_rrt, num_rrt_comp,
		    num_dcct, num_dcct_comp, num_dccsct, num_dccsct_comp );
	write_msg('P', 'S', "    Dropped pkts (Continuity Counter Errors) = %d",
		    cont_cntr_errors);
	ts_control_ptr = &ts_control[0];
	write_msg('P', 'S', "    Active PIDs in this stream");
	for (i=0; i<0x2000; i++, ts_control_ptr++)
	{
	    if ( ts_control_ptr->ts_pkt_cnt > 0 )
	    {
		 write_msg('P', 'S', "        PID=0x%x, ctrl_type=%s, pkt_cnt=%d",  i,
			   ctrl_type_names[ts_control_ptr->ts_ctrl_type],
			   ts_control_ptr->ts_pkt_cnt );
	    }
	}
	return;
}

/* --------------------------------------------------------------------- */

/*
 * This fct initializes the "ts_control" struct. It is used to quickly
 * decode a Transport Pkt via its PID. The PID indexes into the table
 * and the contents of that entry directs the processing of that pkt.
 */

ctrl_init()
{
    int i;
    struct ts_control *ts_control_ptr;

    ts_control_ptr = &ts_control[0];
    for (i=0; i<0x2000; i++, ts_control_ptr++)
    {
	ts_control_ptr->ts_ctrl_type   = UNDEFINED;
	ts_control_ptr->ts_ctrl_status = INACTIVE;
	ts_control_ptr->ts_orig_type   = UNDEFINED;
	ts_control_ptr->ts_orig_stream = 0xFF;
	ts_control_ptr->ts_cont_cntr   = 0xFF;
	ts_control_ptr->ts_buf_hdr_ptr = NULL;
	ts_control_ptr->ts_queue_hdr.q_head_ptr = NULL;
	ts_control_ptr->ts_queue_hdr.q_tail_ptr = NULL;
	ts_control_ptr->ts_queue_hdr.q_valid   = 0;
	ts_control_ptr->ts_queue_hdr.q_tot_bytes = 0;
	ts_control_ptr->ts_queue_hdr.q_ver = 0;
	ts_control_ptr->ts_ctrl_gen.ts_ctrl_ptr = NULL;
	ts_control_ptr->ts_pkt_cnt     = 0;
    }
    ts_control[0x0000].ts_ctrl_type   = PAT ;
    ts_control[0x0000].ts_ctrl_status = ACTIVE;
    ts_control[0x0000].ts_buf_hdr_ptr = NULL;

    ts_control[0x0001].ts_ctrl_type   = CAT ;
    ts_control[0x0001].ts_ctrl_status = ACTIVE;
    ts_control[0x0001].ts_buf_hdr_ptr = NULL;

    ts_control[0x1FFB].ts_ctrl_type   = ATSC;
    ts_control[0x1FFB].ts_ctrl_status = ACTIVE;

    ts_control[0x1FFF].ts_ctrl_type   = EMPTY;

    return;
}

/* --------------------------------------------------------------------- */

/*
 * This fct checks each entry (except 1st two, PAT & CAT, 1FFB & last
 * 1FFF) in the "ts_control" struct. If any active PIDs are found it
 * deletes the related items & associated buffers, etc.
 */

ctrl_clr()
{
    int i;
    struct ts_control *ts_control_ptr;

    ts_control_ptr = &ts_control[2];
    for (i=2; i<0x1FFF; i++, ts_control_ptr++)
    {
	if ( ts_control_ptr->ts_ctrl_type != UNDEFINED )
	{
	    if ( i != 0x1FFB )
	    {
	       switch ( ts_control_ptr->ts_ctrl_type )
	       {
		 case AUDIO_AC3:
		 case VIDEO:
		     if ( ts_control_ptr->ts_ctrl_gen.ts_ctrl_fd != -1 )
		     {
			 close_pid_output_dev(i);
		     }
		     break;

		 case PAT:
		 case CAT:
		 case ATSC:
		 case EIT:
		 case ETT:
		 case DET:
		 case LTST:
		 case PMT:
		     if ( ts_control_ptr->ts_buf_hdr_ptr != NULL )
		     {
			 if (ts_control_ptr->ts_buf_hdr_ptr->buf_ptr != NULL)
			     free (ts_control_ptr->ts_buf_hdr_ptr->buf_ptr);
			 free (ts_control_ptr->ts_buf_hdr_ptr);
			 ts_control_ptr->ts_buf_hdr_ptr = NULL;
		     }
		     clear_queue( ts_control_ptr->ts_queue_hdr );
		     break;

		 case DST_NRT:
		     if ( ts_control_ptr->ts_buf_hdr_ptr != NULL )
		     {
			 if (ts_control_ptr->ts_buf_hdr_ptr->buf_ptr != NULL)
			     free (ts_control_ptr->ts_buf_hdr_ptr->buf_ptr);
			 free (ts_control_ptr->ts_buf_hdr_ptr);
			 ts_control_ptr->ts_buf_hdr_ptr = NULL;
		     }
		     clear_queue( ts_control_ptr->ts_queue_hdr );
		     if ( ts_control_ptr->ts_ctrl_gen.ts_ctrl_ptr != NULL )
		     {
			 delete_dst_nrt_q_item( (struct dst_nrt_q_lst *)
				   ts_control_ptr->ts_ctrl_gen.ts_ctrl_ptr );
		     }
		     break;

		 case DII_DDB:
		     if ( ts_control_ptr->ts_buf_hdr_ptr != NULL )
		     {
			 if (ts_control_ptr->ts_buf_hdr_ptr->buf_ptr != NULL)
			     free (ts_control_ptr->ts_buf_hdr_ptr->buf_ptr);
			 free (ts_control_ptr->ts_buf_hdr_ptr);
			 ts_control_ptr->ts_buf_hdr_ptr = NULL;
		     }
		     clear_queue( ts_control_ptr->ts_queue_hdr );
		     if ( ts_control_ptr->ts_ctrl_gen.ts_ctrl_ptr != NULL )
		     {
			 delete_dii_ddb_lst_item( (struct  dii_ddb_lst *)
				   ts_control_ptr->ts_ctrl_gen.ts_ctrl_ptr );
		     }
		     break;

		 case ADDR_SECT:
		     if ( ts_control_ptr->ts_buf_hdr_ptr != NULL )
		     {
			 if (ts_control_ptr->ts_buf_hdr_ptr->buf_ptr != NULL)
			     free (ts_control_ptr->ts_buf_hdr_ptr->buf_ptr);
			 free (ts_control_ptr->ts_buf_hdr_ptr);
			 ts_control_ptr->ts_buf_hdr_ptr = NULL;
		     }
		     clear_queue( ts_control_ptr->ts_queue_hdr );
		     if ( ts_control_ptr->ts_ctrl_gen.ts_ctrl_ptr != NULL )
		     {
			 delete_addr_sect_lst_item( (struct  addr_sect_lst *)
				    ts_control_ptr->ts_ctrl_gen.ts_ctrl_ptr );
		     }
		     break;

		 case PES_IZED:
		     break;
		 case PES_OUS:
		     break;
		 case DSM_CC_ASYNC:
		     break;

		 case PIPING_RAW:
		 case PIPING_COOKED:
		     if ( ts_control_ptr->ts_buf_hdr_ptr != NULL )
		     {
			 if (ts_control_ptr->ts_buf_hdr_ptr->buf_ptr != NULL)
			     free (ts_control_ptr->ts_buf_hdr_ptr->buf_ptr);
			 free (ts_control_ptr->ts_buf_hdr_ptr);
			 ts_control_ptr->ts_buf_hdr_ptr = NULL;
		     }
		     clear_queue( ts_control_ptr->ts_queue_hdr );
		     break;

		 case EMPTY:
		     break;
		 default:
		     break;
	       }
	       write_msg('P', 'E', "ERROR this %s type should have been deleted w/ annonce Tbls",
		       ctrl_type_names[ts_control_ptr->ts_ctrl_type] );
	       ts_control_ptr->ts_ctrl_type   = UNDEFINED;
	       ts_control_ptr->ts_ctrl_gen.ts_ctrl_ptr = NULL;
	       ts_control_ptr->ts_orig_type   = UNDEFINED;
	       ts_control_ptr->ts_orig_stream = 0xFF;
	       ts_control_ptr->ts_cont_cntr   = 0xFF;
	    }
	}

	ts_control_ptr->ts_ctrl_type   = UNDEFINED;
	ts_control_ptr->ts_ctrl_status = INACTIVE;
	ts_control_ptr->ts_buf_hdr_ptr = NULL;
	ts_control_ptr->ts_ctrl_gen.ts_ctrl_ptr = NULL;
	ts_control_ptr->ts_pkt_cnt     = 0;
    }

    return;
}



/* --------------------------------------------------------------------- */
/* --------------------------------------------------------------------- */

/*
 * Functions above this deal with external devices or processes,
 * such as the API_SIMULATION (the input & output FIFOs), the tuner,
 * the input bitstream, output dev streams, etc.
 */

/* --------------------------------------------------------------------- */
/* --------------------------------------------------------------------- */

/*
 * This is the main processing fct
 *
 * Init the input stream & forever process Transport Pkts based on PIDs
 */

void process_packet(char *packet_ptr, char do_init)
{
    char          *input_ptr, *strg, *cur_pkt_ptr;
    char          cmd_strg1[256], cmd_strg2[256];
    unsigned int  ts_adapt_flags, ts_pcr_hi, ts_pcr_lo;
    int           i, input_remaining_bytes, input_cnt, loop_cnt, pes_hdr_len;
    int           k, max_k, l, partial_input_buf_cnt;
    unsigned int  (*fct_ptr)();
    unsigned int  ts_payload_start, ts_pid, ts_adapt_fld, ts_cont_cntr;
    unsigned int  ts_adapt_len;
    unsigned int  pes_start_code, pes_stream_id ;
    struct        ts_control    *ts_control_ptr;
    struct        mgt_tbl_entry *mgt_tbl_entry_ptr;
    struct        pat_item      *pat_tbl_entry_ptr;
    struct        buf_hdr       *buf_hdr_ptr;
    struct        dii_PID_tbl   *dii_PID_tbl_ptr;
    struct        queue_hdr     *q_hdr;


rate_limit_input_stream = 0;

if (do_init) {
/*
 * Init the input buf params & the input device
 */
    input_buf_init();
/*
 * Init the Transport Stream decoder array
 */
    ctrl_init();

    init_time_throttle();
    current_chan = 13;
}

     input_epoch += TRANSPORT_PKT_SIZE;
     if ( input_epoch > DATA_STREAM_EPOCH )
     {
	input_megs++;
	input_epoch -= DATA_STREAM_EPOCH;
     }

	cur_pkt_ptr = input_ptr = packet_ptr;

	input_cnt  = TRANSPORT_PKT_FIXED_HEADER_SIZE;
	input_remaining_bytes = TRANSPORT_PKT_SIZE - input_cnt;
/*
 * extract info from transport stream fixed header
 */

	ts_payload_start = (input_ptr[1]>>6)&0x1;
	ts_pid           = ((input_ptr[1]&0x1F)<<8) | (input_ptr[2]&0xFF);
	ts_adapt_fld     = (input_ptr[3]>>4)&0x3;
	ts_cont_cntr     = (input_ptr[3] & 0xF);
/*
 * we don't need to parse the adaptation_field, just skip over it
 */
	if ( (ts_adapt_fld&0x2) == 0x2 )
	{
 /* process the minimum necessary of this adaptation_field to skip over it
  */
	    if ( (ts_adapt_len = input_ptr[4]&0xFF) > 0 )
	    {
		ts_adapt_flags = input_ptr[5]&0xFF;
		if ( ts_adapt_flags & 0x90 )
		{
		   if ( ts_adapt_flags & 0x10 )
		   {
		       ts_pcr_hi = ((input_ptr[6]&0xFF)<< 2) |
				   ((input_ptr[7]>>6 )&0x3 );
		       ts_pcr_lo = ((input_ptr[7]&0x3F)<<26)  |
				   ((input_ptr[8]&0xFF)<<18)  |
				   ((input_ptr[9]&0xFF)<<10)  |
				   ((input_ptr[10]&0x80)<< 2) |
				   ((input_ptr[10]&0x1 )<< 8) |
				    (input_ptr[11]&0xFF);
		   }
		   pipe_pcr2api_proc(ts_adapt_flags,ts_pcr_hi,ts_pcr_lo,ts_pid);

/*                 write_msg('P', 'D', "PCR pid=0x%x flags=0x%x, pcr= 0x%x %x",
			ts_pid, ts_adapt_flags, ts_pcr_hi, ts_pcr_lo);
*/
		}
		input_cnt += (ts_adapt_len + 1);
		input_remaining_bytes -= (ts_adapt_len + 1);
		if ( input_remaining_bytes <= 0 )
		{
#ifdef DEBUG_ERR_MSG
			write_msg('P', 'D', "No payload in Transport Pkt, only Adaptaion fld");
			write_msg('P', 'D', "*** %d MBytes processed, input_byte_cnt(epoch)=%d (pkts=%d) ***",
			 input_megs*(DATA_STREAM_EPOCH/1000000) + (input_epoch/1000000),
			 input_epoch, input_epoch/TRANSPORT_PKT_SIZE );
#endif
			return;  /* if no data left in pkt, get next pkt */
		}
	    }
	    else
	    {       /* 0 adapt len => 1 byte of padding */
		input_cnt += 2;
		input_remaining_bytes -= 2;
	    }
	}
/*  #ifdef DEBUG
/*          start_msg("new pkt - ");
/*             append_msg(" %2x", input_buf_ptr[0]&0xFF );
/*             append_msg(" %x", ((input_buf_ptr[1]>>6)&0x01) );
/*             append_msg(" %4x", ((input_buf_ptr[1]&0x1F)<<8) | (input_buf_ptr[2]&0xFF));
/*             i = ((input_buf_ptr[3]>>4)&0x03);
/*             append_msg(" adopt=%x%s", i, ((i&0x2) ? "(Y)" : "(N)") );
/*             append_msg(" cntr=%x", ((input_buf_ptr[3]   )&0x0F) );
/*          append_msg(" -");
/*          for ( i=4; i<10; i++)
/*             append_msg(" %2x", input_buf_ptr[i]&0xFF );
/*          end_msg('P', 'D', "");
/*  #endif
*/

 /*
  * advance input ptr to pkt pay load (which may be a "ptr" if
  *  "ts_payload_start" is set and pkt is a "table"
  */
	input_ptr += input_cnt;

 /* process pay load, if it exists */

	if ( (ts_adapt_fld&0x1) == 0x1 )
	{
	    ts_control_ptr = &ts_control[ts_pid];
	    if ( ts_control_ptr->ts_pkt_cnt < 0x7FFFFFFF )
	    {
		ts_control_ptr->ts_pkt_cnt++;   /* incr pkt_cnt w/o overflow */
	    }

		  /* chk TS pkt Continuity Count */

/*  write_msg('P', 'D', "PID 0x%x pkt Continuity Count - expected %d got %d",
	ts_pid, ts_control_ptr->ts_cont_cntr,  ts_cont_cntr);
*/
	    if ( ts_control_ptr->ts_cont_cntr != ts_cont_cntr )
	    {
		if ( ts_pid == 0x1FFF )
		     ;
		else if ( ts_control_ptr->ts_cont_cntr == 0xFF )
		{     /* init new stream count */
		    ts_control_ptr->ts_cont_cntr = (ts_cont_cntr+1)%16;
		}
		else
		{
		    write_msg('P', 'E', "ERROR in TS 0x%x pkt Continuity Count - expected %d got %d",
			ts_pid, ts_control_ptr->ts_cont_cntr,  ts_cont_cntr);
		    ts_control_ptr->ts_cont_cntr = (ts_cont_cntr+1)%16;
		    cont_cntr_errors++;
		}
	    }
	    else
	    {    /* modulo incr to next expected value */
		ts_control_ptr->ts_cont_cntr = (ts_control_ptr->ts_cont_cntr+1)%16;
	    }

		/* process by type */

	    switch ( ts_control_ptr->ts_ctrl_type )
	    {
	      case UNDEFINED:
		break;

	      case AUDIO_AC3:
		if ( ts_payload_start == 1 )
		{
		   pes_start_code = ((input_ptr[0]&0xFF)<<16)
			 | ((input_ptr[1]&0xFF)<<8) | (input_ptr[+2]&0xFF);
		   pes_stream_id = input_ptr[3] & 0xFF;
		   if ( pes_start_code != 0x000001 )
			 write_msg('P', 'E', "**** Audio PES start error");
		   else
		       num_pes_pkt++;
		   if ( ts_control_ptr->ts_ctrl_status == READY
		       && output_audio_stream != -1 )
		   {
		       ts_control_ptr->ts_ctrl_status = ACTIVE;
		       write_msg('P', 'I', "Initiating Audio output to %s dev",
				 output_audio_dev_ptr );
		   }
		}
		if ( ts_control_ptr->ts_ctrl_status == ACTIVE )
		{
		      if ( pes_header_audio_only == 1 )
		      {
			  if ( ts_payload_start == 1 )
			  {
			     pes_hdr_len = (input_ptr[8]&0xFF) + 8;
			     write(output_audio_stream,input_ptr,pes_hdr_len);
			  }
		      }
		      else
		      {
			 write(output_audio_stream,input_ptr,input_remaining_bytes);
		      }
		}
		break;

	      case VIDEO:
		if ( ts_payload_start == 1 )
		{
		   pes_start_code = ((input_ptr[0]&0xFF)<<16)
			 | ((input_ptr[1]&0xFF)<<8) | (input_ptr[+2]&0xFF);
		   pes_stream_id = input_ptr[3] & 0xFF;
		   num_pes_pkt++;
		   if ( ts_control_ptr->ts_ctrl_status == READY
		       && output_video_stream != -1 )
		   {
		       ts_control_ptr->ts_ctrl_status = ACTIVE;
		       write_msg('P', 'I', "Initiating Video output to %s dev",
				 output_video_dev_ptr );
		   }
		}
		if ( ts_control_ptr->ts_ctrl_status == ACTIVE )
		{
		   if ( pes_header_video_only == 1 )
		   {
		       if ( ts_payload_start == 1 )
		       {
			  pes_hdr_len = (input_ptr[8]&0xFF) + 8;
			  write(output_video_stream,input_ptr,pes_hdr_len);
		       }
		   }
		   else
		   {
		       write(output_video_stream,input_ptr,input_remaining_bytes);
		   }
		}
		break;
  /*
   *  a "table" or data carousel
   */
	      case PAT:
		accum_sect(ts_payload_start,input_ptr,input_remaining_bytes,
						 ts_control_ptr,ts_pid);
		   /*
		    * Any complete Table Sections are now queued on the
		    *   "ts_control_ptr->ts_queue_hdr" queue
		    * the input pkt processing is complete & can rtn
		    * to the drvr or input process.
		    * This would be the place to handoff the queue for
		    * processing in a real-time system
		    */
		while( (buf_hdr_ptr =
		       dequeue_buf(&(ts_control_ptr->ts_queue_hdr))) != NULL )
		{
		    accum_tables(buf_hdr_ptr,ts_control_ptr,&pat_q_hdr,ts_pid);
		}
		break;

	      case CAT:
		accum_sect(ts_payload_start,input_ptr,input_remaining_bytes,
						     ts_control_ptr,ts_pid);
		   /*
		    * Any complete Table Sections are now queued on the
		    *   "ts_control_ptr->ts_queue_hdr" queue
		    * the input pkt processing is complete & can rtn
		    * to the drvr or input process.
		    * This would be the place to handoff the queue for
		    * processing in a real-time system
		    */
		while( (buf_hdr_ptr =
		       dequeue_buf(&(ts_control_ptr->ts_queue_hdr))) != NULL )
		{
		    accum_tables(buf_hdr_ptr,ts_control_ptr,&cat_q_hdr,ts_pid);
		}
		break;

	      case ATSC:   /* ATSC tables (1FFB) MGT, TVCT, CVCT, STT, RRT,
			      DCCT & DCCSCT tables */
		accum_sect(ts_payload_start,input_ptr,input_remaining_bytes,
						      ts_control_ptr,ts_pid);
		   /*
		    * Any complete Table Sections are now queued on the
		    *   "ts_control_ptr->ts_queue_hdr" queue
		    * the input pkt processing is complete & can rtn
		    * to the drvr or input process.
		    * This would be the place to handoff the queue for
		    * processing in a real-time system
		    */
		while( (buf_hdr_ptr =
		       dequeue_buf(&(ts_control_ptr->ts_queue_hdr))) != NULL )
		{
		    atsc_tables(buf_hdr_ptr,ts_control_ptr);
		}
		break;

	      case EIT:
	      case ETT:
	      case DET:
	      case LTST:
		mgt_tbl_entry_ptr =
			(struct mgt_tbl_entry *) ts_control_ptr->ts_ctrl_gen.ts_ctrl_ptr;
		if ( NULL == (q_hdr=mgt_tbl_entry_ptr->table_type_q_ptr) )
		  /* ERROR - no preallocated queue */
		    break;
		if ( q_hdr->q_valid == 1 )
		{       /* This PID is a single table                    */
			/* Since we are counting on the MGT to change &  */
			/* thus deallocate all its tables if any of them */
			/* change. So as long as the table is valid      */
			/* these are all duplicates                      */
		     break;
		}
		accum_sect(ts_payload_start,input_ptr,input_remaining_bytes,
						  ts_control_ptr,ts_pid);
		  /*
		   * Any complete Table Sections are now queued on the
		   *   "ts_control_ptr->ts_queue_hdr" queue
		   * the input pkt processing is complete & can rtn
		   * to the drvr or input process.
		   * This would be the place to handoff the queue for
		   * processing in a real-time system
		   */
		while( (buf_hdr_ptr =
		       dequeue_buf(&(ts_control_ptr->ts_queue_hdr))) != NULL )
		{
		   accum_tables(buf_hdr_ptr,ts_control_ptr,q_hdr,ts_pid);
		}
		if (ts_control_ptr->ts_ctrl_type == EIT) {
			eit_queue = q_hdr;
			}
		break;

	      case PMT:
		pat_tbl_entry_ptr =
			(struct pat_item *) ts_control_ptr->ts_ctrl_gen.ts_ctrl_ptr;
		if ( NULL == (q_hdr = pat_tbl_entry_ptr->pmt_q_hdr_ptr) )
		{ /* ERROR - no preallocated queue hdr */
		    break;
		}
		   /* This PID is a single table  */

		accum_sect(ts_payload_start,input_ptr,input_remaining_bytes,
						      ts_control_ptr,ts_pid);
		   /*
		    * Any complete Table Sections are now queued on the
		    *   "ts_control_ptr->ts_queue_hdr" queue
		    * the input pkt processing is complete & can rtn
		    * to the drvr or input process.
		    * This would be the place to handoff the queue for
		    * processing in a real-time system
		    */
		while( (buf_hdr_ptr =
		       dequeue_buf(&(ts_control_ptr->ts_queue_hdr))) != NULL )
		{
		    accum_tables(buf_hdr_ptr,ts_control_ptr,q_hdr,ts_pid);
		}
		break;

	      case DST_NRT:
		accum_sect(ts_payload_start,input_ptr,input_remaining_bytes,
						      ts_control_ptr,ts_pid);
		   /*
		    * Any complete Table Sections are now queued on the
		    *   "ts_control_ptr->ts_queue_hdr" queue
		    * the input pkt processing is complete & can rtn
		    * to the drvr or input process.
		    * This would be the place to handoff the queue for
		    * processing in a real-time system
		    */
		while( (buf_hdr_ptr =
		       dequeue_buf(&(ts_control_ptr->ts_queue_hdr))) != NULL )
		{
		      accum_tables(buf_hdr_ptr,ts_control_ptr,q_hdr,ts_pid);
		}
		break;

	      case DII_DDB:
		accum_sect(ts_payload_start,input_ptr,input_remaining_bytes,
						      ts_control_ptr,ts_pid);
		   /*
		    * Any complete Table Sections are now queued on the
		    *   "ts_control_ptr->ts_queue_hdr" queue
		    * the input pkt processing is complete & can rtn
		    * to the drvr or input process.
		    * This would be the place to handoff the queue for
		    * processing in a real-time system
		    */
		while( (buf_hdr_ptr =
		       dequeue_buf(&(ts_control_ptr->ts_queue_hdr))) != NULL )
		{
		    accum_dii_tables(buf_hdr_ptr,ts_control_ptr,ts_pid);
		}
		break;

	      case PIPING_RAW:
			/* send pkt to API Simul without any checks */
		  pipe_raw2api_proc(cur_pkt_ptr,ts_pid,188);
		  break;

	      case PIPING_COOKED:
		  if ( ts_payload_start == 1 )
		  {
		      pipe_raw2api_proc(input_ptr+1,ts_pid,input_remaining_bytes-1);
		  }
		  else
		  {
		      pipe_raw2api_proc(input_ptr,ts_pid,input_remaining_bytes);
		  }
		break;

	      case EMPTY:  /* NULL table (0x1FFF) */
		  break;

	    default:
		write_msg('P', 'I', "UNKNOWN pkt -- skipped");
		break;
	  }
	}
}

/* --------------------------------------------------------------------- */

/*
 *  Enqueue a buf_hdr on TS control queue
 */

enqueue_buf(q_hdr,buf_hdr_ptr)
struct queue_hdr *q_hdr;
struct buf_hdr   *buf_hdr_ptr;
{
	// check for bad streams
	if ((q_hdr == 0) || (buf_hdr_ptr == 0)) return;
	
	if ( q_hdr->q_tail_ptr == NULL )
		q_hdr->q_head_ptr = buf_hdr_ptr;
	else
	    q_hdr->q_tail_ptr->q_ptr = buf_hdr_ptr;
	q_hdr->q_tail_ptr = buf_hdr_ptr;
	buf_hdr_ptr->q_ptr = NULL;
}
/* --------------------------------------------------------------------- */

/*
 *  Dequeue a buf_hdr from a TS control queue
 */

struct buf_hdr  *dequeue_buf(q_hdr)
struct queue_hdr *q_hdr;
{
	struct buf_hdr    *buf_hdr_ptr;

	if ( (buf_hdr_ptr=q_hdr->q_head_ptr) == NULL )
		return( buf_hdr_ptr );

	if ( q_hdr->q_head_ptr == q_hdr->q_tail_ptr )
		q_hdr->q_head_ptr = q_hdr->q_tail_ptr = NULL;
	else
		q_hdr->q_head_ptr = buf_hdr_ptr->q_ptr;
	return( buf_hdr_ptr );
}

/* --------------------------------------------------------------------- */

/*
 *  Create and init a queue
 */

struct queue_hdr  *create_queue()
{
	struct queue_hdr *q_hdr;

	q_hdr = malloc( sizeof(struct queue_hdr) );
	q_hdr->q_head_ptr  = NULL;
	q_hdr->q_tail_ptr  = NULL;
	q_hdr->q_valid     = 0;
	q_hdr->q_tot_bytes = 0;
	q_hdr->q_ver       = 0;

	return( q_hdr );
}

/* --------------------------------------------------------------------- */

/*
 * Delete an existing queue and ALL of its associated buffers
 * But don't delete the queue hdr
 */
clear_queue(q_hdr)
struct   queue_hdr     *q_hdr;
{
	struct buf_hdr  *buf_hdr_ptr, *nxt_buf_hdr_ptr;

	nxt_buf_hdr_ptr = q_hdr->q_head_ptr;
	while ( (buf_hdr_ptr=nxt_buf_hdr_ptr) != NULL )
	{
		if ( buf_hdr_ptr->buf_ptr != NULL )
			free( buf_hdr_ptr->buf_ptr );
		nxt_buf_hdr_ptr = buf_hdr_ptr->q_ptr;
		free( buf_hdr_ptr );
	}
	q_hdr->q_head_ptr = NULL;
	q_hdr->q_tail_ptr = NULL;
	q_hdr->q_valid = 0;
}

/* --------------------------------------------------------------------- */

print_queue(q_hdr)
struct queue_hdr *q_hdr;
{
	int i;
	struct buf_hdr    *buf_hdr_ptr;

	if ( q_hdr == NULL )
	{
		write_msg('P', 'D', "No queue HDR");
		return;
	}
	write_msg('P', 'D', "queue HDR@0x%x", q_hdr);
	if ( (buf_hdr_ptr=q_hdr->q_head_ptr) == NULL )
	{
		write_msg('P', 'D', "HDR, but No queue");
		return;
	}
	i = 0;
	write_msg('P', 'D', "Q_HDR valid=%d, size=%d, ver=%d",
		q_hdr->q_valid, q_hdr->q_tot_bytes, q_hdr->q_ver );
	while ( buf_hdr_ptr != NULL )
	{
		write_msg('P', 'D', "%d. pid=%d, loc=%d, chk=%d", i, buf_hdr_ptr->pid,
		       buf_hdr_ptr->byte_loc, buf_hdr_ptr->bytes_checked);
		if ( buf_hdr_ptr->buf_ptr == NULL )
		     write_msg('P', 'D', "  buf Empty");
		else
		     write_msg('P', 'D', "  Tbl_ID=0x%x", buf_hdr_ptr->buf_ptr[0]&0xFF);
		i++;
		buf_hdr_ptr = buf_hdr_ptr->q_ptr;
	}
}
/* --------------------------------------------------------------------- */

/*
 *  Process input Transport Pkt to accumulate a Table Section
 */

accum_sect(ts_payload_start,input_ptr,input_remaining_bytes,ts_control_ptr,pid)
unsigned int  ts_payload_start;
char          *input_ptr;
int           input_remaining_bytes;
struct ts_control   *ts_control_ptr;
unsigned int        pid;
{
    int                i, input_cnt, offset, bytes_left;
    char               *dest_ptr;
    unsigned int       tbl_sect_len;
    struct   buf_hdr   *buf_hdr_ptr;

/* get current accumulate buf */

    buf_hdr_ptr = ts_control_ptr->ts_buf_hdr_ptr;

/* check if this is the beginning of a Table Section */

    if ( ts_payload_start == 1 )
    {
/*
 * If table pack is allowed AND a packed Table can cross TS pkt boundaries,
 * Then one needs to test if a table fragment exists in the beginning
 * of this TS, before the new Table
 */
       offset = (*input_ptr) & 0xFF;
       if ( (buf_hdr_ptr != NULL) && (offset > 0) )
       {     /* accum sect buf open & a fragment exists */
	   if ( buf_hdr_ptr->sect_bytes_left <= offset )
	   {      /* if the frag size can accommodate the num of         */
		  /* remaining bytes in the sect get them otherwise skip */

		 /* copy last chunk of table section */

	       bytes_left = buf_hdr_ptr->sect_bytes_left;
	       dest_ptr = &buf_hdr_ptr->buf_ptr[buf_hdr_ptr->byte_loc];
	       bcopy(input_ptr, dest_ptr, bytes_left);

	       buf_hdr_ptr->byte_loc += bytes_left;
	       buf_hdr_ptr->sect_bytes_left = 0;

		  /* enqueue this completed sect for processing */

	       enqueue_buf(ts_control_ptr->ts_queue_hdr,buf_hdr_ptr);

		  /* remove it from accumulate status */

	       ts_control_ptr->ts_buf_hdr_ptr = buf_hdr_ptr = NULL;
	   }
       }
/*
 * beginning of a table, 1st byte is an offset (ptr) - skip over
 * specified number of byte (should be zero - i.e., skip 1 byte)
 */
       input_remaining_bytes -= (offset + 1);
       input_ptr += (offset + 1);

more_stuff:
 /*
  * This label provides an entry point for a repeat loop when multiple
  * tables are packed into a TS pkt
 */
/*
write_msg('P', 'D', "");
write_msg('P', 'D', "a-1st byte=0x%x", input_ptr[0]&0xFF);
*/
	if ( input_remaining_bytes > 2 )
	{
		tbl_sect_len  = ((input_ptr[1]&0xF)<<8) | (input_ptr[2]&0xFF);
	}
	else
	{
		tbl_sect_len  = 0;
	}
/* if buf_hdr_ptr not NULL, then a previous table has not completed */

	if ( buf_hdr_ptr == NULL )
	{
	  /* allocate a new accumlate buf_hdr & place it in accumulate status */

		buf_hdr_ptr = malloc( sizeof(struct buf_hdr) );
		ts_control_ptr->ts_buf_hdr_ptr = buf_hdr_ptr;

		if ( tbl_sect_len  > 0 )
		{     /* allocate a correct size buf */
		    buf_hdr_ptr->buf_ptr = malloc( tbl_sect_len+3 );
		    buf_hdr_ptr->buf_size = tbl_sect_len + 3;
		}
		else
		{     /* allocate a full size buf */
		    buf_hdr_ptr->buf_ptr = malloc( MAX_SECTION_SIZE );
		    buf_hdr_ptr->buf_size = MAX_SECTION_SIZE;
		}
	}
	else
	{
		if ( tbl_sect_len > 0 )
		{
			if ( buf_hdr_ptr->buf_size < tbl_sect_len )
			{
				free( buf_hdr_ptr->buf_ptr );
				buf_hdr_ptr->buf_ptr = malloc( tbl_sect_len );
				buf_hdr_ptr->buf_size = tbl_sect_len;
			}
		}
		else
		{
			if ( buf_hdr_ptr->buf_size != MAX_SECTION_SIZE )
			{
			    free( buf_hdr_ptr->buf_ptr );
			    buf_hdr_ptr->buf_ptr = malloc( MAX_SECTION_SIZE );
			    buf_hdr_ptr->buf_size = MAX_SECTION_SIZE;
			}
		}
	}
	   /* init buf_hdr, if one already existed then it was a previous   */
	   /* sect which didn't successfully complete  -- rcv error, ignore */

	buf_hdr_ptr->valid = 2;      /* set to building state */
	buf_hdr_ptr->sect_bad_missing = 0;
	buf_hdr_ptr->byte_loc = 0;
	buf_hdr_ptr->bytes_checked = 0;

/* if we were guranteed at least 3 byte fragments, then we could always */
/* allocate the correct size buf                                                   */

	if ( input_remaining_bytes > 2 )
	{
		buf_hdr_ptr->sect_bytes_left = tbl_sect_len + 3;
	}
	else
	{
	   /* initial fragment too small to get Sect len */
	   /* store fragment, but set bytes_left = 0     */

		buf_hdr_ptr->sect_bytes_left = 0;
		buf_hdr_ptr->buf_ptr[buf_hdr_ptr->byte_loc++] = *input_ptr++;
		if ( input_remaining_bytes == 2 )
		    buf_hdr_ptr->buf_ptr[buf_hdr_ptr->byte_loc++] = *input_ptr++;
		return;
	}
    }

/*
 * Check if a buf_hdr is active signaling that
 * a table accumulation is in the process of being built
 */

    if ( (buf_hdr_ptr!=NULL) && (buf_hdr_ptr->valid==2) )
    {
if ( buf_hdr_ptr->buf_ptr == NULL )
	write_msg('P', 'E', "buf is NULL");

		/* an active buf exists, implying accum a sect */
	if ( buf_hdr_ptr->sect_bytes_left == 0 )
	{
	   /* initial fragment was too small to get Sect len */
	   /* get it now & init bytes_left                   */

	    if ( buf_hdr_ptr->byte_loc = 1 )
	    {
		tbl_sect_len = ((input_ptr[0]&0xF)<<8) | (input_ptr[1]&0xFF);
		buf_hdr_ptr->sect_bytes_left = tbl_sect_len + 2;
	    }
	    else
	    {
		tbl_sect_len = ((buf_hdr_ptr->buf_ptr[1]&0xF)<<8)
						     | (input_ptr[0]&0xFF);
		buf_hdr_ptr->sect_bytes_left = tbl_sect_len + 1;
	    }
	}

/*  ***
write_msg('P', 'D', "buf_hdr_ptr continuing, sect_bytes_left = 0x%x", buf_hdr_ptr->sect_bytes_left);
*/

/* make sure there is enough room in buf for data */

	if ( (buf_hdr_ptr->buf_size-buf_hdr_ptr->byte_loc)
					< buf_hdr_ptr->sect_bytes_left )
	{
		write_msg('P', 'E', "buf too small in accum_sect()");
		write_msg('P', 'E', "buf_size=%d buf_loc=%d bytes_left=%d",
			buf_hdr_ptr->buf_size, buf_hdr_ptr->byte_loc,
				buf_hdr_ptr->sect_bytes_left);

		return;
	}
/* get the location in the table buf to append new data to */

	dest_ptr = &buf_hdr_ptr->buf_ptr[buf_hdr_ptr->byte_loc];
	if ( buf_hdr_ptr->sect_bytes_left > input_remaining_bytes )
	{
/* copy next chunk of data from input, not yet end of table section */

	    bcopy(input_ptr, dest_ptr, input_remaining_bytes);
	    buf_hdr_ptr->byte_loc += input_remaining_bytes;
	    buf_hdr_ptr->sect_bytes_left -= input_remaining_bytes;
	    return;
	}
	else
	{
 /* copy last chunk of table section */

	    bcopy(input_ptr, dest_ptr, buf_hdr_ptr->sect_bytes_left);
	    buf_hdr_ptr->byte_loc += buf_hdr_ptr->sect_bytes_left;

	    input_ptr += buf_hdr_ptr->sect_bytes_left;
	    input_remaining_bytes -= buf_hdr_ptr->sect_bytes_left;
	    buf_hdr_ptr->sect_bytes_left = 0;
	    buf_hdr_ptr->pid             = pid;

	       /* enqueue this completed sect for processing */

	    enqueue_buf(&(ts_control_ptr->ts_queue_hdr),buf_hdr_ptr);
/*
 start_msg("inp(0x%x)=", dest_ptr);
 for (i=0; i<30; i++)
     append_msg("%2x ", dest_ptr[i]&0xFF);
 end_msg('P', 'D', "");
write_msg('P', 'D', "c-1st byte=0x%x", buf_hdr_ptr->buf_ptr[0]&0xFF);
write_msg('P', 'D', "queue HDR at 0x%x", &(ts_control_ptr->ts_queue_hdr) );
print_queue( &(ts_control_ptr->ts_queue_hdr) );
*/
	       /* remove it from accumulate status */

	    ts_control_ptr->ts_buf_hdr_ptr = buf_hdr_ptr = NULL;
	}
     }
     else
     {
	return;  /* Not processing a table, skip data */
     }

     if ( (input_remaining_bytes>0) && ((*input_ptr&0xFF)!=0xFF) )
     {
   /*     prt_extra_pkt(input_ptr,input_remaining_bytes);       */
	goto more_stuff;
     }
     return;
}

/* --------------------------------------------------------------------- */

int tvct_table_type;
int cvct_table_type;
int rrt_table_type;
int dcct_table_type;
int dccsct_table_type;

/* --------------------------------------------------------------------- */

/*
 *  Process a full Table Section from the Transport Stream for
 *  MGT, TVCT, CVCT, STT, RRT tables
 */

atsc_tables(buf_hdr_ptr,ts_control_ptr)
struct buf_hdr    *buf_hdr_ptr;
struct ts_control *ts_control_ptr;
{
    int          i, bytes, sect_cnt;
    char         *dest_ptr, *crc_ptr, *crc__ptr;
    unsigned int dest_len;
    unsigned int tbl_id, tbl_sect_len, tbl_cur_next, tbl_ver_num;
    unsigned int tbl_sect_num, tbl_last_sect, tbl_proto_ver;
    unsigned int rating_region, crc, sect_num, sect_len, last_table_type;
    struct  mgt_tbl_entry  *mgt_tbl_entry_ptr;
    struct  queue_hdr      *q_hdr;
    struct  buf_hdr        *cur_link, *prev_link;

	dest_ptr = buf_hdr_ptr->buf_ptr;
	if ( buf_hdr_ptr->buf_ptr == NULL )
	{
      /* no data, discard sect - deallocate buf_hdr */

	     goto discard2;
	}
/* extract basic table header info */

	tbl_id        = (*dest_ptr) & 0xFF;
	tbl_sect_len  = ((dest_ptr[1]&0xF)<<8) | (dest_ptr[2]&0xFF);
	
// check length (for bad streams)
	dest_len = buf_hdr_ptr->buf_size;
	if ((tbl_sect_len > dest_len) || (tbl_sect_len < 2)) {
		goto discard;
	}	

/*  ***
write_msg('P', 'D', "tbl_sect_len = 0x%x", tbl_sect_len);
*/
	tbl_cur_next  = (dest_ptr[5]&0x1);
	tbl_ver_num   = ((dest_ptr[5]>>1) & 0x1F);
	tbl_sect_num  = (dest_ptr[6] & 0xFF);
	tbl_last_sect = (dest_ptr[7] & 0xFF);
	tbl_proto_ver = (dest_ptr[8] & 0xFF);
	if ( tbl_proto_ver != 0 )
	{
      /* discard sect - deallocate buf_hdr & buf */

	     goto discard;
	}

#ifdef DEEP_DEBUG
     write_msg('P', 'D', " id=%2x, len=%x(%d), cur=%x, ver=%x, sect=%x, last=%x, proto=%x",
     tbl_id, tbl_sect_len, tbl_sect_len, tbl_cur_next, tbl_ver_num, tbl_sect_num,
     tbl_last_sect, tbl_proto_ver);

    start_msg("inp(0x%x)=", dest_ptr);
    for (i=0; i<30; i++)
	append_msg("%2x ", dest_ptr[i]&0xFF);
    end_msg('P', 'D', "");
#endif

 /* extract CRC from end of table section in table buf */

	crc_ptr = &buf_hdr_ptr->buf_ptr[buf_hdr_ptr->byte_loc-4];
	crc = ( (((*crc_ptr)&0xFF)<<24) | ((crc_ptr[1]&0xFF)<<16)
	       | ((crc_ptr[2]&0xFF)<<8) | (crc_ptr[3]&0xFF) );

  /*
   * pass buf and length, not including the CRC field, to check CRC
   * which follows the data
   */
	if ( check_crc(dest_ptr,tbl_sect_len-1,crc) == ERROR )
	{
	 /* bad CRC chk, ignore table */
	     goto discard;
	}

/* CRC check is good, update number of checked bytes in table buf */

	buf_hdr_ptr->bytes_checked = buf_hdr_ptr->byte_loc;
	buf_hdr_ptr->valid         = 1;

  /* use table ID to find associated buf for this new table */

	switch ( tbl_id )
	{
	   case 0xC7:      /* MGT */
	       num_mgt++;
	       q_hdr = &mgt_q_hdr;
	       if ( q_hdr->q_valid == 1 )
	       {
		   if ( tbl_ver_num != ((q_hdr->q_ver+ 1)%32) )
		   {
#ifdef DEBUG_ERR_MSG
			write_msg('P', 'D', "not next version -- ignore!!");
#endif
			goto discard;  /* not next version -- ignore!! */
		   }
		/* need to deallocate existing table Q & assoc struc */

		   clear_queue( q_hdr );
		   delete_mgt_table();
	       }
	       break;

	   case 0xC8:      /* TVCT */
	       num_vct++;
	       tvct_table_type = 0x0000|(tbl_cur_next^1);
	       if ( (mgt_tbl_entry_ptr=search_mgt_for_type(tvct_table_type)) == NULL )
	       {                        /* 0x0000 or 0x0001 table_type */
		   //write_msg('P', 'E', "error no TVCT entry in MGT in atsc_tables()");
	    /* no entry in MGT -- ignore!!, but must have one?? */
		   goto discard;
	       }
	       if ( (tbl_ver_num != mgt_tbl_entry_ptr->table_type_version_number)
		 || (mgt_tbl_entry_ptr->table_type_q_ptr->q_valid==1)  )
	       {
#ifdef DEBUG_ERR_MSG
		   write_msg('P', 'D', "TVCT version doesn't match or buf already valid in atsc_tables()");
#endif
	  /* version doesn't match or buf already valid -- ignore!! */
		   goto discard;
	       }
	       q_hdr = mgt_tbl_entry_ptr->table_type_q_ptr;
	       break;

	   case 0xC9:      /* CVCT */
	       num_mgt++;
	       cvct_table_type = 0x0002|(tbl_cur_next^1);
	       if ( (mgt_tbl_entry_ptr=search_mgt_for_type(cvct_table_type)) == NULL )
	       {                        /* 0x0002 or 0x0003 table_type */
#ifdef DEBUG_ERR_MSG
		   write_msg('P', 'D', "CVCT tbl_cur_next=0x%x", tbl_cur_next);
		   write_msg('P', 'D', "CVCT no entry in MGT in atsc_tables()");
#endif
	  /* no entry in MGT -- ignore!!?? */
		   goto discard;
	       }
	       if ( (tbl_ver_num != mgt_tbl_entry_ptr->table_type_version_number)
		 || (mgt_tbl_entry_ptr->table_type_q_ptr->q_valid==1)  )
	       {
#ifdef DEBUG_ERR_MSG
		   write_msg('P', 'D', "CVCT version doesn't match or buf already valid in atsc_tables()");
#endif
	  /* version doesn't match or buf already valid -- ignore!! */
		   goto discard;
	       }
	       q_hdr = mgt_tbl_entry_ptr->table_type_q_ptr;
	       break;

	   case 0xCA:      /* RRT */
	       num_rrt++;
	       rating_region =  dest_ptr[4] & 0xFF;
	       rrt_table_type = 0x300|rating_region;
	       if ( (mgt_tbl_entry_ptr=search_mgt_for_type(rrt_table_type)) == NULL )
	       {                        /* 0x0301 thru 0x03FF table_type */
#ifdef DEBUG_ERR_MSG
		   write_msg('P', 'D', "RRT no entry in MGT in atsc_tables()");
#endif
	  /* no entry in MGT -- ignore!! */
		   goto discard;
	       }
	       q_hdr = mgt_tbl_entry_ptr->table_type_q_ptr;

	       if ( (q_hdr->q_valid == 1)  && (q_hdr->q_head_ptr!=NULL) )
	       {
		   if ( (tbl_ver_num == mgt_tbl_entry_ptr->table_type_version_number)
			   &&  (q_hdr->q_ver != tbl_ver_num) )
		   {
		      clear_queue( q_hdr ); /* delete any prev Q */
		   }
		   else
		   {
#ifdef DEBUG_ERR_MSG
		      write_msg('P', 'D', "RRT version doesn't match or buf already valid in atsc_tables()");
#endif
		     /* version doesn't match or buf already valid -- ignore!! */
		      goto discard;
		   }
	       }
	       else if(tbl_ver_num != mgt_tbl_entry_ptr->table_type_version_number)
	       {
			  /* version doesn't match -- ignore!! */
		   goto discard;
	       }
	       break;

	   case 0xCD:      /* STT, no ver num -- always valid */
	       num_stt++;
	       q_hdr = &stt_q_hdr;
	       clear_queue( q_hdr ); /* delete any prev Q */
	       break;

	   case 0xD3:      /* DCCT - Directed Channel Change Table */
	       num_dcct++;
	       dcct_table_type = 0x0005;
	       if ( (mgt_tbl_entry_ptr=search_mgt_for_type(dcct_table_type)) == NULL )
	       {                        /* 0x0005 table_type */
		   write_msg('P', 'E', "error no DCCT entry in MGT in atsc_tables()");
		      /* no entry in MGT -- ignore!!?? */
		   goto discard;
	       }
	       if ( (tbl_ver_num != mgt_tbl_entry_ptr->table_type_version_number)
		 || (mgt_tbl_entry_ptr->table_type_q_ptr->q_valid==1)  )
	       {
#ifdef DEBUG_ERR_MSG
		   write_msg('P', 'D', "DCCT version doesn't match or buf already valid in atsc_tables()");
#endif
		    /* version doesn't match or buf already valid -- ignore!! */
		   goto discard;
	       }
	       q_hdr = mgt_tbl_entry_ptr->table_type_q_ptr;
	       break;

	   case 0xD4:    /* DCCSCT - Directed Channel Change Selection Code Table */
	       num_dccsct++;
	       dccsct_table_type = 0x0005;
	       if ( (mgt_tbl_entry_ptr=search_mgt_for_type(dccsct_table_type)) == NULL )
	       {                        /* 0x0006 table_type */
		   write_msg('P', 'E', "error no DCCSCT entry in MGT in atsc_tables()");
		      /* no entry in MGT -- ignore!!?? */
		   goto discard;
	       }
	       if ( (tbl_ver_num != mgt_tbl_entry_ptr->table_type_version_number)
		 || (mgt_tbl_entry_ptr->table_type_q_ptr->q_valid==1)  )
	       {
#ifdef DEBUG_ERR_MSG
		   write_msg('P', 'D', "DCCSCT version doesn't match or buf already valid in atsc_tables()");
#endif
		    /* version doesn't match or buf already valid -- ignore!! */
		   goto discard;
	       }
	       q_hdr = mgt_tbl_entry_ptr->table_type_q_ptr;
	       break;

	    default:
#ifdef DEBUG_RCV
	       write_msg('P', 'D', "ERROR unknown table ID in atsc_tables()");
#endif
	  /* unknown table_id -- skip */
	       goto discard;
	       break;
	}

/* insert table section in the list ordered by section number */

	if ( (cur_link=q_hdr->q_head_ptr) == NULL )
	{
	      /* Q empty add table section */
	    enqueue_buf(q_hdr,buf_hdr_ptr);
	      /* init table size */
	    q_hdr->q_tot_bytes = buf_hdr_ptr->bytes_checked;
	}
	else
	{
	      /* walk thru ordered list and add in before a higher sect_num */

	    prev_link = NULL;
	    do
	    {
		if ( tbl_sect_num < (sect_num=(cur_link->buf_ptr[6]&0xFF)) )
		{
		    /* insert before current entry */
			buf_hdr_ptr->q_ptr = cur_link;
			if ( prev_link == NULL )
				q_hdr->q_head_ptr = buf_hdr_ptr;
			else
			    prev_link->q_ptr = buf_hdr_ptr;
				/* add to table size */
			q_hdr->q_tot_bytes += buf_hdr_ptr->bytes_checked;
			break;
		}
		if ( tbl_sect_num == sect_num )
		{
			  /* duplicate entry, discard */
			goto discard;
		}
		prev_link = cur_link;
		if ( (cur_link=cur_link->q_ptr) == NULL )
		{
		      /* add to end of queue */
		    enqueue_buf(q_hdr,buf_hdr_ptr);
		      /* add to table size */
		    q_hdr->q_tot_bytes += buf_hdr_ptr->bytes_checked;
		}

	    } while ( cur_link != NULL );
	}

/* check if all sections are assembled and verified */

	sect_cnt = 0;
	cur_link = q_hdr->q_head_ptr;
	while ( cur_link != NULL )
	{
	     if ( sect_cnt != (cur_link->buf_ptr[6]&0xFF) )
	     {
		 return; /* at least 1 section is still missing */
	     }
	     if ( (cur_link=cur_link->q_ptr) == NULL )
	     {
		 break;
	     }
	     sect_cnt++;
	}
	if ( sect_cnt != (q_hdr->q_head_ptr->buf_ptr[7]&0xFF) )
	{
	     return; /* Last Section is still missing */
	}

/* table, as an ordered (by sect num) linked list, is verified,
 * mark it valid & process it as required */

	q_hdr->q_valid = 1;          /* full table checked & valid */
	q_hdr->q_ver   = tbl_ver_num;

	last_table_type = 0;
	switch ( (*dest_ptr)&0xFF )
	{
	   case 0xC7:      /* MGT */
	     num_mgt_comp++;
	     build_mgt_table(q_hdr);
	     break;
	   case 0xC8:      /* TVCT */
	     num_vct_comp++;
	     last_table_type = tvct_table_type;
	     if ( tbl_cur_next == 0 )
		     break;     /* don't build struct for next ver of tbl */
	     extract_vct_stream_info(q_hdr,&current_tvct_hdr);
	     break;
	   case 0xC9:      /* CVCT */
	     num_vct_comp++;
	     last_table_type = cvct_table_type;
	     if ( tbl_cur_next == 0 )
		     break;     /* don't build struct for next ver of tbl */
	     extract_vct_stream_info(q_hdr,&current_cvct_hdr);
	     break;
	   case 0xCA:      /* RRT */
	     num_rrt_comp++;
#ifdef DEBUG_SUMMARY
	     prt_rrt_table(q_hdr);
#endif
	     last_table_type = rrt_table_type;
	     break;
	   case 0xCD:      /* STT */
	     num_stt_comp++;
	     build_stt_table(q_hdr);
	     break;
	   case 0xD3:      /* DCCT */
	     num_dcct_comp++;
	     build_dcct_table(q_hdr);
	     last_table_type = dcct_table_type;
	     break;
	   case 0xD4:      /* DCCSCT */
	     num_dccsct_comp++;
	     build_dccsct_table(q_hdr);
	     last_table_type = dccsct_table_type;
	     break;
	   default:
	     break;
	}
/* pass verified table to API simulator process, PID=0x1FFB, last_table_type
 */
	pipe_q2api_proc(q_hdr, 0x1FFB, last_table_type);
	return;

/* discard Table Sect */

discard:
     free( buf_hdr_ptr->buf_ptr );
discard2:
     free( buf_hdr_ptr );
     return;
}

/* --------------------------------------------------------------------- */

prt_extra_pkt(input_ptr,input_remaining_bytes)
char *input_ptr;
int  input_remaining_bytes;
{
	int i;

	start_msg("extra (0x%x) stuff in pkt =", input_remaining_bytes);
	for ( i=0; i<input_remaining_bytes; i++ )
	{
		append_msg(" %2x", *input_ptr++ & 0xFF);
		if ( (i%17) == 16 )
		{
		   end_msg('P', 'D', "");
		   start_msg("");
		   if ( ((input_ptr[-1]& 0xFF)==0xFF)
			&&  ((input_ptr[0]& 0xFF)==0xFF)
			&&  ((input_ptr[1]& 0xFF)==0xFF) )
		      break;
		}
	}
	end_msg('P', 'D', " end");
	return;
}

/* --------------------------------------------------------------------- */

   /* most other tables */

/*
 *  Process a full Table Section from the Transport Stream for most
 *  other tables (i.e., EIT, ETT, DET, LTST, PAT, CAT, PMT, DST, NRT)
 */

accum_tables(buf_hdr_ptr,ts_control_ptr,q_hdr,ts_pid)
struct   buf_hdr    *buf_hdr_ptr;
struct   ts_control *ts_control_ptr;
struct   queue_hdr  *q_hdr;
unsigned int      ts_pid;
{
    int           i, bytes, sect_cnt;
    char          *dest_ptr, *crc_ptr;
    unsigned int dest_len;
    unsigned char c_char;
    unsigned int  tbl_id, tbl_sect_len, tbl_cur_next, tbl_ver_num;
    unsigned int  tbl_sect_num, tbl_last_sect, tbl_proto_ver;
    unsigned int  rating_region, crc, sect_num, sect_len, ver_num;
    unsigned int  next_ver_num;
    struct buf_hdr        *cur_link, *prev_link;
    struct dst_nrt_q_lst  *q_lst_ptr;
    struct mgt_tbl_entry  *mgt_tbl_entry_ptr;

	dest_ptr = buf_hdr_ptr->buf_ptr;
	if ( buf_hdr_ptr->buf_ptr == NULL )
	{
      /* no data, discard sect - deallocate buf_hdr */

	     goto discard2;
	}
/* extract basic table header info */

	tbl_id        = (*dest_ptr) & 0xFF;
	tbl_sect_len  = ((dest_ptr[1]&0xF)<<8) | (dest_ptr[2]&0xFF);

// check length (for bad streams)
	dest_len = buf_hdr_ptr->buf_size;
	if ((tbl_sect_len > dest_len) || (tbl_sect_len < 2)) {
		goto discard;
	}	

/*  ***
write_msg('P', 'D', "tbl_sect_len = 0x%x", tbl_sect_len);
*/
	tbl_cur_next  = (dest_ptr[5]&0x1);
	tbl_ver_num   = ((dest_ptr[5]>>1) & 0x1F);
	tbl_sect_num  = (dest_ptr[6] & 0xFF);
	tbl_last_sect = (dest_ptr[7] & 0xFF);
	tbl_proto_ver = (dest_ptr[8] & 0xFF);

#ifdef DEEP_DEBUG
     write_msg('P', 'D', " id=%2x, len=%x(%d), cur=%x, ver=%x, sect=%x, last=%x, proto=%x",
     tbl_id, tbl_sect_len, tbl_sect_len, tbl_cur_next, tbl_ver_num, tbl_sect_num,
     tbl_last_sect, tbl_proto_ver);

    start_msg("inp(0x%x)=", dest_ptr);
    for (i=0; i<30; i++)
	append_msg("%2x ", dest_ptr[i]&0xFF);
    end_msg('P', 'D', "");
#endif

 /* extract CRC from end of table section in table buf */

	crc_ptr = &buf_hdr_ptr->buf_ptr[buf_hdr_ptr->byte_loc-4];
	crc = ( (((*crc_ptr)&0xFF)<<24) | ((crc_ptr[1]&0xFF)<<16)
	       | ((crc_ptr[2]&0xFF)<<8) | (crc_ptr[3]&0xFF) );

  /*
   * pass buf and length, not including the CRC field, to check CRC
   * which follows the data
   */
	if ( check_crc(dest_ptr,tbl_sect_len-1,crc) == ERROR )
	{
#ifdef DEBUG_ERR_MSG
	     write_msg('P', 'D', "error bad CRC in accum_tables()");
#endif
	 /* bad CRC chk, ignore table */
	     goto discard;
	}

/* CRC check is good, update number of checked bytes in table buf */

	buf_hdr_ptr->bytes_checked = buf_hdr_ptr->byte_loc;
	buf_hdr_ptr->valid         = 1;

	switch ( tbl_id )
	{
	   case 0xCB:      /* EIT */
	   case 0xCC:      /* ETT */
	   case 0xCE:      /* DET */
	   case 0xD2:      /* LTST*/
	       if ( tbl_proto_ver != 0 )
	       {       /* DET, LTST, EIT & ETT protocol_version limit */
#ifdef DEBUG_ERR_MSG
		    write_msg('P', 'D', "error protocol_version not zero in accum_tables()");
#endif
		    goto discard;
	       }
	       if ( q_hdr->q_valid == 1 )
	       {
		   if ( tbl_ver_num != ((q_hdr->q_ver+ 1)%32) )
		   {
#ifdef DEBUG_ERR_MSG
		   write_msg('P', 'D', "tbl_ver_num=0x%x", tbl_ver_num);
		   write_msg('P', 'D', "version doesn't match or buf already valid in accum_tables()");
#endif
		     goto discard;  /* version doesn't match or buf already valid -- Ignore!! */
		   }
		/* new ver, need to deallocate existing table Q & */
		/* any internal assoc struc */
		   clear_queue( q_hdr );   /* Q, but no internal struct here */
	       }
	       break;
	   case 0x00:      /* PAT */
	   case 0x01:      /* CAT */
	   case 0x02:      /* PMT */
	       if ( q_hdr->q_valid == 1 )
	       {
		   if ( tbl_ver_num != ((q_hdr->q_ver+ 1)%32) )
		   {
#ifdef DEBUG_ERR_MSG
			write_msg('P', 'D', "not next version -- ignore!!");
#endif
			goto discard;  /* not next version -- ignore!! */
		   }
		/* new ver, need to deallocate existing table Q & */
		/* any internal assoc struc */
		   clear_queue( q_hdr );
		   if ( tbl_id == 0x00 )  /* only PAT internal struct exists */
		   {
			delete_pat();
		   }
	       }
	       break;
  /* use table ID to process specific table */

	   case 0xCF:      /* DST */
	       if ( tbl_proto_ver != 1 )
	       {       /* DSTprotocol_version */
#ifdef DEBUG_ERR_MSG
		     write_msg('P', 'D', "error protocol_version not one in accum_tables()");
#endif
		     goto discard;  /* not correct version -- ignore!! */
	       }

	   case 0xD1:      /* NRT */
	       q_lst_ptr =
		 (struct dst_nrt_q_lst *) ts_control_ptr->ts_ctrl_gen.ts_ctrl_ptr;
	       if ( tbl_id  == 0xCF )
		   q_hdr = q_lst_ptr->dst_q_hdr;
	       else
		   q_hdr = q_lst_ptr->nrt_q_hdr;
	       if ( q_hdr->q_valid == 1 )
	       {
		   if ( tbl_ver_num != ((q_hdr->q_ver+ 1)%32) )
		   {
#ifdef DEBUG_ERR_MSG
			write_msg('P', 'D', "tbl_ver_num=0x%x", tbl_ver_num);
			write_msg('P', 'D', "version doesn't match or buf already valid in accum_tables()");
#endif
		     goto discard;  /* version doesn't match or buf already valid -- Ignore!! */
		   }
		/* new ver, need to deallocate existing table Q & */
		/* any internal assoc struc */
		   clear_queue( q_hdr );   /* Q, but no internal struct here */
	       }
	       break;

	   default:
#ifdef DEBUG_RCV
	       write_msg('P', 'D', "unknown table_id in accum_tables()-- skip!!");
#endif
	       goto discard;  /* unknown table_id -- skip */
	       break;
	}

/* insert table section in the list ordered by section number */

	if ( (cur_link=q_hdr->q_head_ptr) == NULL )
	{
	      /* Q empty add table section */
	    enqueue_buf(q_hdr,buf_hdr_ptr);
	      /* init table size */
	    q_hdr->q_tot_bytes = buf_hdr_ptr->bytes_checked;
	}
	else
	{
	      /* walk thru ordered list and add in before a higher sect_num */
			   /* special handling for EIT & ETT tables */
	    prev_link = NULL;
	    do
	    {
		if ( tbl_sect_num <= (sect_num=(cur_link->buf_ptr[6]&0xFF)) )
		{
		    if ( tbl_sect_num == sect_num )
		    {
			if (  (tbl_id!=0xCB && tbl_id!=0xCC) )
			{
			      /* duplicate entry, discard */
			    goto discard;
			}
			   /* special handling for EIT & ETT tables   */
			   /* this should really order by & check for */
			   /* duplicate 2nd key (table_id_extention)  */
		    }
		    /* insert before current entry */
			buf_hdr_ptr->q_ptr = cur_link;
			if ( prev_link == NULL )
			    q_hdr->q_head_ptr = buf_hdr_ptr;
			else
			    prev_link->q_ptr = buf_hdr_ptr;
				/* add to table size */
			q_hdr->q_tot_bytes += buf_hdr_ptr->bytes_checked;
			break;
		}
		prev_link = cur_link;
		if ( (cur_link=cur_link->q_ptr) == NULL )
		{
		      /* add to end of queue */
		    enqueue_buf(q_hdr,buf_hdr_ptr);
		      /* add to table size */
		    q_hdr->q_tot_bytes += buf_hdr_ptr->bytes_checked;
		}

	    } while ( cur_link != NULL );
	}

/* check if all sections are assembled and verified */

	if (  (tbl_id==0xCB || tbl_id==0xCC) )
	{       /* special handling for EIT & ETT tables */
	    mgt_tbl_entry_ptr =
	       (struct mgt_tbl_entry *) ts_control_ptr->ts_ctrl_gen.ts_ctrl_ptr;
	    if ( q_hdr->q_tot_bytes < mgt_tbl_entry_ptr->number_bytes )
	    {
		return; /* Section is still missing */
	    }
	}
	else
	{
	    sect_cnt = 0;
	    cur_link = q_hdr->q_head_ptr;
	    while ( cur_link != NULL )
	    {
		 bytes += cur_link->bytes_checked;
		 if ( sect_cnt != (cur_link->buf_ptr[6]&0xFF) )
		 {
		     return; /* at least 1 section is still missing */
		 }
		 if ( (cur_link=cur_link->q_ptr) == NULL )
		 {
		      break;
		 }
		 sect_cnt++;
	    }
	    if ( sect_cnt != (q_hdr->q_head_ptr->buf_ptr[7]&0xFF) )
	    {
		 return; /* Section is still missing */
	    }
	}
/* table, as an ordered (by sect num) linked list, is verified,
 * mark it valid & process it as required */

	q_hdr->q_valid = 1;         /* full table checked & valid */
	q_hdr->q_ver   = tbl_ver_num;

	switch ( tbl_id )
	{
	   case 0x00:      /* PAT */
	       build_pat(q_hdr);
	       break;
	   case 0x01:      /* CAT */
	       break;
	   case 0x02:      /* PMT */
	       build_pmt(q_hdr);
	       break;
	   case 0xCB:      /* EIT */
	       build_eit(q_hdr);
	       break;
	   case 0xCC:      /* ETT */
	       build_ett(q_hdr);
	       break;
	   case 0xCE:      /* DET */
	       break;
	   case 0xCF:      /* DST */
		build_dst_table(q_hdr);
	       break;
	   case 0xD1:      /* NRT */
		build_nrt_table(q_hdr);
	       break;
	   case 0xD2:      /* LTST*/
	       break;
	   default:
	       break;
	}

/* pass verified table to API simulator process */

	pipe_q2api_proc(q_hdr,ts_pid, 0);
	return;

/* discard Table Sect */

discard:
     free( buf_hdr_ptr->buf_ptr );
discard2:
     free( buf_hdr_ptr );
     return;
}

/* --------------------------------------------------------------------- */

/*
 *  Process a full Table Section from the Transport Stream for Data
 *  Tables & Modules (i.e., DII, DSI, DC and Data Modules DDB)
 */

accum_dii_tables(buf_hdr_ptr,ts_control_ptr,ts_pid)
struct   buf_hdr    *buf_hdr_ptr;
struct   ts_control *ts_control_ptr;
unsigned int        ts_pid;
{
    int           i, j, bytes, sect_cnt;
    char          *dest_ptr, *crc_ptr;
    unsigned int dest_len;
    unsigned char c_char;
    unsigned int  tbl_id, tbl_sect_len, tbl_cur_next, tbl_ver_num;
    unsigned int  tbl_sect_num, tbl_last_sect, tbl_syntax_ind, msg_id;
    unsigned int  crc, sect_num, sect_len, ver_num, trans_ver;
    unsigned int  next_ver_num, blk_len, data_len, trans_msg_id, trans_update;
    unsigned int  adapt_len, ddb_offset, new_mod_id, new_mod_ver, mod;
    unsigned int  num_mods, blk_num, num_of_diis;
    struct queue_hdr      *q_hdr;
    struct dii_level      *dii_level;
    struct data_mod       *carousel_mod;
    struct blk_hdr        *blk_hdr;
    struct buf_hdr        *cur_link, *prev_link;
    struct dii_ddb_lst    *dii_ddb_lst_ptr;

	dest_ptr = buf_hdr_ptr->buf_ptr;
	if ( buf_hdr_ptr->buf_ptr == NULL )
	{
      /* no data, discard sect - deallocate buf_hdr */

	     goto discard2;
	}
/* extract basic table header info */

	tbl_id        = (*dest_ptr) & 0xFF;
	tbl_sect_len  = ((dest_ptr[1]&0xF)<<8) | (dest_ptr[2]&0xFF);

// check length (for bad streams)
	dest_len = buf_hdr_ptr->buf_size;
	if ((tbl_sect_len > dest_len) || (tbl_sect_len < 2)) {
		goto discard;
	}	

/*  ***
write_msg('P', 'D', "tbl_sect_len = 0x%x", tbl_sect_len);
*/
	tbl_syntax_ind = (dest_ptr[1]>>6) & 0x01;
	tbl_cur_next   = (dest_ptr[5]&0x1);
	tbl_ver_num    = ((dest_ptr[5]>>1) & 0x1F);
	tbl_sect_num   = (dest_ptr[6] & 0xFF);
	tbl_last_sect  = (dest_ptr[7] & 0xFF);

#ifdef DEEP_DEBUG
	write_msg('P', 'D', " id=%2x, len=%x(%d), cur=%x, ver=%x, sect=%x, last=%x",
	tbl_id, tbl_sect_len, tbl_sect_len, tbl_cur_next, tbl_ver_num,
	tbl_sect_num, tbl_last_sect);

	start_msg("inp(0x%x)=", dest_ptr);
	for (i=0; i<30; i++)
	    append_msg("%2x ", dest_ptr[i]&0xFF);
	end_msg('P', 'D', "");
#endif

 /* extract CRC from end of table section in table buf */

	crc_ptr = &buf_hdr_ptr->buf_ptr[buf_hdr_ptr->byte_loc-4];
	crc = ( (((*crc_ptr)&0xFF)<<24) | ((crc_ptr[1]&0xFF)<<16)
	       | ((crc_ptr[2]&0xFF)<<8) | (crc_ptr[3]&0xFF) );

  /*
   * pass buf and length, not including the CRC field that follows the data,
   * to check CRC or CHKSUM whichever is specified
   */
	if ( tbl_syntax_ind == 0 )
	{          /* CHKSUM */
	    if ( check_chksum(dest_ptr,tbl_sect_len-1,crc) == ERROR )
	    {
#ifdef DEBUG_ERR_MSG
	     write_msg('P', 'D', "error bad CHKSUM in accum_dii_tables()");
#endif
		  goto discard;  /* bad CHKSUM chk, ignore table */
	    }
	}
	else
	{          /* CRC */
	    if ( check_crc(dest_ptr,tbl_sect_len-1,crc) == ERROR )
	    {
#ifdef DEBUG_ERR_MSG
	     write_msg('P', 'D', "error bad CRC in accum_dii_tables()");
#endif
		 goto discard;  /* bad CRC chk, ignore table */
	    }
	}
/* check is good, update number of checked bytes in sect buf */

	buf_hdr_ptr->bytes_checked = buf_hdr_ptr->byte_loc;
	buf_hdr_ptr->valid         = 1;

	dii_ddb_lst_ptr =
	  (struct dii_ddb_lst *) ts_control_ptr->ts_ctrl_gen.ts_ctrl_ptr;
	msg_id = ((dest_ptr[10]&0xFF)<<8) | (dest_ptr[11]&0xFF);
	dii_ddb_lst_ptr->dii_pid = ts_pid;

	if ( tbl_id == 0x3B )
	{       /* DII, DSI or DC */
	    trans_ver = ((dest_ptr[12]&0x3F)<<8) | (dest_ptr[13]&0xFF);
	    trans_msg_id = ((dest_ptr[14]&0xFF)<<7) | ((dest_ptr[15]>>1)&0x7F);
	    trans_update =  (dest_ptr[15]&0x1);

	    if ( msg_id == 0x1002 )
	    {        /* DII */
		if ( trans_msg_id == 0 )
		{    /* single level carousel */
		    if ( dii_ddb_lst_ptr->carousel != NULL )
		    {
			if ( trans_update == 
			    ((dii_ddb_lst_ptr->carousel->transaction_id)&0x1) )
			{      /* same ver */
			     goto discard; /* discard current buf */
			}
		    }
		       /* build single level carousel */
		       /* send tab sect to API, if valid, & deallocate buf */

		    build2_L1_carousel(buf_hdr_ptr,dii_ddb_lst_ptr);
		}
		else
		{    /* two level carousel */
		    if ( dii_ddb_lst_ptr->carousel == NULL )
		    {     /* waiting for 1st level (group) to arrvive */
			  /* place dii on queue */
			enqueue_buf(dii_ddb_lst_ptr->dii_q_hdr,buf_hdr_ptr);
		    }
		    else
		    {   /* 1st level (group) has arrv & carousel is built */
			/* add to 2nd level carousel */
			/* send tab sect to API, if valid, & deallocate buf */

			build2_L2_carousel(buf_hdr_ptr,dii_ddb_lst_ptr);
		    }
		}
	    }
	    else if ( msg_id == 0x1005 )
	    {        /* DC */
		if ( dii_ddb_lst_ptr->dc_buf_hdr != NULL )
		{     /* one already exists */
		    goto discard; /* discard current buf */
		}
		dii_ddb_lst_ptr->dc_buf_hdr = buf_hdr_ptr;

		     /* Carousel Cancelled - delete it */

		delete_carousel( dii_ddb_lst_ptr->carousel );
		clear_queue( dii_ddb_lst_ptr->dii_q_hdr );

		/* pipe grp table sect to API Simul, then save buf */

		pipe_tab2api_proc(buf_hdr_ptr, dii_ddb_lst_ptr->dii_pid, 0);
		dii_ddb_lst_ptr->dc_buf_hdr = buf_hdr_ptr;
	    }
	    else if ( msg_id == 0x1006 )
	    {        /* DSI */
		 if ( dii_ddb_lst_ptr->carousel != NULL )
		 {
		     if ( trans_update == 
                        dii_ddb_lst_ptr->carousel->transaction_id&0x1 )
		     {      /* same ver */
			   goto discard; /* discard current buf */
		     }
		 }
		    /* create group carousel & build 1st of 2 levels */
		    /* send tab sect to API, if valid, & deallocate buf */

		 build2_grp_carousel(buf_hdr_ptr,dii_ddb_lst_ptr);
	    }
	    else
	    {        /* Unknown */
		 goto discard; /* discard current buf */
	    }
	}
	else if ( tbl_id == 0x3C )
	{       /* DDB - Data Blk */
	    if ( dii_ddb_lst_ptr->carousel == NULL )
	    {       /* Carousel doesn't exists yet, ignore */
		 goto discard; /* discard current buf */
	    }
 /* extract module ID of this module block from DDB table header
  */
	    adapt_len = dest_ptr[8+9]&0xFF;
	    ddb_offset = (8 + 12 + adapt_len);
	    new_mod_id = ((dest_ptr[ddb_offset]&0xFF)<<8)
				| (dest_ptr[ddb_offset+1]&0xFF);
	    new_mod_ver = ((dest_ptr[ddb_offset+2]&0xFF)<<8);
 /*
  * search the corresponding carousel for the module ID of this module block
  */
	    dii_level = dii_ddb_lst_ptr->carousel->dii_level_ptr;
	    num_of_diis = dii_ddb_lst_ptr->carousel->num_of_diis;
	    for (j=0; j<num_of_diis; j++, dii_level++)
	    {       /* for a two level carousel, step thru each dii group */
		if ( (num_mods=dii_level->num_of_mods) > 0 )
		{     /* step thru each module hdr of a dii group, */
		      /* allocated contiguously                    */
		    carousel_mod = dii_level->mod_array;
		    mod = -1;
		    for (i=0; i<num_mods; i++, carousel_mod++)
		    {
			if ( carousel_mod->module_id == new_mod_id )
			{
				mod = i;
				goto mod_found;
			}
		    }
		}
	    }

	    goto discard; /* no Mod match found, discard current buf */

mod_found:
 /*
  * extract the block number of this module block from DDB table header
  */
	    blk_num =  ((dest_ptr[ddb_offset+4]&0xFF)<<8)
			| (dest_ptr[ddb_offset+5]&0xFF);
 /*
  * check if expected version & not duplicate data
  */
	    if ( carousel_mod->mod_ver != new_mod_ver )
	    {
#ifdef DEBUG_RCV
		write_msg('P', 'D', "UNEXPECTED blk data version, discarded in accum_dii_tables()");
#endif
		goto discard;
	    }
	    if ( carousel_mod->blk_hdr_ptr[blk_num].valid_flag == 1 )
	    {
#ifdef DEBUG_RCV
		write_msg('P', 'D', "DUPLICATE blk data, discarded in accum_dii_tables()");
#endif
		goto discard;
	    }
 /*
  * compare blk size against payload data size
  */
	    blk_len = dii_level->blk_size;
	    if ( (blk_num+1) == carousel_mod->num_of_blks )
	    {      /* correct for last blk */
		blk_len = carousel_mod->mod_size - (blk_num*blk_len);
	    }
		/* data_payload_len = sect len - hdr_len - crc */
	    data_len = (tbl_sect_len+3) - (ddb_offset+6) - 4;
	    if ( data_len != blk_len )
	    {
#ifdef DEBUG_RCV
		write_msg('P', 'D', "ERROR data_len=%d NOT equal to blk_len=%d in accum_dii_tables()",
			data_len, blk_len);
#endif
		goto discard;
	    }
	    if ( carousel_mod->mod_size == 0 )
	    {
		carousel_mod->buf_size = data_len;
	    }
#ifdef DEBUG_SUMMARY
	write_msg('P', 'D', "");
	write_msg('P', 'D', "DDB - blk_num=0x%x, mod_ver=0x%x, mod_id=0x%x, blk_len=0x%x",
		blk_num, new_mod_ver, new_mod_id, blk_len );
#endif
 /*
  * copy data from sect into buffer
  */
	    bcopy(&dest_ptr[ddb_offset+6],
		  &carousel_mod->buffer[blk_num*dii_level->blk_size], blk_len);
 /*
  * extract PTS, if any
  */
	    if ( (adapt_len > 0) && ((dest_ptr[8+12]&0xFF)==0x04) )
	    {
		carousel_mod->blk_hdr_ptr[blk_num].PTS_flag = 1;
		carousel_mod->blk_hdr_ptr[blk_num].PTS_hi = dest_ptr[8+15]&0x1;
		carousel_mod->blk_hdr_ptr[blk_num].PTS_low =
		   ((dest_ptr[8+16]&0xFF)<<24) | ((dest_ptr[8+17]&0xFF)<<16)
			| ((dest_ptr[8+18]&0xFF)<<8)| (dest_ptr[8+19]&0xFF);
	    }
	    else
	    {  /* no PTS, mark it empty */
		carousel_mod->blk_hdr_ptr[blk_num].PTS_hi = 0x80;   /* arbitrary NULL */
	    }
	    carousel_mod->blk_hdr_ptr[blk_num].valid_flag = 1;

 /* decr "mod_not_done" (mod complete when = 0) & tst if all blks rcvd
  * send a complete module to the API proc
  */
	     if ( (--carousel_mod->mod_not_done) == 0 )
	     {
		 pipe_car2api_proc(dii_ddb_lst_ptr->carousel,ts_pid,carousel_mod);
		 if ( carousel_mod->mod_size == 0 )
		 {     /* special handling for Streaming (module_size==0) */
		     carousel_mod->mod_not_done = 1;
		 }
#ifdef DEBUG_SUMMARY
	write_msg('P', 'D', "");
	write_msg('P', 'D', "DDB - Full Module RCVD, mod_ver=0x%x, mod_id=0x%x, mod_len=0x%x",
		new_mod_ver, new_mod_id, carousel_mod->buf_size );
#endif
	     }
	     goto discard;  /* sect info extracted, discard buf */
	}
	else
	{

#ifdef DEBUG_RCV
	       write_msg('P', 'D', "unknown table_id in accum_dii_tables()-- skip!!");
#endif
	       goto discard;  /* unknown table_id -- skip */
	}
	return;

/* discard Table Sect */

discard:
     free( buf_hdr_ptr->buf_ptr );
discard2:
     free( buf_hdr_ptr );
     return;
}

/* --------------------------------------------------------------------- */

/*
 * delete a addr_sect_lst item & all its linkage
 */

delete_addr_sect_lst_item(addr_sect_lst)
struct  addr_sect_lst *addr_sect_lst;
{
     struct addr_sect_lst *next_addr_sect_ptr, *back_addr_sect_ptr;

     if ( addr_sect_lst != NULL )
     {
	   /* delink lst item from doubly linked list */

	 next_addr_sect_ptr = addr_sect_lst->nxt_addr_sect_ptr;
	 back_addr_sect_ptr = addr_sect_lst->bck_addr_sect_ptr;
	 if ( next_addr_sect_ptr != NULL )
	 {
	    next_addr_sect_ptr->bck_addr_sect_ptr = back_addr_sect_ptr;
	 }
	 if ( back_addr_sect_ptr != NULL )
	 {
	    back_addr_sect_ptr->nxt_addr_sect_ptr = next_addr_sect_ptr;
	 }
	 else
	 {
	    addr_sect_lst_head = next_addr_sect_ptr;
	 }
    }
}

/* --------------------------------------------------------------------- */

/*
 * delete a dii_ddb_lst item & all its linkage
 */

delete_dii_ddb_lst_item(dii_ddb_lst_ptr)
struct  dii_ddb_lst  *dii_ddb_lst_ptr;
{
     struct dii_ddb_lst *next_dii_ddb_lst_ptr, *back_dii_ddb_lst_ptr;

     if ( dii_ddb_lst_ptr != NULL )
     {
	   /* delink lst item from doubly linked list */

	 next_dii_ddb_lst_ptr = dii_ddb_lst_ptr->nxt_dii_ddb_lst;
	 back_dii_ddb_lst_ptr = dii_ddb_lst_ptr->bck_dii_ddb_lst;
	 if ( next_dii_ddb_lst_ptr != NULL )
	 {
	    next_dii_ddb_lst_ptr->bck_dii_ddb_lst = back_dii_ddb_lst_ptr;
	 }
	 if ( back_dii_ddb_lst_ptr != NULL )
	 {
	    back_dii_ddb_lst_ptr->nxt_dii_ddb_lst = next_dii_ddb_lst_ptr;
	 }
	 else
	 {
	    dii_ddb_lst_head = next_dii_ddb_lst_ptr;
	 }

	   /* deallocate & clear all references */

	 if ( dii_ddb_lst_ptr->dc_buf_hdr != NULL )
	 {
	     if ( dii_ddb_lst_ptr->dc_buf_hdr->buf_ptr != NULL )
		free( dii_ddb_lst_ptr->dc_buf_hdr->buf_ptr );
	     free( dii_ddb_lst_ptr->dc_buf_hdr );
	 }
	 if ( dii_ddb_lst_ptr->dsi_buf_hdr != NULL )
	 {
	     if ( dii_ddb_lst_ptr->dsi_buf_hdr->buf_ptr != NULL )
		free( dii_ddb_lst_ptr->dsi_buf_hdr->buf_ptr );
	     free( dii_ddb_lst_ptr->dsi_buf_hdr );
	 }
	 if ( dii_ddb_lst_ptr->dii_q_hdr != NULL )
	 {
	     clear_queue( dii_ddb_lst_ptr->dii_q_hdr );
	     free( dii_ddb_lst_ptr->dii_q_hdr );
	 }
	 delete_carousel( dii_ddb_lst_ptr->carousel );

	free( dii_ddb_lst_ptr ); /* deallocate list item */
    }
}

/* --------------------------------------------------------------------- */

/*
 * create a carousel structure & add to linked list
 */

struct carousel *create_carousel()
{
     struct carousel *carousel_ptr;

  /* allocate a new carousel & add it to the head of the linked
   * carousel list
   */
	carousel_ptr = malloc( sizeof(struct carousel) );
	carousel_ptr->nxt_carousel = carousel_hdr;
	carousel_ptr->bck_carousel = NULL;
	if ( carousel_hdr != NULL )
	    carousel_hdr->bck_carousel = carousel_ptr;
	carousel_hdr = carousel_ptr;
	carousel_ptr->private_len = 0;
	carousel_ptr->private_data = NULL;

	return( carousel_ptr );
}

/* --------------------------------------------------------------------- */

/*
 * delete a carousel structure and delink it 
 */

delete_carousel(carousel_ptr)
struct carousel *carousel_ptr;
{

    if ( carousel_ptr != NULL )
    {
        if ( carousel_ptr->dii_level_ptr != NULL )
            clear_carousel(carousel_ptr);
	if ( carousel_ptr->nxt_carousel != NULL )
        {
            carousel_ptr->nxt_carousel->bck_carousel = 
                                    carousel_ptr->bck_carousel;
        }
	if ( carousel_ptr->bck_carousel != NULL )
        {
            carousel_ptr->bck_carousel->nxt_carousel = 
                                    carousel_ptr->nxt_carousel;
        }
        else
        {
            carousel_hdr = carousel_ptr->nxt_carousel; 
        }
        free( carousel_ptr );
    }
}

/* --------------------------------------------------------------------- */

/*
 * clear a carousel structure, but not its header or its linkage 
 */

clear_carousel(carousel_ptr)
struct carousel *carousel_ptr;
{
    unsigned int   i, j, num_of_diis, num_of_mods;
    struct dii_level *dii_level;
    struct data_mod  *mod_ptr;

    if ( carousel_ptr != NULL )
    {
  /* If a carousel exists, we deallocate all the dynamic data
   * structures & only leave the "carousel" structrue to be re-init
   *
   * We could consider trying to see if the new data module sizes are <=
   * to the existing module sizes and reuse the dynamic data structures
   * but this maybe more time & effort than its worth??
   */
	if ( carousel_ptr->private_data != NULL )
	{
		free( carousel_ptr->private_data );
		carousel_ptr->private_data = NULL;
            carousel_ptr->private_len = 0;           
	}
 /*
  * data structures for all group diis, modules of each dii and
  * blks hdrs of a module are each allocated in a single
  * contiguous memory segment
  */
      num_of_diis = carousel_ptr->num_of_diis;
      dii_level = carousel_ptr->dii_level_ptr;
      for (i=0; i<num_of_diis; i++, dii_level++ )
      {
         num_of_mods = dii_level->num_of_mods;
         if ( (mod_ptr = dii_level->mod_array) != NULL )
         {
             for (j=0; j<num_of_mods; j++, mod_ptr++ )
             {
                 if ( mod_ptr->buffer != NULL )
		        free( mod_ptr->buffer );
                 if ( mod_ptr->blk_hdr_ptr!= NULL )
	              free( mod_ptr->blk_hdr_ptr );
             }
         }
         free( dii_level->mod_array );
      }
      free( carousel_ptr->dii_level_ptr );
    }
}

/* --------------------------------------------------------------------- */

/*
 * Build a group (2 level) carousel entry from this DSI table 
 * sect in the buf_hdr 
 */

build2_grp_carousel(buf_hdr_ptr,dii_ddb_lst_ptr)
struct buf_hdr        *buf_hdr_ptr;
struct dii_ddb_lst    *dii_ddb_lst_ptr;
{       /* allocate struct for a carousel */

    char            *tab_data, *dest_buf;
    unsigned int    i, j, k, adapt_len, msg_len, descr_len, tab_size;
    unsigned int    num_of_diis, num_of_blks, priv_len;
    unsigned int    grp_info_len, order, len, tag, compat_len;
    unsigned int    nxt_trans_id;
    struct carousel  *carousel_ptr;
    struct dii_level *dii_level;

    order = 0;
    if ( dii_ddb_lst_ptr->carousel != NULL )
    {
  /* If a carousel exists, we deallocate all the dynamic data
   * structures & only leave the "carousel" structrue to be re-init
   *
   * We could consider trying to see if the new data module sizes are <=
   * to the existing module sizes and reuse the dynamic data structures
   * but this maybe more time & effort than its worth??
   */
	clear_carousel(carousel_ptr);
    }
    else
    {
  /* allocate a new carousel & add it to the head of the linked
   * carousel list
   */
	dii_ddb_lst_ptr->carousel = create_carousel();
    }
  /* allocate all the dynamic data structures from the "carousel"
   * structrue & init everything
   */
     carousel_ptr = dii_ddb_lst_ptr->carousel;
     carousel_ptr->assoc_tag = dii_ddb_lst_ptr->assoc_tag;
     carousel_ptr->dii_pid   = dii_ddb_lst_ptr->dii_pid;
 /*
  * init string ptr to beginning of table buf
  */
     tab_data = buf_hdr_ptr->buf_ptr;
     carousel_ptr->transaction_id =
	     ((tab_data[8+4]&0xFF)<<24) | ((tab_data[8+5]&0xFF)<<16)
		     | ((tab_data[8+6]&0xFF)<<8) | (tab_data[8+7]&0xFF);
     adapt_len = tab_data[8+9]&0xFF;
     msg_len   = ((tab_data[8+10]&0xFF)<<8) | (tab_data[8+11]&0xFF);
 /*
  * advance string ptr past adapt field & server_id
  */
     tab_data +=  (8 + 12 + adapt_len + 20);

 /*
  * advance string ptr beyond compatability_descriptors
  */
     descr_len   = ((tab_data[0]&0xFF)<<8)| (tab_data[1]&0xFF);
     tab_data +=  (2 + descr_len);
 /*
  * Allocate space for a hdr for each dii in this carousel.
  * The entire space is contiguous (an array), allowing easy 
  * allocation and deallocation. Each hdr points to a dii module
  * group that will contain the module hdrs. Those hdrs are 
  * allocated contiguous as an array, allowing easy allocation 
  * and deallocation.  
  */
     priv_len = ((tab_data[0]&0xFF)<<8)| (tab_data[1]&0xFF);
     if ( priv_len <= 0 )
     {
#ifdef DEBUG_RCV
         write_msg('P', 'D', "no group info in DSI in build2_grp_carsousel()");
#endif
	 goto discard;  /* deallocate table sect in buf */
     }
     tab_data +=  2;  /* adv table to start of GII */

     num_of_diis = ((tab_data[0]&0xFF)<<8)| (tab_data[1]&0xFF);
     carousel_ptr->num_of_diis = num_of_diis;
     carousel_ptr->dii_level_ptr = malloc(sizeof(struct dii_level) 
                                                     * num_of_diis);
#ifdef DEBUG_SUMMARY
     write_msg('P', 'D', "DSI table - Trans_id=0x%x, num_of_grps=0x%x",
	carousel_ptr->transaction_id,         num_of_diis);
#endif
     tab_data += 2;  /* adv table to start of loop */
     dii_level = carousel_ptr->dii_level_ptr;
     for (i=0; i<num_of_diis; i++, dii_level++ )
     {
        dii_level->transaction_id =
	     ((tab_data[0]&0xFF)<<24) | ((tab_data[1]&0xFF)<<16)
		     | ((tab_data[2]&0xFF)<<8) | (tab_data[3]&0xFF);
	  dii_level->grp_size = 
            ((tab_data[4]&0xFF)<<24) | ((tab_data[5]&0xFF)<<16)
			   | ((tab_data[6]&0xFF)<<8)| (tab_data[7]&0xFF);
#ifdef DEBUG_SUMMARY
     write_msg('P', 'D', "    %d. grp_id=0x%x, grp_size=0x%x",
	i+1, dii_level->transaction_id, dii_level->grp_size);
#endif
	 compat_len = ((tab_data[8]&0xFF)<<8)| (tab_data[9]&0xFF);

  /* adv table past GroupCompatibility */

       tab_data +=  (10 + compat_len);
 /*
  * search GroupInfoBytes for GroupLinkDescr
  */
	 grp_info_len = ((tab_data[0]&0xFF)<<8)| (tab_data[1]&0xFF);
	 tab_data += 2;
	 dii_level->pos      = -1;      /* pos */
	 dii_level->blk_size = -1;      /* nxt trans_id */
	 while ( grp_info_len > 0 )
	   {
	       tag = (tab_data[0]&0xFF);
	       len = (tab_data[1]&0xFF);
	     if ( tag == 0xB8 )
	     {     /* for ordering info we temp use other fields */
		 order = 1;
		 dii_level->pos      =(tab_data[2]&0xFF); /* pos */
		 dii_level->blk_size =      /* nxt trans_id - temp use */
		    ((tab_data[3]&0xFF)<<24) | ((tab_data[4]&0xFF)<<16)
			   | ((tab_data[5]&0xFF)<<8) | (tab_data[6]&0xFF);
	     }
	       tab_data += (len + 2);
	       grp_info_len -= (len + 2);
	   }
     }
 /*
  * allocate space for "private" data, if any
  */
     priv_len = ((tab_data[0]&0xFF)<<8)| (tab_data[1]&0xFF);
     carousel_ptr->private_len = priv_len;
     if ( priv_len > 0 )
     {
	   carousel_ptr->private_data = malloc ( priv_len );
	   bcopy(&tab_data[2],carousel_ptr->private_data,priv_len);
     }
     else
     {
	   carousel_ptr->private_data = NULL;
     }

/*
 * check if group ordering/concatenation is specified
 */
     if ( order == 1 )
     {     /* sort flag set, find 1st item in sublist & then follow links  */
	   /* there may be multiple sublist (i.e, multiple starting items) */
	 for (i=0; i<(num_of_diis-2); i++ )
         {
	     dii_level = &carousel_ptr->dii_level_ptr[0];
	     for (j=i; j<(num_of_diis-1); j++, dii_level++ )
	     {     /* search for a starting position (pos==0) */
		 if ( dii_level->pos == 0 )
		 {      /* starter found */
		      break;
		 }
	     }
	     if ( j >= (num_of_diis-1) )
	     {      /* no starter found ==> done */
		 break;
	     }
	     if ( i != j )
	     {     /* place starter at beginning of current sublist, ith pos */
		 swap_grp(&carousel_ptr->dii_level_ptr,i,j);
	     }
	       /* 1st found & placed in "i", now get rest of sublist */
	     for (k=i+1; k<(num_of_diis-2); k++ )
	     {     /* get next dii in  sublist */
		 dii_level = &carousel_ptr->dii_level_ptr[k-1];
		 nxt_trans_id = dii_level->blk_size;
		 dii_level++;
		 for (j=k; j<(num_of_diis-1); j++, dii_level++ )
		 {     /* search remainder of list for next module */
		     if ( nxt_trans_id == dii_level->blk_size )
		     {
			 if ( k != j )
			 {     /* place next module in proper list position */
			     swap_grp(&carousel_ptr->dii_level_ptr,k,j);
			 }
			 break; /* next dii found */
		     }
		 }
		 if ( j >= (num_of_diis-1) )
		 {      /* next not found, terminate current chain ?? */
		     k--;
		     break;
		 }
		 if ( carousel_ptr->dii_level_ptr[k].pos == 2 )
		 {     /* end of current chain (pos==2) */
		     break;
		 }
	     }
	     i = k;
         }
     }
/*
 * complete init of rest of dii hdr 
 */
     dii_level = carousel_ptr->dii_level_ptr;
     for (i=0; i<num_of_diis; i++, dii_level++ )
     {
	     dii_level->blk_size  = 0;
	     dii_level->num_of_mods  = 0;
	     dii_level->private_len  = 0;
	     dii_level->private_data  = NULL;
	     dii_level->mod_array     = NULL;
     }

/* pipe grp table sect to API Simul, then deallocate table */

     pipe_tab2api_proc(buf_hdr_ptr, dii_ddb_lst_ptr->dii_pid, 0);
     free( buf_hdr_ptr->buf_ptr );
     free( buf_hdr_ptr );

/* process any DIIs that were queued & waiting for the group DSI,
 * then deallocate table
 */
     while ( dii_ddb_lst_ptr->dii_q_hdr->q_head_ptr != NULL )
     {
	 buf_hdr_ptr = dequeue_buf(&(dii_ddb_lst_ptr->dii_q_hdr));
	 build2_L2_carousel(buf_hdr_ptr,dii_ddb_lst_ptr);
     }
     return;

discard:
     free( buf_hdr_ptr->buf_ptr );
     free( buf_hdr_ptr );
     return;
}

/* --------------------------------------------------------------------- */

swap_grp(dii_array,pos_0,pos_1)
struct dii_level *dii_array;
int      pos_0;
int      pos_1;
{
    struct dii_level temp_dii;

    bcopy(&dii_array[pos_0],&temp_dii, sizeof(struct dii_level) );
    bcopy(&dii_array[pos_1],&dii_array[pos_0], sizeof(struct dii_level) );
    bcopy(&temp_dii,&dii_array[pos_1], sizeof(struct dii_level) );
}

/* --------------------------------------------------------------------- */

swap_mod(mod_array,pos_0,pos_1)
struct data_mod *mod_array;
int      pos_0;
int      pos_1;
{
    struct data_mod temp_mod;

    bcopy(&mod_array[pos_0],&temp_mod, sizeof(struct data_mod) );
    bcopy(&mod_array[pos_1],&mod_array[pos_0], sizeof(struct data_mod) );
    bcopy(&temp_mod,&mod_array[pos_1], sizeof(struct data_mod) );
}

/* --------------------------------------------------------------------- */

/*
 * Build a single level carousel entry from this DII table 
 * sect in the buf_hdr 
 */

build2_L1_carousel(buf_hdr_ptr,dii_ddb_lst_ptr)
struct buf_hdr        *buf_hdr_ptr;
struct dii_ddb_lst    *dii_ddb_lst_ptr;
{       /* allocate struct for a carousel */

    char             *tab_data, *dest_buf;
    unsigned int     i, j, k, adapt_len, msg_len, descr_len, tab_size;
    unsigned int     blk_size, num_of_mods, num_of_blks, priv_len;
    unsigned int     mod_info_len, order, tag, len;
    unsigned int     nxt_mod_id;
    struct carousel  *carousel_ptr;
    struct dii_level *dii_level;
    struct data_mod  *mod_hdr;

    order = 0;
    if ( dii_ddb_lst_ptr->carousel != NULL )
    {
  /* If a carousel exists, we deallocate all the dynamic data
   * structures & only leave the "carousel" structrue to be re-init
   *
   * We could consider trying to see if the new data module sizes are <=
   * to the existing module sizes and reuse the dynamic data structures
   * but this maybe more time & effort than its worth??
   */
	clear_carousel(carousel_ptr);
    }
    else
    {
  /* allocate a new carousel & add it to the head of the linked
   * carousel list
   */
	dii_ddb_lst_ptr->carousel = create_carousel();
    }
  /* allocate all the dynamic data structures from the "carousel"
   * structrue & init everything
   */
     carousel_ptr = dii_ddb_lst_ptr->carousel;
     carousel_ptr->assoc_tag = dii_ddb_lst_ptr->assoc_tag;
     carousel_ptr->dii_pid   = dii_ddb_lst_ptr->dii_pid;
     carousel_ptr->num_of_diis = 1;
     carousel_ptr->dii_level_ptr = malloc( sizeof(struct dii_level) );
     dii_level = carousel_ptr->dii_level_ptr;
 /*
  * init string ptr to beginning of table buf
  */
     tab_data = buf_hdr_ptr->buf_ptr;
     carousel_ptr->transaction_id =
	     ((tab_data[8+4]&0xFF)<<24) | ((tab_data[8+5]&0xFF)<<16)
		     | ((tab_data[8+6]&0xFF)<<8) | (tab_data[8+7]&0xFF);
     dii_level->transaction_id = carousel_ptr->transaction_id;
     adapt_len = tab_data[8+9]&0xFF;
     msg_len   = ((tab_data[8+10]&0xFF)<<8) | (tab_data[8+11]&0xFF);
 /*
  * advance string ptr past adapt field
  */
     tab_data +=  (8 + 12 + adapt_len);
     carousel_ptr->carousel_id =
	     ((tab_data[0]&0xFF)<<24) | ((tab_data[1]&0xFF)<<16)
		     | ((tab_data[2]&0xFF)<<8)| (tab_data[3]&0xFF);
     blk_size = ((tab_data[4]&0xFF)<<8) | (tab_data[5]&0xFF);
     dii_level->blk_size = blk_size;
 /*
  * advance string ptr beyond compatability_descriptors
  */
     descr_len   = ((tab_data[16]&0xFF)<<8)| (tab_data[17]&0xFF);
     tab_data +=  (20 + descr_len);
 /*
  * Allocate space for a hdr for each module in this carousel.
  * The entire space is contiguous (an array), allowing easy 
  * allocation and deallocation. Each hdr points to a module buffer 
  * that will contain the module data. The data is transmitted as
  * separate block and we allocate a hdr for each blk. Unfortunately
  * we must copy the blk of data into the module buffer, extra
  * handling of the data. Those hdrs are allocated contiguous as 
  * an array, allowing easy allocation and deallocation.  
  */
     num_of_mods = ((tab_data[0]&0xFF)<<8)| (tab_data[1]&0xFF);
     dii_level->num_of_mods = num_of_mods;
     dii_level->mod_array = malloc(sizeof(struct data_mod) * num_of_mods);
     tab_data +=  2;  /* adv table to start of loop */
#ifdef DEBUG_SUMMARY
	 write_msg('P', 'D', "");
     write_msg('P', 'D', "DII table (Level 1) - Trans_id=0x%x, Carousel_id=0x%x, blk_size=0x%x, num_mods=0x%x",
	carousel_ptr->transaction_id, carousel_ptr->carousel_id,
	blk_size, num_of_mods);
#endif
     mod_hdr = dii_level->mod_array;
     for (i=0; i<num_of_mods; i++, mod_hdr++ )
     {
	 mod_hdr->module_id = ((tab_data[0]&0xFF)<<8)| (tab_data[1]&0xFF);
	 mod_hdr->mod_size = ((tab_data[2]&0xFF)<<24) | ((tab_data[3]&0xFF)<<16)
			     | ((tab_data[4]&0xFF)<<8)| (tab_data[5]&0xFF);
	 mod_hdr->mod_ver = tab_data[6]&0xFF;

	 if ( mod_hdr->mod_size == 0 )
	 {     /* special handling for streaming, size=0 */
	       mod_hdr->num_of_blks  = 1;
	       mod_hdr->mod_not_done = 1;
	       mod_hdr->buf_size = 4096;
	 }
	 else
	 {
	       num_of_blks = 1 + (mod_hdr->mod_size-1)/blk_size;
	       mod_hdr->num_of_blks  = num_of_blks;
	       mod_hdr->mod_not_done = num_of_blks;
	       mod_hdr->buf_size = mod_hdr->mod_size;
	 }
#ifdef DEBUG_SUMMARY
     write_msg('P', 'D', "    %d. mod_id=0x%x, mod_size=0x%x, mod_ver=0x%x, num_of_blks=0x%x",
	i+1, mod_hdr->module_id, mod_hdr->mod_size, mod_hdr->mod_ver, num_of_blks);
#endif

 /*
  * allocate the module buffer & a list of blks hdrs, one/blk
  */
	 mod_hdr->buffer = malloc( mod_hdr->buf_size );
	 mod_hdr->blk_hdr_ptr = malloc( sizeof(struct blk_hdr) * mod_hdr->num_of_blks);
 /*
  * search ModuleInfoBytes for ModuleDescr
  */
	 mod_info_len = (tab_data[7]&0xFF);
	 tab_data += 8;
	 mod_hdr->pos = -1;             /* pos */
	 mod_hdr->nxt_module_id = -1;   /* nxt mod_id */
	 while ( mod_info_len > 0 )
	 {
	     tag = (tab_data[0]&0xFF);
	     len = (tab_data[1]&0xFF);
	     if ( tag == 0xB4 )
	     {
		 order = 1;
		 mod_hdr->nxt_module_id = ((tab_data[3]&0xFF)<<8)
			       | (tab_data[4]&0xFF); /* nxt mod_id */
		 mod_hdr->pos =(tab_data[2]&0xFF);   /* pos */
	     }
	     tab_data += (len + 2);
   	     mod_info_len -= (len + 2);   
	 }
     }
 /*
  * allocate space for "private" data, if any
  */
     priv_len = ((tab_data[0]&0xFF)<<8)| (tab_data[1]&0xFF);
     dii_level->private_len = priv_len;
     if ( priv_len > 0 )
     {
	 dii_level->private_data = malloc ( priv_len );
	 bcopy(&tab_data[2],dii_level->private_data,priv_len);
     }
     else
     {
	 dii_level->private_data = NULL;
     }
/*
 * check if module ordering/concatenation is specified
 */
     if ( order == 1 )
     {     /* sort flag set, find 1st item in sublist & then follow links  */
	   /* there may be multiple sublist (i.e, multiple starting items) */
	 for (i=0; i<(num_of_mods-2); i++ )
         {
	     mod_hdr = &dii_level->mod_array[i];
	     for (j=i; j<(num_of_mods-1); j++, mod_hdr++ )
	     {     /* search for a starting position (pos==0) */
		 if ( mod_hdr->pos == 0 )
		 {      /* starter found */
		      break;
		 }
	     }
	     if ( j >= (num_of_mods-1) )
	     {      /* no starter found ==> done */
		 break;
	     }
	     if ( i != j )
	     {     /* place starter at beginning of current sublist, ith pos */
		 swap_mod(&dii_level->mod_array,i,j);
	     }
	       /* 1st found & placed in "i", now get rest of sublist */
	     for (k=i+1; k<(num_of_mods-2); k++ )
	     {     /* get next module in  sublist */
		 mod_hdr = &dii_level->mod_array[k-1];
		 nxt_mod_id = mod_hdr->nxt_module_id;
		 mod_hdr++;
		 for (j=k; j<(num_of_mods-1); j++, mod_hdr++ )
		 {     /* search remainder of list for next module */
		     if ( nxt_mod_id == mod_hdr->module_id )
		     {
			 if ( k != j )
			 {     /* place next module in proper list position */
			     swap_mod(&dii_level->mod_array,k,j);
			 }
			 break; /* next module found */
		     }
		 }
		 if ( j >= (num_of_mods-1) )
		 {      /* next not found, terminate current chain ?? */
		     k--;
		     break;
		 }
		 if ( dii_level->mod_array[k].pos == 2 )
		 {     /* end of current chain */
		     break;
		 }
	     }
	     i = k;
         }
     }

/* send this DII tab sect to the API Simul */

     pipe_tab2api_proc(buf_hdr_ptr, dii_ddb_lst_ptr->dii_pid, 0);

/* deallocate tab sect buf */

     free( buf_hdr_ptr->buf_ptr );
     free( buf_hdr_ptr );
     return;
}
/* --------------------------------------------------------------------- */

/*
 * Add to a group (2 level) carousel from this DII table 
 * sect in the buf_hdr 
 */

build2_L2_carousel(buf_hdr_ptr,dii_ddb_lst_ptr)
struct buf_hdr        *buf_hdr_ptr;
struct dii_ddb_lst    *dii_ddb_lst_ptr;
{       /* allocate struct for a carousel */

    char             *tab_data, *dest_buf;
    unsigned int     i, j, k, adapt_len, msg_len, descr_len, tab_size;
    unsigned int     blk_size, num_of_mods, num_of_blks, priv_len;
    unsigned int     mod_info_len, order, tag, len;
    unsigned int     num_ddi, transaction_id, nxt_mod_id;
    struct carousel  *carousel_ptr;
    struct dii_level *dii_level;
    struct data_mod  *mod_hdr;

    order = 0;
    num_ddi = dii_ddb_lst_ptr->carousel->num_of_diis;
    dii_level = dii_ddb_lst_ptr->carousel->dii_level_ptr;
/*
 * init string ptr to beginning of table buf
 */
    tab_data = buf_hdr_ptr->buf_ptr;
    transaction_id =
	     ((tab_data[8+4]&0xFF)<<24) | ((tab_data[8+5]&0xFF)<<16)
		     | ((tab_data[8+6]&0xFF)<<8) | (tab_data[8+7]&0xFF);

/* search group for proper dii entry */
    j = -1;
    for ( i=0; i<(num_ddi-1); i++, dii_level++ )
    {
         if ( dii_level->transaction_id == transaction_id )
         {
             j = i;
             break;
         }
    }
    if ( j == -1 )
    {
#ifdef DEBUG_RCV
        write_msg('P', 'D', "ERROR dii not found in carousel group in build2_L2_carousel()");
#endif
	goto discard;
    }
     adapt_len = tab_data[8+9]&0xFF;
     msg_len   = ((tab_data[8+10]&0xFF)<<8) | (tab_data[8+11]&0xFF);
 /*
  * advance string ptr past adapt field
  */
     tab_data +=  (8 + 12 + adapt_len);
     blk_size = ((tab_data[4]&0xFF)<<8) | (tab_data[5]&0xFF);
     dii_level->blk_size = blk_size;
 /*
  * advance string ptr beyond compatability_descriptors
  */
     descr_len   = ((tab_data[16]&0xFF)<<8)| (tab_data[17]&0xFF);
     tab_data +=  (20 + descr_len);
 /*
  * Allocate space for a hdr for each module in this carousel.
  * The entire space is contiguous (an array), allowing easy 
  * allocation and deallocation. Each hdr points to a module buffer 
  * that will contain the module data. The data is transmitted as
  * separate block and we allocate a hdr for each blk. Unfortunately
  * we must copy the blk of data into the module buffer, extra
  * handling of the data. Those hdrs are allocated contiguous as 
  * an array, allowing easy allocation and deallocation.  
  */
     num_of_mods = ((tab_data[0]&0xFF)<<8)| (tab_data[1]&0xFF);
     dii_level->num_of_mods = num_of_mods;
     dii_level->mod_array = malloc(sizeof(struct data_mod) * num_of_mods);
#ifdef DEBUG_SUMMARY
 	 write_msg('P', 'D', "");
     write_msg('P', 'D', "DII table (Level 2) - Trans_id=0x%x, Carousel_id=0x%x, blk_size=0x%x, num_mods=0x%x",
	transaction_id, dii_ddb_lst_ptr->carousel->carousel_id,
	blk_size, num_of_mods);
#endif
/*
 * extract info about each module from table sect 
 */
     tab_data +=  2;  /* adv table to start of loop */
     mod_hdr = dii_level->mod_array;
     for (i=0; i<num_of_mods; i++, mod_hdr++ )
     {
	 mod_hdr->module_id = ((tab_data[0]&0xFF)<<8)| (tab_data[1]&0xFF);
	 mod_hdr->mod_size = ((tab_data[2]&0xFF)<<24) | ((tab_data[3]&0xFF)<<16)
			     | ((tab_data[4]&0xFF)<<8)| (tab_data[5]&0xFF);
	 mod_hdr->mod_ver = tab_data[6]&0xFF;

	 if ( mod_hdr->mod_size == 0 )
	 {     /* special handling for streaming, size=0 */
	       mod_hdr->num_of_blks  = 1;
	       mod_hdr->mod_not_done = 1;
	       mod_hdr->buf_size = 4096;
	 }
	 else
	 {
	       num_of_blks = 1 + (mod_hdr->mod_size-1)/blk_size;
	       mod_hdr->num_of_blks  = num_of_blks;
	       mod_hdr->mod_not_done = num_of_blks;
	       mod_hdr->buf_size = mod_hdr->mod_size;
	 }
#ifdef DEBUG_SUMMARY
     write_msg('P', 'D', "    %d. mod_id=0x%x, mod_size=0x%x, mod_ver=0x%x, num_of_blks=0x%x",
	i+1, mod_hdr->module_id, mod_hdr->mod_size, mod_hdr->mod_ver, num_of_blks);
#endif
 /*
  * allocate the module buffer & a list of blks hdrs, one/blk
  */
	 mod_hdr->buffer = malloc( mod_hdr->buf_size );
	 mod_hdr->blk_hdr_ptr = malloc( sizeof(struct blk_hdr) * mod_hdr->num_of_blks);
 /*
  * search ModuleInfoBytes for ModuleDescr
  */
	 mod_info_len = (tab_data[7]&0xFF);
	 tab_data += 8;
	 mod_hdr->nxt_module_id = -1;   /* nxt mod_id */
	 mod_hdr->pos      = -1;        /* pos */
	 while ( mod_info_len > 0 )
	 {
	     tag = (tab_data[0]&0xFF);
	     len = (tab_data[1]&0xFF);
	     if ( tag == 0xB4 )
	     {     /* for ordering info we temp use other fields */
		 order = 1;
		 mod_hdr->nxt_module_id = ((tab_data[3]&0xFF)<<8)
			       | (tab_data[4]&0xFF);        /* nxt mod_id */
		 mod_hdr->pos =(tab_data[2]&0xFF);          /* pos */
	     }
	     tab_data += (len + 2);
   	     mod_info_len -= (len + 2);   
	 }
     }
 /*
  * allocate space for "private" data, if any
  */
     priv_len = ((tab_data[0]&0xFF)<<8)| (tab_data[1]&0xFF);
     dii_level->private_len = priv_len;
     if ( priv_len > 0 )
     {
	 dii_level->private_data = malloc ( priv_len );
	 bcopy(&tab_data[2],dii_level->private_data,priv_len);
     }
     else
     {
	 dii_level->private_data = NULL;
     }
/*
 * check if module ordering/concatenation is specified
 */
     if ( order == 1 )
     {     /* sort flag set, find 1st item in sublist & then follow links  */
	   /* there may be multiple sublist (i.e, multiple starting items) */
	 for (i=0; i<(num_of_mods-2); i++ )
         {
	     mod_hdr = &dii_level->mod_array[i];
	     for (j=i; j<(num_of_mods-1); j++, mod_hdr++ )
	     {     /* search for a starting position (pos==0) */
		 if ( mod_hdr->pos == 0 )
		 {      /* starter found */
		      break;
		 }
	     }
	     if ( j >= (num_of_mods-1) )
	     {      /* no starter found ==> done */
		 break;
	     }
	     if ( i != j )
	     {     /* place starter at beginning of current sublist, ith pos */
		 swap_mod(&dii_level->mod_array,i,j);
	     }
	       /* 1st found & placed in "i", now get rest of sublist */
	     for (k=i+1; k<(num_of_mods-2); k++ )
	     {     /* get next module in  sublist */
		 mod_hdr = &dii_level->mod_array[k-1];
		 nxt_mod_id = mod_hdr->nxt_module_id;
		 mod_hdr++;
		 for (j=k; j<(num_of_mods-1); j++, mod_hdr++ )
		 {     /* search remainder of list for next module */
		     if ( nxt_mod_id == mod_hdr->module_id )
		     {
			 if ( k != j )
			 {     /* place next module in proper list position */
			     swap_mod(&dii_level->mod_array,k,j);
			 }
			 break; /* next module found */
		     }
		 }
		 if ( j >= (num_of_mods-1) )
		 {      /* next not found, terminate current chain ?? */
		     k--;
		     break;
		 }
		 if ( dii_level->mod_array[k].pos == 2 )
		 {     /* end of current chain */
		     break;
		 }
	     }
	     i = k;
         }
     }

/* send this DII tab sect to the API Simul */

     pipe_tab2api_proc(buf_hdr_ptr, dii_ddb_lst_ptr->dii_pid, 0);

/* deallocate tab sect buf */

discard:
     free( buf_hdr_ptr->buf_ptr );
     free( buf_hdr_ptr );
     return;
}

/* --------------------------------------------------------------------- */

/*
 * search DST/NRT list for the specified entry & delete it
 * also delete the queues for the DST/NRT Tables & the associated
 * ptr in the ts_control struct that points to this list item
 */

delete_dst_nrt_q_item( q_lst_ptr )
struct  dst_nrt_q_lst  *q_lst_ptr;
{
       struct  dst_nrt_q_lst  *back_q_lst_ptr, *next_q_lst_ptr;

/* delink item from doubly linked list */

       next_q_lst_ptr = q_lst_ptr->nxt_q_lst_ptr;
       back_q_lst_ptr = q_lst_ptr->bck_q_lst_ptr;
       if ( next_q_lst_ptr != NULL )
       {
          next_q_lst_ptr->bck_q_lst_ptr = back_q_lst_ptr;
       }
       if ( back_q_lst_ptr != NULL )
       {
          back_q_lst_ptr->nxt_q_lst_ptr = next_q_lst_ptr;
       }
       else
       {
	  dst_nrt_q_lst_head = next_q_lst_ptr;
       }

/* delete associated queues & deallocate queue header */
       if ( q_lst_ptr->dst_q_hdr != NULL )
       {
	   clear_queue( q_lst_ptr->dst_q_hdr );
	   free( q_lst_ptr->dst_q_hdr );
       }
       if ( q_lst_ptr->nrt_q_hdr != NULL )
       {
	   clear_queue( q_lst_ptr->nrt_q_hdr );
	   free( q_lst_ptr->nrt_q_hdr );
       }
/* clear ptr to this list */
       ts_control[q_lst_ptr->pid].ts_ctrl_gen.ts_ctrl_ptr = NULL;

/* deallocate this list item */
       free( q_lst_ptr );
       return;
}

/* --------------------------------------------------------------------- */

/*
 * Search the MGT table for a matching "table_type" entry
 * rtn ptr to a MGT table entry
 */

struct mgt_tbl_entry *search_mgt_for_type(value)
unsigned int  value;
{
    int i;
    struct mgt_tbl_entry  *entry_ptr;

    if ( (entry_ptr=current_mgt_tbl.mgt_tbl_ptr) != NULL )
    {
	for ( i=0; i<current_mgt_tbl.mgt_num_entries; i++,entry_ptr++)
	{
	    if ( entry_ptr->table_type == value )
	    {
		return ( entry_ptr );
	    }
	}
    }
    return (NULL );
}

/* --------------------------------------------------------------------- */

/*
 * Delete the existing MGT table and ALL of its associated tables
 */

delete_mgt_table()
{
     int                     i, type_pid;
     struct   mgt_tbl_entry *mgt_tbl_entry_ptr;
     struct   ts_control    *ts_control_ptr;
     struct   queue_hdr     *q_hdr;

/* deallocate existing MGT & all tables it points to
 */
     if ( (mgt_tbl_entry_ptr=current_mgt_tbl.mgt_tbl_ptr) != NULL )
     {
 /* deallocate each MGT table entry
  */
	for ( i=0; i<current_mgt_tbl.mgt_num_entries; i++,mgt_tbl_entry_ptr++)
	{
 /* for each table entry (a announcement table), deallocate its queue
  */
	     if ( (q_hdr=mgt_tbl_entry_ptr->table_type_q_ptr) != NULL )
	     {
		clear_queue( q_hdr );
		free( q_hdr );
	     }
 /* deallocate entry descriptors
  */
	     if ( mgt_tbl_entry_ptr->table_type_descriptor_ptr != NULL )
	     {
		free( mgt_tbl_entry_ptr->table_type_descriptor_ptr );
	     }
 /* clear entry to the "ts_control" struct for this entry
  */
	     type_pid = mgt_tbl_entry_ptr->table_type_pid;
	     if (  (type_pid!=0x1FFB) && (type_pid!=0x1FFF) && (type_pid>1) )
	     {
		 ts_control[type_pid].ts_ctrl_type =  UNDEFINED;
		 if ( ts_control[type_pid].ts_buf_hdr_ptr != NULL )
		 {
		     if ( ts_control[type_pid].ts_buf_hdr_ptr->buf_ptr
								  != NULL )
			 free( ts_control[type_pid].ts_buf_hdr_ptr->buf_ptr );
		     free( ts_control[type_pid].ts_buf_hdr_ptr );
		 }
		 clear_queue( &(ts_control[type_pid].ts_queue_hdr) );
//       can't free this- allocated as a global
//		 free( &(ts_control[type_pid].ts_queue_hdr) );
	     }
	}
 /* deallocate MGT hdr descriptors
  */
	if ( current_mgt_tbl.mgt_descriptor_ptr != NULL )
	{
	   free( current_mgt_tbl.mgt_descriptor_ptr );
	   current_mgt_tbl.mgt_descriptor_ptr = NULL;
	}
 /* deallocate MGT table struct & set the ptr to it in the MGT hdr to NULL
  */
	free( current_mgt_tbl.mgt_tbl_ptr );
	current_mgt_tbl.mgt_tbl_ptr = NULL;
 /*
  * deallocate VCT table struct & clr the corresponding references
  * in the "ts_control" struct, test for both cable & teresterial VCTs
  */
	if ( current_tvct_hdr.vct_tbl_ptr != NULL )
	   clr_vct(&current_tvct_hdr);
	if ( current_cvct_hdr.vct_tbl_ptr != NULL )
	   clr_vct(&current_cvct_hdr);
     }
}

/* --------------------------------------------------------------------- */

/*
 * Expand the byte linear MGT table into a more convienent data struct
 */

build_mgt_table(q_hdr)
struct  queue_hdr  *q_hdr;
{
     int                    i, type, type_pid;
     char                   *buf_ptr;
     unsigned int           num_def, tbl_ver_num;
     struct   mgt_tbl_entry *mgt_tbl_entry_ptr;
     struct   ts_control    *ts_control_ptr;

/* If an MGT already exists, deallocate it 1st
 */
     if ( current_mgt_tbl.mgt_tbl_ptr != NULL )
     {
	 delete_mgt_table();
     }

/*
 * allocate a new MGT table
 */
/*
 * Extract version & number of entries from the table section in table buf
 */
     buf_ptr =  q_hdr->q_head_ptr->buf_ptr;
     tbl_ver_num   = ((buf_ptr[5]>>1) & 0x1F);
     num_def = ((buf_ptr[9] & 0xFF)<<8) | (buf_ptr[10]& 0xFF);
 /*
  * Set values in MGT hdr & allocate space for MGT table struct
  */
     current_mgt_tbl.mgt_version = tbl_ver_num;
     current_mgt_tbl.mgt_num_entries = num_def;
     current_mgt_tbl.mgt_tbl_ptr = malloc (num_def * sizeof(struct mgt_tbl_entry));
     mgt_tbl_entry_ptr = current_mgt_tbl.mgt_tbl_ptr;
 /*
  * advance ptr to table buf past the table header
  */
     buf_ptr += MGT_TBL_HDR_SIZE;
 /*
  * allocate buf_hdrs & bufs for each table entry in the MGT table struct
  */
     for ( i=0; i<num_def; i++,mgt_tbl_entry_ptr++)
     {
 /* extract info about this entry & place it into the MGT table struct
  */
	  mgt_tbl_entry_ptr->table_type =
			       ((buf_ptr[0] & 0xFF)<<8) | (buf_ptr[1]& 0xFF);
	  mgt_tbl_entry_ptr->table_type_pid =
			       ((buf_ptr[2] & 0x1F)<<8) | (buf_ptr[3]& 0xFF);
	  mgt_tbl_entry_ptr->table_type_version_number = (buf_ptr[4] & 0x1F);
	  mgt_tbl_entry_ptr->number_bytes =
			((buf_ptr[5] & 0xFF)<<24) | ((buf_ptr[6] & 0xFF)<<16)
		       | ((buf_ptr[7] & 0xFF)<<8) | (buf_ptr[8] & 0xFF);
	  mgt_tbl_entry_ptr->table_type_descriptor_length =
				((buf_ptr[9] & 0xF)<<8) | (buf_ptr[10]& 0xFF);
/*
 * advance ptr to table buf past item info & allocate a queue
 * for this table entry
 */
	  buf_ptr += MGT_ITEM_SIZE;
	  mgt_tbl_entry_ptr->table_type_q_ptr =
					   malloc( sizeof(struct queue_hdr) );
	  mgt_tbl_entry_ptr->table_type_q_ptr->q_head_ptr= NULL;
	  mgt_tbl_entry_ptr->table_type_q_ptr->q_tail_ptr= NULL;
	  mgt_tbl_entry_ptr->table_type_q_ptr->q_valid   = 0;
	  mgt_tbl_entry_ptr->table_type_q_ptr->q_tot_bytes = 0;
	  mgt_tbl_entry_ptr->table_type_q_ptr->q_ver =
			mgt_tbl_entry_ptr->table_type_version_number;
 /*
  * insert entry into the "ts_control" struct
  */
	 type_pid = mgt_tbl_entry_ptr->table_type_pid;
	 if ( (type_pid!=0x1FFB) && (type_pid!=0x1FFF) && (type_pid>1) )
	 {
	     ts_control_ptr = &ts_control[type_pid];
	     ts_control_ptr->ts_ctrl_type   =
		 (  ((type=mgt_tbl_entry_ptr->table_type)==0x0004) ? ETT :
			( ((type&0xFF80)==0x0100) ? EIT :
			( ((type&0xFF80)==0x0200) ? ETT :  /* for EITs */
			( ((type&0xFF80)==0x1300) ? DET :
			( ((type&0xFF80)==0x1200) ? ETT :  /* for DETs */
			( ((type       )==0x1180) ? LTST:  /* LTST */
						  UNDEFINED))))) );
	     ts_control_ptr->ts_ctrl_status = ACTIVE;
	     ts_control_ptr->ts_buf_hdr_ptr = NULL;
	     ts_control_ptr->ts_ctrl_gen.ts_ctrl_ptr  = (unsigned char *) mgt_tbl_entry_ptr;
	 }
/*
 * allocate space for any descriptors with this entry & copy the descr contents
 */
	  if ( mgt_tbl_entry_ptr->table_type_descriptor_length > 0 )
	  {
	      mgt_tbl_entry_ptr->table_type_descriptor_ptr =
		   malloc( mgt_tbl_entry_ptr->table_type_descriptor_length );
	      bcopy(buf_ptr,mgt_tbl_entry_ptr->table_type_descriptor_ptr,
			   mgt_tbl_entry_ptr->table_type_descriptor_length );
/*
 * advance ptr to table buf past any descriptors
 */
	      buf_ptr += mgt_tbl_entry_ptr->table_type_descriptor_length;
	  }
	  else
	  {
	      mgt_tbl_entry_ptr->table_type_descriptor_ptr = NULL;
	  }
     }
/*
 * allocate space for any descriptors at the MGT table level (vs. per entry)
 * & copy the descr contents
 */
     if ( (current_mgt_tbl.descriptor_length =
		      (((buf_ptr[0] & 0xF)<<8) | (buf_ptr[1]& 0xFF)) ) > 0 )
     {
	 current_mgt_tbl.mgt_descriptor_ptr =
				 malloc( current_mgt_tbl.descriptor_length );
	 bcopy(&buf_ptr[2],current_mgt_tbl.mgt_descriptor_ptr,
					  current_mgt_tbl.descriptor_length);
     }
     else
     {
	 current_mgt_tbl.mgt_descriptor_ptr = NULL;
     }
#ifdef DEBUG_SUMMARY
	write_msg('P', 'D', "");
	start_msg("MGT table -");
	append_msg(" ver=0x%x", current_mgt_tbl.mgt_version);
	append_msg(" items=0x%x", current_mgt_tbl.mgt_num_entries);
	end_msg('P', 'D', " descr_len=0x%x", current_mgt_tbl.descriptor_length);
	prt_all_descr(current_mgt_tbl.mgt_descriptor_ptr,
				current_mgt_tbl.descriptor_length);
	mgt_tbl_entry_ptr = current_mgt_tbl.mgt_tbl_ptr;
	for (i=0; i<current_mgt_tbl.mgt_num_entries; i++, mgt_tbl_entry_ptr++)
	{
	   start_msg(" %2d ", i);
	   append_msg(" typ=0x%x", mgt_tbl_entry_ptr->table_type);
	   append_msg(" pid=0x%x", mgt_tbl_entry_ptr->table_type_pid);
	   append_msg(" ver=0x%x", mgt_tbl_entry_ptr->table_type_version_number);
	   append_msg(" size=0x%x", mgt_tbl_entry_ptr->number_bytes);
	   end_msg('P', 'D', " descr_len=0x%x", mgt_tbl_entry_ptr->table_type_descriptor_length);
	   prt_all_descr(mgt_tbl_entry_ptr->table_type_descriptor_ptr,
			    mgt_tbl_entry_ptr->table_type_descriptor_length);
	}
#endif
}

/* ---------------------------------------------------------------- ---- */

/*
 * Expand the byte linear STT table into a more convienent data struct
 */

build_stt_table(q_hdr)
struct  queue_hdr  *q_hdr;
{
   int           i;
   char          *buf_ptr;
   unsigned int  sect_len;

/* GPS sec since 12:00 am 6 Jan 1980  +/- 4 sec accuracy */

   buf_ptr = q_hdr->q_head_ptr->buf_ptr;
/*
 * extract table info from the table buffer & place it in the STT struct
 */
   stt_data.system_time      =
		       ((buf_ptr[9] & 0xFF)<<24) | ((buf_ptr[10] & 0xFF)<<16)
		     | ((buf_ptr[11] & 0xFF)<<8) | (buf_ptr[12] & 0xFF);
   stt_data.gps_utc_offset   = (buf_ptr[13] & 0xFF);
   stt_data.ds_status        = (buf_ptr[14] & 0x01);
   stt_data.ds_day_of_month  = ((buf_ptr[14]>>3) & 0x1F);
   stt_data.ds_hour          = (buf_ptr[15] & 0xFF);
/*
 * get current system time that corresponds to rcvd time
 */
   gettimeofday(&time_val_struct, NULL );
/*
 * free any previous descriptors
 */
   if ( stt_data.descriptor_ptr != NULL )
   {
       free( stt_data.descriptor_ptr );
   }
/*
 * allocate & copy any new descriptors
 */
   sect_len = ((buf_ptr[1]&0xF)<<8) | (buf_ptr[2]&0xFF);
   if ( (stt_data.descriptor_length = sect_len-13-4) > 0 )
   {                          /* assuming 4 bytes for CRC */
       buf_ptr += (13+3);
       stt_data.descriptor_ptr = malloc(stt_data.descriptor_length);
       bcopy(buf_ptr, stt_data.descriptor_ptr, stt_data.descriptor_length);
   }
#ifdef DEBUG_SUMMARY

   start_msg(" STT table - time=%x", stt_data.system_time      );
/*
   append_msg(" offset=0x%x", stt_data.gps_utc_offset   );
   append_msg(" stat=0x%x", stt_data.ds_status        );
   append_msg(" day=0x%x", stt_data.ds_day_of_month  );
   append_msg(" hr=0x%x", stt_data.ds_hour          );
   end_msg('P', 'D', " d_len=0x%x", stt_data.descriptor_length);
   start_msg("");
*/
   end_msg('P', 'D', " system_time = %d", time_val_struct.tv_sec);
   prt_all_descr(stt_data.descriptor_ptr, stt_data.descriptor_length);
#endif
}

/* --------------------------------------------------------------------- */

/*
 * Expand the byte linear TVCT or CVCT table into a more convienent data struct
 * the only difference between the 2 is that the CVCT has 2 extra 1-bit
 * fields (they are just unused bits in the TVCT); also there may
 * be different descriptor -- such as the mandatory "Service Location
 * Descriptor" for the TVCT, but that shouldn't pose a problem here
 * since we don't process descriptors except for the "Service Location
 * Descriptor", if its there.
 */

extract_vct_stream_info(q_hdr,current_vct_hdr)
struct  queue_hdr  *q_hdr;
struct  vct_hdr  *current_vct_hdr;
{
    int          i, j, k, entry_cnt, byte_cnt, last_sect_num, ch_cnt;
    int          descr_length, descr_bytes, num_elts,type_stream_id, type_pid;
    char         *buf_ptr, *descr_ptr;
    unsigned int tbl_sect_len, addl_descr_len;
    struct       vct_item       *vct_item_ptr;
    struct       elt_descr      *elt_ptr;
    struct       ts_control     *ts_control_ptr;
    struct       buf_hdr        *buf_hdr_ptr, *cur_link;
    struct       dst_nrt_q_lst  *q_lst_ptr;
    struct       dii_ddb_lst    *dii_ddb_lst;
    struct       addr_sect_lst  *addr_sect_lst;

/*
 * The appropriate VCT hdr (either TVCT or CVCT) has been passed to
 * this fct, so it can generically use "current_vct_hdr"
 */

 /*
  * deallocate VCT table struct, if it exists, & clr the corresponding
  * references in the "ts_control" struct
  */
    if ( current_vct_hdr->vct_tbl_ptr != NULL )
       clr_vct(&current_vct_hdr);
/*
 * set ptr to start of table buffer
 */
    buf_hdr_ptr = q_hdr->q_head_ptr;
    buf_ptr = buf_hdr_ptr->buf_ptr;
/*
#ifdef DEBUG_SUMMARY
    start_msg("VCT(raw)=");
    for (i=0; i<(buf_hdr_ptr->bytes_checked); i++)
    {
	append_msg(" %2x", buf_ptr[i]&0xFF);
	if ( (i%23) == 22 ) {
		end_msg('P', 'D', "");	
	    start_msg("         ");
	    }
    }
    end_msg('P', 'D', "");
#endif
*/
/*
 * extract basic table info & place it in the VCT hdr
 */
    current_vct_hdr->transport_stream_id = ((buf_ptr[3]&0xFF)<<8) | (buf_ptr[4]&0xFF);
    current_vct_hdr->version_number = ((buf_ptr[5]>>1) & 0x1F);
/*
 * step thru each table section (in table buffer) to count the
 * total number of VCT entries across the entire table
 */
    entry_cnt = 0;
    cur_link = buf_hdr_ptr;
    while ( cur_link != NULL )
    {
	buf_ptr = cur_link->buf_ptr;
	entry_cnt += (buf_ptr[9]&0xFF);
	cur_link = cur_link->q_ptr;
    }
/*
 * allocate the required space for the VCT struct
 */
    current_vct_hdr->vct_tbl_ptr = malloc( sizeof(struct vct_item) * entry_cnt );
    vct_item_ptr = current_vct_hdr->vct_tbl_ptr;
    current_vct_hdr->num_entries = entry_cnt;
/*
 * step thru each table section (in table buffer), extracting info
 * on each item
 */
    byte_cnt = 0;
    entry_cnt = 0;
    cur_link = buf_hdr_ptr;
    while ( cur_link != NULL )
    {
	buf_ptr = cur_link->buf_ptr;
	tbl_sect_len  = ((buf_ptr[1]&0xF)<<8) | (buf_ptr[2]&0xFF);
	byte_cnt += (tbl_sect_len + 3);
	entry_cnt += (buf_ptr[9]&0xFF);
	buf_ptr += VCT_HDR_OFFSET;
	ch_cnt = 0;
/*
 * step thru each item in this table section (in table buffer),
 * extracting info on each item & place it in the VCT struct
 */
	for ( j=0; j<entry_cnt; j++, vct_item_ptr++)
	{
	    for ( k=0; k<7 ; k++)
	    {
		vct_item_ptr->short_name[k] =
				  ((buf_ptr[0]&0xFF)<<8) | (buf_ptr[1]&0xFF);

		buf_ptr += 2;
	    }
	    vct_item_ptr->major_channel_number =
			      ((buf_ptr[0]&0xF)<<6) | ((buf_ptr[1]>>2)&0x3F);
	    vct_item_ptr->minor_channel_number =
				  ((buf_ptr[1]&0x03)<<8) | (buf_ptr[2]&0xFF);
	    vct_item_ptr->modulation_mode      = (buf_ptr[3]&0xFF);
	    vct_item_ptr->carrier_frequency    =
			  ((buf_ptr[4]&0xFF)<<24) | ((buf_ptr[5]&0xFF)<<16)
				| ((buf_ptr[6]&0xFF)<<8) | (buf_ptr[7]&0xFF);
	    vct_item_ptr->channel_TSID         =
				  ((buf_ptr[8]&0xFF)<<8) | (buf_ptr[9]&0xFF);
	    vct_item_ptr->program_number       =
				((buf_ptr[10]&0xFF)<<8) | (buf_ptr[11]&0xFF);
	    vct_item_ptr->etm_location         = ((buf_ptr[12]>>6)&0x03);
	    vct_item_ptr->access_controlled    = ((buf_ptr[12]>>5)&0x01);
	    vct_item_ptr->hidden               = ((buf_ptr[12]>>4)&0x01);
  /* these next 2 are only applicable to cable */
  /* the bits are there, just not used */
	    vct_item_ptr->path_select          = ((buf_ptr[12]>>3)&0x01);
	    vct_item_ptr->out_of_band          = ((buf_ptr[12]>>2)&0x01);

	    vct_item_ptr->service_type         = (buf_ptr[13]&0x3F);
	    vct_item_ptr->source_id            =
				((buf_ptr[14]&0xFF)<<8) | (buf_ptr[15]&0xFF);
/*
 * get length of descriptors for this item
 */
	    descr_length                        =
				((buf_ptr[16]&0x03)<<8) | (buf_ptr[17]&0xFF);
/*
 * advance ptr to table & associated byte count past item info to descriptors
 */
	    buf_ptr += VCT_ITEM_SIZE;
	    ch_cnt += VCT_ITEM_SIZE;
	    vct_item_ptr->num_ch_elts = 0;
	    if ( (vct_item_ptr->descr_length=descr_length) == 0 )
	    {
		vct_item_ptr->descr_ptr = NULL;
	    }
	    else
	    {
		vct_item_ptr->descr_ptr = buf_ptr;
	    }
/*
 * step thru each descriptor for this item searching for "Service Location
 * Descriptor" for the TVCT, if its there expand it into the VCT struct
 */
	    descr_bytes = 0;
	    descr_ptr = buf_ptr;
	    while ( descr_bytes < descr_length )
	    {
		if ( ((*descr_ptr)&0xFF) == 0xA1 )
		{
 /* "Service Location Descriptor", extract common info
  */
		    vct_item_ptr->pcr_pid =
			      ((descr_ptr[2]&0x1F)<<8) | (descr_ptr[3]&0xFF);
		    num_elts = descr_ptr[4] & 0xFF;
		    if (  num_elts > 10 )
		    {
			write_msg('P', 'E', "num of pes_pids = %d -- truncated to 10",
								   num_elts);
			num_elts =10;
		     }
/* step thru each channel element & extract info into the VCT struct
 */
		     vct_item_ptr->num_ch_elts = num_elts;
		     for ( k=0; k<num_elts; k++, elt_ptr++ )
		     {
			 type_stream_id = vct_item_ptr->elts[k].stream_id  =
						(descr_ptr[k*6+5]&0xFF);
			 type_pid = vct_item_ptr->elts[k].pes_pid =
					   ((descr_ptr[k*6+6]&0x1F)<<8)
					      | (descr_ptr[k*6+7]&0xFF);
			 vct_item_ptr->elts[k].code_639[0] =
						(descr_ptr[k*6+8]&0xFF);
			 vct_item_ptr->elts[k].code_639[1] =
						(descr_ptr[k*6+9]&0xFF);
			 vct_item_ptr->elts[k].code_639[2] =
						(descr_ptr[k*6+10]&0xFF);
			 if ( (type_pid!=0x1FFB) && (type_pid!=0x1FFF) && (type_pid>1) )
			 {
			     ts_control_ptr = &ts_control[type_pid];
			     ts_control_ptr->ts_orig_stream = type_stream_id;
			     if ( ts_control_ptr->ts_ctrl_type == UNDEFINED )
				 ts_control_ptr->ts_cont_cntr   = 0xFF;
			     switch ( type_stream_id )
			     {
				case 0x02:
				   if ( ts_control_ptr->ts_ctrl_type == VIDEO )
				   {  /* already defined by PMT */
					break;
				   }
				   ts_control_ptr->ts_ctrl_type = VIDEO;
				   ts_control_ptr->ts_orig_type   = VIDEO;
				   ts_control_ptr->ts_ctrl_status = INACTIVE;
				   ts_control_ptr->ts_buf_hdr_ptr = NULL;
				   ts_control_ptr->ts_ctrl_gen.ts_ctrl_fd = -1;
				   clear_queue( &(ts_control_ptr->ts_queue_hdr) );
				   if ( first_video_stream == 1 )
				   {
				      if ( output_video_stream != -1 )
				      {
					  first_video_stream = 0;
					  ts_control_ptr->ts_ctrl_status = READY;
					  ts_control_ptr->ts_ctrl_gen.ts_ctrl_fd =
								output_video_stream;
					  open_pid_output_dev(type_pid,output_video_dev_ptr);
				      }
				   }
				   break;

				case 0x06:
				   if ( ts_control_ptr->ts_ctrl_type == PES_IZED )
				   {  /* already defined by PMT */
					break;
				   }
				   ts_control_ptr->ts_ctrl_type = PES_IZED;
				   ts_control_ptr->ts_orig_type = PES_IZED;
				   ts_control_ptr->ts_ctrl_status = ACTIVE;
				   ts_control_ptr->ts_buf_hdr_ptr = NULL;
				   ts_control_ptr->ts_ctrl_gen.ts_ctrl_ptr  =
						      (unsigned char *) NULL;
				   clear_queue( &(ts_control_ptr->ts_queue_hdr) );
				   break;

				case 0x0D:
				   if ( ts_control_ptr->ts_ctrl_type == ADDR_SECT )
				   {  /* already defined by PMT */
					break;
				   }
				   ts_control_ptr->ts_ctrl_type = ADDR_SECT;
				   ts_control_ptr->ts_orig_type = ADDR_SECT;
				   ts_control_ptr->ts_ctrl_status = ACTIVE;
				   ts_control_ptr->ts_buf_hdr_ptr = NULL;

			      /* create a list item for this addr_sect entry */
			      /* and add it to the linked list               */

				   addr_sect_lst =  malloc( sizeof(struct addr_sect_lst) );
				   addr_sect_lst->nxt_addr_sect_ptr = addr_sect_lst_head;
				   addr_sect_lst->bck_addr_sect_ptr = NULL;
				   if ( addr_sect_lst_head != NULL )
				   {
				       addr_sect_lst_head->bck_addr_sect_ptr = addr_sect_lst;
				   }
				   addr_sect_lst_head = addr_sect_lst;
				   ts_control_ptr->ts_ctrl_gen.ts_ctrl_ptr =
						(unsigned char *) addr_sect_lst;
				   addr_sect_lst->pid = type_pid;
				   addr_sect_lst->assoc_tag = 0;
				   clear_queue( &(ts_control_ptr->ts_queue_hdr) );
				   break;

				case 0x0B:
				case 0x14:
				   if ( ts_control_ptr->ts_ctrl_type == DII_DDB )
				   {  /* already defined by PMT */
					break;
				   }
				   ts_control_ptr->ts_ctrl_type = DII_DDB;
				   ts_control_ptr->ts_orig_type = DII_DDB;
				   ts_control_ptr->ts_ctrl_status = ACTIVE;
				   ts_control_ptr->ts_buf_hdr_ptr = NULL;

			      /* create a list item for this set of tables */
			      /* and add it to the linked list              */

				   dii_ddb_lst = malloc( sizeof(struct dii_ddb_lst) );
				   dii_ddb_lst->nxt_dii_ddb_lst = dii_ddb_lst_head;
				   dii_ddb_lst->bck_dii_ddb_lst = NULL;
				   if ( dii_ddb_lst_head != NULL )
				   {
				       dii_ddb_lst_head->bck_dii_ddb_lst = dii_ddb_lst;
				   }
				   dii_ddb_lst_head = dii_ddb_lst;
				   ts_control_ptr->ts_ctrl_gen.ts_ctrl_ptr =
						(unsigned char *) dii_ddb_lst;

				   dii_ddb_lst->dii_pid = type_pid;
				   dii_ddb_lst->assoc_tag = 0;
				   dii_ddb_lst->dii_q_hdr   = create_queue();
				   dii_ddb_lst->dsi_buf_hdr = NULL;
				   dii_ddb_lst->dc_buf_hdr  = NULL;
				   dii_ddb_lst->carousel    = NULL;

				   clear_queue( &(ts_control_ptr->ts_queue_hdr) );
				   break;

				case 0x81:
				   if ( ts_control_ptr->ts_ctrl_type == AUDIO_AC3 )
				   {  /* already defined by PMT */
					break;
				   }
				   ts_control_ptr->ts_ctrl_type = AUDIO_AC3;
				   ts_control_ptr->ts_orig_type = AUDIO_AC3;
				   ts_control_ptr->ts_ctrl_status = INACTIVE;
				   ts_control_ptr->ts_buf_hdr_ptr = NULL;
				   ts_control_ptr->ts_ctrl_gen.ts_ctrl_fd = -1;
				   if ( first_audio_stream == 1 )
				   {
				      if ( output_audio_stream != -1 )
				      {
					  first_audio_stream = 0;
					  ts_control_ptr->ts_ctrl_status = READY;
					  ts_control_ptr->ts_ctrl_gen.ts_ctrl_fd =
								output_audio_stream;
					  open_pid_output_dev(type_pid,output_audio_dev_ptr);
				      }
				   }
				   break;

				case 0x95:
				   if ( ts_control_ptr->ts_ctrl_type == DST_NRT )
				   {  /* already defined by PMT */
					break;
				   }
				   ts_control_ptr->ts_ctrl_type = DST_NRT;
				   ts_control_ptr->ts_orig_type = DST_NRT;
				   ts_control_ptr->ts_ctrl_status = ACTIVE;
				   ts_control_ptr->ts_buf_hdr_ptr = NULL;
				   clear_queue( &(ts_control_ptr->ts_queue_hdr) );

			      /* create a list item for this pair of tables */
			      /* and add it to the linked list              */

				   q_lst_ptr = malloc( sizeof(struct dst_nrt_q_lst) );
				   q_lst_ptr->nxt_q_lst_ptr = dst_nrt_q_lst_head;
				   q_lst_ptr->bck_q_lst_ptr = NULL;
				   if ( dst_nrt_q_lst_head != NULL )
				   {
				       dst_nrt_q_lst_head->bck_q_lst_ptr
					   = q_lst_ptr;
				   }
				   dst_nrt_q_lst_head = q_lst_ptr;
				   ts_control_ptr->ts_ctrl_gen.ts_ctrl_ptr =
						(unsigned char *) q_lst_ptr;
				   q_lst_ptr->pid = type_pid;
				       /* allocate & init DST queue */
				   q_lst_ptr->dst_q_hdr = create_queue();
				       /* allocate & init NRT queue */
				   q_lst_ptr->nrt_q_hdr = create_queue();
		/* need to keep a list of new & old DST/NRTs to know when */
		/* to delete previous DST/NRTs that are no longer used */
				   break;

				case 0xC2:
				   if ( ts_control_ptr->ts_ctrl_type == PES_OUS )
				   {  /* already defined by PMT */
					break;
				   }
				   ts_control_ptr->ts_ctrl_type = PES_OUS;
				   ts_control_ptr->ts_orig_type = PES_OUS;
				   ts_control_ptr->ts_ctrl_status = ACTIVE;
				   ts_control_ptr->ts_buf_hdr_ptr = NULL;
				   ts_control_ptr->ts_ctrl_gen.ts_ctrl_ptr  =
						      (unsigned char *) NULL;
				   clear_queue( &(ts_control_ptr->ts_queue_hdr) );
				   break;

				default:
				   if ( type_stream_id >= 0xC0 )
				   {       /* User Private -- assume data piping */
				       ts_control_ptr->ts_ctrl_type = PIPING_COOKED;
				       ts_control_ptr->ts_orig_type = PIPING_COOKED;
				   }
				   else
				   {
				       ts_control_ptr->ts_ctrl_type = UNDEFINED;
				       ts_control_ptr->ts_orig_type = UNDEFINED;
				   }
				   ts_control_ptr->ts_buf_hdr_ptr = NULL;
				   clear_queue( &(ts_control_ptr->ts_queue_hdr) );
				   ts_control_ptr->ts_ctrl_gen.ts_ctrl_ptr = NULL;
				   break;
			     }
			 }
		     }
		}
		descr_bytes += (descr_ptr[1]&0xFF) + 2;   /* descr_len */
		descr_ptr   += (descr_ptr[1]&0xFF) + 2;   /* descr_len */
	    }
/*
 * advance ptr to table past descriptors
 */
	    buf_ptr += descr_length;
	}
/*
 * advance ptr past additional descriptors for this table section
 * (we ignore them here)
 */
	addl_descr_len = ((buf_ptr[0]&0x03)<<8) | (buf_ptr[1]&0xFF);
	if ( (current_vct_hdr->descr_length=addl_descr_len) == 0 )
	{
	    current_vct_hdr->descr_ptr = NULL;
	}
	else
	{
	    current_vct_hdr->descr_ptr = buf_ptr;
	}
/*        buf_ptr += (addl_descr_len+2+4);  /* 2 addl_descr_len & 4 crc */

	cur_link = cur_link->q_ptr;
    }
#ifdef DEBUG_SUMMARY
	start_msg("VCT table -");
	append_msg(" ver=0x%x", current_vct_hdr->version_number);
	append_msg(" items=0x%x", current_vct_hdr->num_entries);
	end_msg('P', 'D', " ts_stream_id=0x%x", current_vct_hdr->transport_stream_id);
	prt_all_descr(current_vct_hdr->descr_ptr,
					   current_vct_hdr->descr_length);
	vct_item_ptr = current_vct_hdr->vct_tbl_ptr;
	for (i=0; i<current_vct_hdr->num_entries; i++, vct_item_ptr++)
	{
	    start_msg(" %2d. name= ", i);
	    for (j=0; j<7; j++ )
	    {
	       append_msg("%c", vct_item_ptr->short_name[j] );
	    }
	    append_msg(" maj=0x%x",  vct_item_ptr->major_channel_number);
	    append_msg(" min=0x%x",  vct_item_ptr->minor_channel_number);
	    append_msg(" mod=0x%x",  vct_item_ptr->modulation_mode     );
	    append_msg(" freq=0x%x", vct_item_ptr->carrier_frequency   );
	    append_msg(" TSid=0x%x", vct_item_ptr->channel_TSID        );
	    append_msg(" prog=0x%x", vct_item_ptr->program_number      );
	    append_msg(" etm=0x%x",  vct_item_ptr->etm_location        );
	    append_msg(" acc=0x%x",  vct_item_ptr->access_controlled   );
	    append_msg(" hide=0x%x", vct_item_ptr->hidden              );
	    append_msg(" serv=0x%x", vct_item_ptr->service_type        );
	    append_msg(" s_id=0x%x", vct_item_ptr->source_id           );
	    append_msg(" pcr=0x%x",  vct_item_ptr->pcr_pid             );
	    parser_end msg('E', " descr_len=0x%x", vct_item_ptr->descr_length );
	    for ( k=0; k<vct_item_ptr->num_ch_elts; k++, elt_ptr++ )
	    {
		start_msg("         %2d.", k);
		append_msg(" stream=0x%x",vct_item_ptr->elts[k].stream_id  );
		append_msg(" pid=0x%x",vct_item_ptr->elts[k].pes_pid    );
		append_msg(" lang_code=%c",vct_item_ptr->elts[k].code_639[0]);
		append_msg("%c",vct_item_ptr->elts[k].code_639[1]);
		end_msg('P', 'D', "%c",vct_item_ptr->elts[k].code_639[2]);
	    }
	    prt_all_descr(vct_item_ptr->descr_ptr,
						vct_item_ptr->descr_length);
	}
#endif
}

/* --------------------------------------------------------------------- */

/*
 * deallocate VCT table struct & clr the corresponding
 * references in the "ts_control" struct
 */

clr_vct(current_vct_hdr)
struct  vct_hdr  *current_vct_hdr;
{
    int                    i, k, entry_cnt;
    int                    num_elts,type_pid;
    struct   vct_item      *vct_item_ptr;

/*
 * step thru each entry in the VCT & if there are any stream_type items
 * clr their ref in the "ts_control" struct
 */
    entry_cnt = current_vct_hdr->num_entries;
    vct_item_ptr = current_vct_hdr->vct_tbl_ptr;
    for ( i=0; i<entry_cnt; i++, vct_item_ptr++ )
    {
	num_elts = vct_item_ptr->num_ch_elts;
	for ( k=0; k<num_elts; k++ )
	{
	     type_pid = vct_item_ptr->elts[k].pes_pid;
	     if ( (type_pid!=0x1FFB) && (type_pid!=0x1FFF) && (type_pid>1) )
	     {
		ts_control[type_pid].ts_ctrl_type = UNDEFINED;
	     }
	}
    }
/*
 * deallocate VCT table struct
 */
    free( current_vct_hdr->vct_tbl_ptr );
    current_vct_hdr->vct_tbl_ptr = NULL;
}

/* --------------------------------------------------------------------- */

	/* ETT tables */

/*
 * Expand the byte linear ETT tables currently only for printing summary
 */

build_ett(q_hdr)
struct  queue_hdr  *q_hdr;
{
    char         *buf_ptr, *c_ptr, *descr_ptr;
    unsigned int  last_sect, i, j, tbl_sect_len, num, sect;
    unsigned int  len_complete, id_ext;
    struct        buf_hdr      *cur_link;

    cur_link = q_hdr->q_head_ptr;
    len_complete = 0;
    sect = 0;

#ifdef DEBUG_SUMMARY
    write_msg('P', 'D', "ETT table - bytes_checked=0x%x", q_hdr->q_tot_bytes);
    while ( cur_link != NULL )
    {
	buf_ptr = cur_link->buf_ptr;
	sect++;
	tbl_sect_len = ((buf_ptr[1]&0x0F)<<8) | (buf_ptr[2]&0xFF);
	id_ext = ((buf_ptr[3]&0xFF)<<8) | (buf_ptr[4]&0xFF);
	last_sect = (buf_ptr[7]&0xFF);
	start_msg(" %d. sect=%d len=0x%x ID_ext=0x%x", sect, (buf_ptr[6]&0xFF),
					tbl_sect_len, id_ext );
	append_msg(" ver=0x%x",  ((buf_ptr[5]>>1)&0x1F) );
	append_msg(" last_sect=0x%x",  last_sect );
	append_msg(" ETM_id=0x%x", ( ((buf_ptr[9]&0xFF)<<24) |
	 ((buf_ptr[10]&0xFF)<<16) | ((buf_ptr[11]&0xFF)<<8) |
	 (buf_ptr[12]&0xFF) ) );

	prt_multiple_string_struct( &buf_ptr[13] );

	len_complete += tbl_sect_len+3;
	cur_link = cur_link->q_ptr;
    }
#endif
}

/* --------------------------------------------------------------------- */

prt_multiple_string_struct( char_ptr )
char *char_ptr;
{
    char *c_ptr, char_strg[1024*8];
    unsigned int i, j, num_strg, num_bytes, num_segs, type, mode, txt_len;

    c_ptr = char_ptr;
    num_strg = (*c_ptr++)&0xFF;
    end_msg('P', 'D', " strings = %d", num_strg);
    for (i=0; i< num_strg; i++)
    {
	num_segs = c_ptr[3]&0xFF;
	write_msg('P', 'D', "  strg %d. ISO_639_code=0x%x %x %x, segments=%d", i+1,
		c_ptr[0]&0xFF, c_ptr[1]&0xFF, c_ptr[2]&0xFF, num_segs);
	c_ptr += 4;
	for (j=0; j< num_segs; j++)
	{
	    type = c_ptr[0]&0xFF;
	    mode = c_ptr[1]&0xFF;
	    num_bytes = c_ptr[2]&0xFF;
	    write_msg('P', 'D', "    seg %d. comp_type=0x%x mode=0x%x num_bytes=0x%x",
			j+1, type, mode, num_bytes);
 /* need to insert huffman decoding */
	    if ( num_bytes > 0 )
	    {
		if ( type == 0 )
		{
		     bcopy(&c_ptr[3],char_strg,num_bytes);
		     txt_len = num_bytes;
		}
		else if ( (type==1) || (type==2) )
		{
		    txt_len = huffman_decode(&c_ptr[3],type,num_bytes,char_strg);
		}
		else
		{
		    write_msg('P', 'D', "Can't handle comp_type");
		    c_ptr += (num_bytes +4);
		    continue;
		}
		if ( mode == 0 )
		{
		    prt_txt(char_strg,txt_len);
		}
		else
		{
		    write_msg('P', 'D', "Can't handle mode");
		}
	    }
	    c_ptr += (num_bytes +4);
	}
    }
}

/*-----------------------------------------------------*/

#include "huffman_tables.h"

huffman_decode(strg,type,len,dest_strg)
unsigned char *strg;
int            type;
int            len;
unsigned char *dest_strg;
{
	int            more_data, more_nodes, bit, bit_index, base_offset;
	int            tree_offset, node_offset, j, val, decode_len;
	unsigned char  *strg_ptr, last_char, *root_table_addr, *tree_table;

/* select table */

	decode_len = 0;
	if ( (type==1) || (type==2) )
	{
		if ( type == 1)
		{
			root_table_addr = title_table;
/*   write_msg('P', 'D', "Title table");
*/              }
		else
		{
			root_table_addr = descr_table;
		}
/*
 *  init for outer loop, which is the per char loop
 */
		last_char = 0;
		strg_ptr = strg;
		bit_index = 7;
		more_data = 1;
		while ( more_data )
		{
			base_offset = (last_char<<1)&0xFF;
/*  write_msg('P', 'D', "base_offset=%d", base_offset);
*/                      tree_offset =  ((root_table_addr[base_offset]&0xFF)<<8)
					|  (root_table_addr[base_offset+1]&0xFF);
/*  write_msg('P', 'D', "tree_offset=%d", tree_offset);
*/                      tree_table = &root_table_addr[tree_offset];
			node_offset = 0;
			more_nodes = 1;
/*
 *  the inner loop follows node ptrs within the single tree until
 *  it reaches a leaf, which is the char
 */
			while (more_nodes  )
			{
/*  get next bit from encoded input strg */
				bit = ((*strg_ptr)>>bit_index)&0x1;
/* advance the byte & bit indices */
				if (  bit_index  == 0 )
				{
					bit_index = 7;
					strg_ptr++;
				}
				else
				{
					bit_index--;
				}
/* follow link to next node, remember that the node_offset is a
 * word addr vs a byte addr
 */
				node_offset = tree_table[2*node_offset+bit];
/*   write_msg('P', 'D', "node_offset=%d", node_offset);
*/                              if ( (node_offset&0x80) != 0 )
				{
/* node is a leaf, check for 2 special char */
					more_nodes = 0;
					last_char = node_offset&0x7F;
/* esc char implies an uncompressed char in the encode input strg */
					if ( last_char == 27 )
					{
						val = ((strg_ptr[0]&0xFF)<<8)
							| (strg_ptr[1]&0xFF);
						strg_ptr++;
						last_char = (val>>(bit_index+1))&0xFF;

					}
/* zero char terminates the string */
					if ( last_char == 0 )
						return( decode_len );
/* process decoded char */
					*dest_strg++ = last_char;
					decode_len++;
/*                                      write_msg('P', 'D', "Char=%c", last_char);
*/                              }
			}
		}	
	}
	else
		write_msg('P', 'E', "ERROR type=%d UNKNOWN in huffman_decode()", type);
	return( decode_len );
}

/* --------------------------------------------------------------------- */

prt_txt(strg,num_bytes)
unsigned char *strg;
unsigned int  num_bytes;
{
       int i;

       start_msg("        strg_val=");
       for (i=0; i<num_bytes; i++)
	   append_msg("%c", *strg++ );
       end_msg('P', 'D', "");
}

/* --------------------------------------------------------------------- */

	/* EIT tables */

/*
 * Expand the byte linear EIT tables currently only for printing summary
 */

build_eit(q_hdr)
struct  queue_hdr  *q_hdr;
{
    char         *buf_ptr, *c_ptr, *descr_ptr;
    unsigned int total_items, last_sect, i, j, tbl_sect_len, title_len, sect;
    unsigned int len_left, descr_len, descr_tag, num_events, len_complete;
    struct        buf_hdr      *cur_link;

    cur_link = q_hdr->q_head_ptr;
    buf_ptr = cur_link->buf_ptr;
    last_sect = (buf_ptr[7]&0xFF);

#ifdef DEBUG_SUMMARY
    len_complete = 0;
    sect = 0;
    write_msg('P', 'D', "EIT table - last_sect=0x%x bytes_checked=0x%x",
				last_sect, q_hdr->q_tot_bytes);
    while ( cur_link != NULL )
    {
	buf_ptr = cur_link->buf_ptr;
	sect++;
	tbl_sect_len = ((buf_ptr[1]&0x0F)<<8) | (buf_ptr[2]&0xFF);
	start_msg(" %d. sect %d. len=0x%x", sect, (buf_ptr[6]&0xFF), tbl_sect_len );
	append_msg(" src_id=0x%x", ((buf_ptr[3]&0xFF)<<8) | (buf_ptr[4]&0xFF) );
	append_msg(" ver=0x%x",  ((buf_ptr[5]>>1)&0x1F) );
	num_events = buf_ptr[9]&0xFF;
	end_msg('P', 'D', " proto_ver=0x%x num_events=%d", (buf_ptr[8]&0xFF), num_events );
	buf_ptr += 10;
	len_complete += 10;
	for (j=0; j<num_events; j++)
	{
	    start_msg("  %d. event=0x%x start=0x%x EMT_loc=0x%x",
		     j+1, (((buf_ptr[0]&0x3F)<<8) | (buf_ptr[1]&0xFF)),
		     (((buf_ptr[2]&0xFF)<<24) | ((buf_ptr[3]&0xFF)<<16)
		      | ((buf_ptr[4]&0xFF)<< 8) | (buf_ptr[5]&0xFF)),
		     ((buf_ptr[6]>>4)&0x03) );
	    title_len = buf_ptr[9]&0xFF;
	    append_msg(" len_sec=0x%x len_title==0x%x ",
		     (((buf_ptr[6]&0x0F)<<16) | ((buf_ptr[7]&0xFF)<< 8)
			| (buf_ptr[8]&0xFF)), title_len );
	    if ( title_len > 0 )
		prt_multiple_string_struct( &buf_ptr[10] );
	    else
		end_msg('P', 'D', "");
	    buf_ptr += (title_len+10);
	    len_complete += (title_len+10);

	    descr_len = ((buf_ptr[0]&0x0F)<<8) | (buf_ptr[1]&0xFF) ;
	    descr_ptr =  buf_ptr + 2;
	    write_msg('P', 'D', "   descr_len=0x%x", descr_len);
	    prt_all_descr(descr_ptr,descr_len);
	    buf_ptr += (2+descr_len);
	    len_complete += (2+descr_len);
	}
	len_complete += 4;
	cur_link = cur_link->q_ptr;
    }
#endif

}

/* --------------------------------------------------------------------- */

	/* PAT tables */

/*
 * Delete the PAT table & any associated PMT tables
 */

delete_pat()
{
    unsigned int i;
    struct       pat_item       *pat_item_ptr;
    struct       ts_control     *ts_control_ptr;

/* deallocate any exisiting PAT & associated PMT */

    if ( (pat_item_ptr=pat_hdr.pat_item_ptr) != NULL )
    {
/* step thru each item (a PMT) of the PAT
 */
	for (i=0; i<pat_hdr.num_items; i++,pat_item_ptr++)
	{
	    if ( pat_item_ptr->pmt_q_hdr_ptr != NULL  )
	    {
/* deallocate PMT queue
 */
		clear_queue( pat_item_ptr->pmt_q_hdr_ptr );
		free( pat_item_ptr->pmt_q_hdr_ptr );
	    }
/* clear entry to the "ts_control" struct for this PMT
 */
	    ts_control[pat_item_ptr->program_map_pid].ts_ctrl_type =
								   UNDEFINED;
	}
/* deallocate PAT & set the ptr in the PAT_hdr to NULL
 */
	free( pat_hdr.pat_item_ptr );
	pat_hdr.pat_item_ptr = NULL;
    }
}

/* --------------------------------------------------------------------- */

	/* PAT tables */

/*
 * Expand the byte linear PAT table into a more convienent data struct
 */

build_pat(q_hdr)
struct  queue_hdr  *q_hdr;
{
    char         *buf_ptr;
    unsigned int total_items, last_sect, i, j, tbl_sect_len, num;
    struct       pat_item       *pat_item_ptr;
    struct       ts_control     *ts_control_ptr;
    struct       buf_hdr        *buf_hdr_ptr;

/* deallocate any exisiting PAT & the associated PMT */

    if ( pat_hdr.pat_item_ptr != NULL )
    {
	delete_pat();
    }

/* build new PAT */

    buf_hdr_ptr = q_hdr->q_head_ptr;
    buf_ptr = buf_hdr_ptr->buf_ptr;

#ifdef DEEP_DEBUG
    start_msg("buf(0x%x)=", buf_ptr);
    for (i=0; i<16; i++)
	append_msg("%2x ", buf_ptr[i]&0xFF);
    end_msg('P', 'D', "");
#endif
/*
 * from the number of PAT table sections compute the total table entries,
 * this can be done since there are no variable size entries in the PAT
 */
    last_sect = (buf_ptr[7] & 0xFF);
    total_items = (q_hdr->q_tot_bytes - ((last_sect+1)*PAT_OVERHEAD_BYTES))
					/PAT_ENTRY_BYTES;
/*
 * place common info into PAT hdr struct
 */
    pat_hdr.num_items = total_items;
    pat_hdr.version_number = ((buf_ptr[5]>>1) & 0x1F);
    pat_hdr.transport_stream_id = ((buf_ptr[3]&0xFF)<<8) | (buf_ptr[4]&0xFF);
/*
 * allocate space for PAT struct
 */
    pat_hdr.pat_item_ptr = malloc( total_items * sizeof(struct pat_item) );
/*
 * step thru each table section (in table buffer)
 */
    pat_item_ptr = pat_hdr.pat_item_ptr;
    while ( buf_hdr_ptr != NULL )
    {
	buf_ptr = buf_hdr_ptr->buf_ptr;
	tbl_sect_len = ((buf_ptr[1]&0xF)<<8) | (buf_ptr[2]&0xFF);
/*
 * step thru each program in the section & allocate space for
 * individual PMT buf_hdr & buf and a ptr to them
 */
	num = (tbl_sect_len + 3 - PAT_OVERHEAD_BYTES) / PAT_ENTRY_BYTES;
	buf_ptr += PAT_HEADER_BYTES;
	for (j=0; j<num; j++, pat_item_ptr++)
	{
	    pat_item_ptr->program_number = ((buf_ptr[0]&0xFF)<<8) | (buf_ptr[1]&0xFF);
	    pat_item_ptr->program_map_pid = ((buf_ptr[2]&0x1F)<<8) | (buf_ptr[3]&0xFF);
	    pat_item_ptr->pmt_q_hdr_ptr = malloc( sizeof(struct queue_hdr) );
	    pat_item_ptr->pmt_q_hdr_ptr->q_head_ptr = NULL;
	    pat_item_ptr->pmt_q_hdr_ptr->q_tail_ptr = NULL;
	    pat_item_ptr->pmt_q_hdr_ptr->q_valid    = 0;
	    pat_item_ptr->pmt_q_hdr_ptr->q_tot_bytes = 0;
 /*
  * insert entry into the "ts_control" struct
  */
	    ts_control_ptr = &ts_control[pat_item_ptr->program_map_pid];
	    ts_control_ptr->ts_ctrl_type   = PMT ;
	    ts_control_ptr->ts_ctrl_status = ACTIVE;
	    ts_control_ptr->ts_ctrl_gen.ts_ctrl_ptr = (unsigned char *) pat_item_ptr;
 /*
  * advance to next entry in table buf
  */
	    buf_ptr += 4;
	}
	   /* skip past crc to next section */
	buf_hdr_ptr = buf_hdr_ptr->q_ptr;
    }
#ifdef DEBUG_SUMMARY
	start_msg("PAT table -");
	append_msg(" ver=0x%x", pat_hdr.version_number);
	append_msg(" items=%d", pat_hdr.num_items);
	end_msg('P', 'D', " ts_stream_id=0x%x", pat_hdr.transport_stream_id);
	pat_item_ptr = pat_hdr.pat_item_ptr;
	for (i=0; i<pat_hdr.num_items; i++, pat_item_ptr++)
	{
	    start_msg(" %2d ", i);
	    append_msg(" prog=0x%4x", pat_item_ptr->program_number);
	    end_msg('P', 'D', " pid=0x%4x", pat_item_ptr->program_map_pid);
	}
#endif
}

/* --------------------------------------------------------------------- */

	/* PMT tables */

/*
 * Expand the byte linear PMT tables currently only for printing summary
 */

build_pmt(q_hdr)
struct  queue_hdr  *q_hdr;
{
    char         *buf_ptr, *c_ptr, *descr_ptr;
    short        assoc_tag;
    unsigned int total_items, last_sect, i, j, tbl_sect_len, num;
    unsigned int len_left, descr_len, d_len_left, type_stream_id, type_pid;
    struct       pat_item       *pat_item_ptr;
    struct       ts_control     *ts_control_ptr;
    struct       buf_hdr        *buf_hdr_ptr;
    struct       dst_nrt_q_lst  *q_lst_ptr;
    struct       dii_ddb_lst    *dii_ddb_lst;
    struct       addr_sect_lst  *addr_sect_lst;

    buf_hdr_ptr = q_hdr->q_head_ptr;
    buf_ptr = buf_hdr_ptr->buf_ptr;

    assoc_tag = 0;
    tbl_sect_len = ((buf_ptr[1]&0x0F)<<8) | (buf_ptr[2]&0xFF);
    last_sect = (buf_ptr[7]&0xFF);

#ifdef DEBUG_SUMMARY
	start_msg("PMT table -");
	append_msg(" sect_len=0x%x", tbl_sect_len );
	append_msg(" prog #=0x%x", ((buf_ptr[3]&0xFF)<<8) | (buf_ptr[4]&0xFF) );
	append_msg(" ver=0x%x",  ((buf_ptr[5]>>1)&0x1F) );
	append_msg(" last_sect=0x%x",  last_sect );
	append_msg(" tbl_len=0x%x",  buf_hdr_ptr->buf_size );
	end_msg('P', 'D', " PRC_PID=0x%x", ((buf_ptr[8]&0x1F)<<8) | (buf_ptr[9]&0xFF) );
	descr_len = ((buf_ptr[10]&0x0F)<<8) | (buf_ptr[11]&0xFF) ;
	descr_ptr =  buf_ptr + 12;
	c_ptr = buf_ptr + descr_len + 12;
	len_left = (tbl_sect_len - 9 - 4) - (descr_len) ;
	start_msg("   len_left=0x%x", len_left );
	end_msg('P', 'D', " descr_len=0x%x", descr_len);

	prt_all_descr(descr_ptr,descr_len);
	while ( len_left > 0 )
	{
	    start_msg(" stream_type =0x%x", (c_ptr[0]&0xFF) );
	    end_msg('P', 'D', " PID =0x%x", ((c_ptr[1]&0x1F)<<8) | (c_ptr[2]&0xFF) );
	    descr_len = ((c_ptr[3]&0x0F)<<8) | (c_ptr[4]&0xFF);
	    descr_ptr = c_ptr+5;
	    c_ptr += (descr_len + 5) ;
	    len_left -= (descr_len + 5) ;

	    prt_all_descr(descr_ptr,descr_len);
	}
#endif

	descr_len = ((buf_ptr[10]&0x0F)<<8) | (buf_ptr[11]&0xFF) ;
	descr_ptr =  buf_ptr + 12;
	c_ptr = buf_ptr + descr_len + 12;
	len_left = (tbl_sect_len - 9 - 4) - (descr_len) ;
	while ( len_left > 0 )
	{
	    type_stream_id = c_ptr[0] & 0xFF;
	    type_pid  = ((c_ptr[1]&0x1F)<<8) | (c_ptr[2]&0xFF);
	    descr_len = ((c_ptr[3]&0x0F)<<8) | (c_ptr[4]&0xFF);
	    descr_ptr = c_ptr+5;

	    if ( (type_pid!=0x1FFB) && (type_pid!=0x1FFF) && (type_pid>1) )
	    {
		ts_control_ptr = &ts_control[type_pid];
		ts_control_ptr->ts_orig_stream = type_stream_id;
		if ( ts_control_ptr->ts_ctrl_type == UNDEFINED )
		    ts_control_ptr->ts_cont_cntr   = 0xFF;

	/* search descr for assoc tag info */

		assoc_tag = 0;
		d_len_left = descr_len;
		while ( d_len_left > 0 )
		{
		     if ( (descr_ptr[0]&0xFF) == 0x14 )
		     {
			 assoc_tag = ((descr_ptr[2]&0xFF)<<8) | (descr_ptr[3]&0xFF);
			 break;
		     }
		     d_len_left -= ( (descr_ptr[1]&0xFF)+2 );
		     descr_ptr  += ( (descr_ptr[1]&0xFF)+2 );
		}

	/* process stream_type */

		switch ( type_stream_id )
		{
		   case 0x02:
		      if ( ts_control_ptr->ts_ctrl_type == VIDEO )
		      {  /* already defined by VCT */
			   //write_msg('P', 'I', "VIDEO already defined by VCT");
			   break;
		      }
		      ts_control_ptr->ts_ctrl_type = VIDEO;
		      ts_control_ptr->ts_orig_type = VIDEO;
		      ts_control_ptr->ts_ctrl_status = INACTIVE;
		      ts_control_ptr->ts_buf_hdr_ptr = NULL;
		      ts_control_ptr->ts_ctrl_gen.ts_ctrl_fd = -1;
		      clear_queue( &(ts_control_ptr->ts_queue_hdr) );
		      if ( first_video_stream == 1 )
		      {
			 if ( output_video_stream != -1 )
			 {
			     first_video_stream = 0;
			     ts_control_ptr->ts_ctrl_status = READY;
			     ts_control_ptr->ts_ctrl_gen.ts_ctrl_fd =
						   output_video_stream;
			     open_pid_output_dev(type_pid,output_video_dev_ptr);
			 }
		      }
		      break;

		   case 0x06:
		      if ( ts_control_ptr->ts_ctrl_type == PES_IZED )
		      {  /* already defined by VCT */
			   //write_msg('P', 'I', "PES_IZED already defined by VCT");
			   break;
		      }
		      ts_control_ptr->ts_ctrl_type = PES_IZED;
		      ts_control_ptr->ts_orig_type = PES_IZED;
		      ts_control_ptr->ts_ctrl_status = ACTIVE;
		      ts_control_ptr->ts_buf_hdr_ptr = NULL;
		      ts_control_ptr->ts_ctrl_gen.ts_ctrl_ptr  =
					 (unsigned char *) NULL;
		      clear_queue( &(ts_control_ptr->ts_queue_hdr) );
		      break;

		   case 0x0D:
		      if ( ts_control_ptr->ts_ctrl_type == ADDR_SECT )
		      {  /* already defined by VCT */
			   addr_sect_lst = (struct addr_sect_lst *)
				     ts_control_ptr->ts_ctrl_gen.ts_ctrl_ptr;
			   addr_sect_lst->assoc_tag = assoc_tag;
			   //write_msg('P', 'I', "ADDR_SECT already defined by VCT");
			   break;
		      }
		      ts_control_ptr->ts_ctrl_type = ADDR_SECT;
		      ts_control_ptr->ts_orig_type = ADDR_SECT;
		      ts_control_ptr->ts_ctrl_status = ACTIVE;
		      ts_control_ptr->ts_buf_hdr_ptr = NULL;

		 /* create a list item for this addr_sect entry */
		 /* and add it to the linked list               */

		      addr_sect_lst =  malloc( sizeof(struct addr_sect_lst) );
		      addr_sect_lst->nxt_addr_sect_ptr = addr_sect_lst_head;
		      addr_sect_lst->bck_addr_sect_ptr = NULL;
		      if ( addr_sect_lst_head != NULL )
		      {
			  addr_sect_lst_head->bck_addr_sect_ptr = addr_sect_lst;
		      }
		      addr_sect_lst_head = addr_sect_lst;
		      ts_control_ptr->ts_ctrl_gen.ts_ctrl_ptr =
				   (unsigned char *) addr_sect_lst;
		      addr_sect_lst->pid = type_pid;
		      addr_sect_lst->assoc_tag = assoc_tag;
		      clear_queue( &(ts_control_ptr->ts_queue_hdr) );
		      break;

		   case 0x0B:
		   case 0x14:
		      if ( ts_control_ptr->ts_ctrl_type == DII_DDB )
		      {      /* already defined by VCT */
			   dii_ddb_lst = (struct dii_ddb_lst *)
				     ts_control_ptr->ts_ctrl_gen.ts_ctrl_ptr;
			   dii_ddb_lst->assoc_tag = assoc_tag;
			   //write_msg('P', 'I', "DII_DDB already defined by VCT");
			   break;
		      }
		      ts_control_ptr->ts_ctrl_type = DII_DDB;
		      ts_control_ptr->ts_orig_type = DII_DDB;
		      ts_control_ptr->ts_ctrl_status = ACTIVE;
		      ts_control_ptr->ts_buf_hdr_ptr = NULL;
			    /* create a list item for this set of tables */
			    /* and add it to the linked list              */

		      dii_ddb_lst = malloc( sizeof(struct dii_ddb_lst) );
		      dii_ddb_lst->nxt_dii_ddb_lst = dii_ddb_lst_head;
		      dii_ddb_lst->bck_dii_ddb_lst = NULL;
		      if ( dii_ddb_lst_head != NULL )
		      {
			  dii_ddb_lst_head->bck_dii_ddb_lst
				= dii_ddb_lst;
		      }
		      dii_ddb_lst_head = dii_ddb_lst;
		      ts_control_ptr->ts_ctrl_gen.ts_ctrl_ptr =
			      (unsigned char *) dii_ddb_lst;

		      dii_ddb_lst->dii_pid = type_pid;
		      dii_ddb_lst->assoc_tag = assoc_tag;
		      dii_ddb_lst->dii_q_hdr   = create_queue();
		      dii_ddb_lst->dsi_buf_hdr = NULL;
		      dii_ddb_lst->dc_buf_hdr  = NULL;
		      dii_ddb_lst->carousel    = NULL;

		      clear_queue( &(ts_control_ptr->ts_queue_hdr) );
		      break;

		   case 0x81:
		      if ( ts_control_ptr->ts_ctrl_type == AUDIO_AC3 )
		      {  /* already defined by VCT */
			   //write_msg('P', 'I', "AUDIO_AC3 already defined by VCT");
			   break;
		      }
		      ts_control_ptr->ts_ctrl_type = AUDIO_AC3;
		      ts_control_ptr->ts_orig_type = AUDIO_AC3;
		      ts_control_ptr->ts_ctrl_status = INACTIVE;
		      ts_control_ptr->ts_buf_hdr_ptr = NULL;
		      ts_control_ptr->ts_ctrl_gen.ts_ctrl_fd = -1;
		      if ( first_audio_stream == 1 )
		      {
			 if ( output_audio_stream != -1 )
			 {
			     first_audio_stream = 0;
			     ts_control_ptr->ts_ctrl_status = READY;
			     ts_control_ptr->ts_ctrl_gen.ts_ctrl_fd =
						   output_audio_stream;
			     open_pid_output_dev(type_pid,output_audio_dev_ptr);
			 }
		      }
		      break;

		   case 0x95:
		      if ( ts_control_ptr->ts_ctrl_type == DST_NRT )
		      {      /* already defined by VCT */
 		         q_lst_ptr->assoc_tag = assoc_tag;
			   //write_msg('P', 'I', "DST_NRT already defined by VCT");
			   break;
		      }
		      ts_control_ptr->ts_ctrl_type = DST_NRT;
		      ts_control_ptr->ts_orig_type = DST_NRT;
		      ts_control_ptr->ts_ctrl_status = ACTIVE;
		      ts_control_ptr->ts_buf_hdr_ptr = NULL;
		      clear_queue( &(ts_control_ptr->ts_queue_hdr) );
		 /* create a list item for this pair of tables */
		 /* and add it to the linked list              */

		      q_lst_ptr = malloc( sizeof(struct dst_nrt_q_lst) );
		      q_lst_ptr->nxt_q_lst_ptr = dst_nrt_q_lst_head;
		      q_lst_ptr->bck_q_lst_ptr = NULL;
		      if ( dst_nrt_q_lst_head != NULL )
		      {
			  dst_nrt_q_lst_head->bck_q_lst_ptr
				= q_lst_ptr;
		      }
		      dst_nrt_q_lst_head = q_lst_ptr;
		      ts_control_ptr->ts_ctrl_gen.ts_ctrl_ptr =
				   (unsigned char *) q_lst_ptr;
		      q_lst_ptr->pid = type_pid;
		      q_lst_ptr->assoc_tag = assoc_tag;
			  /* allocate & init DST queue */
		      q_lst_ptr->dst_q_hdr = create_queue();
			  /* allocate & init NRT queue */
		      q_lst_ptr->nrt_q_hdr = create_queue();
   /* need to keep a list of new & old DST/NRTs to know when */
   /* to delete previous DST/NRTs that are no longer used */
		      break;

		   case 0xC2:
		      if ( ts_control_ptr->ts_ctrl_type == PES_OUS )
		      {  /* already defined by VCT */
			   //write_msg('P', 'I', "PES_OUS already defined by VCT");
			   break;
		      }
		      ts_control_ptr->ts_ctrl_type = PES_OUS;
		      ts_control_ptr->ts_orig_type = PES_OUS;
		      ts_control_ptr->ts_ctrl_status = ACTIVE;
		      ts_control_ptr->ts_buf_hdr_ptr = NULL;
		      ts_control_ptr->ts_ctrl_gen.ts_ctrl_ptr  =
					 (unsigned char *) NULL;
		      clear_queue( &(ts_control_ptr->ts_queue_hdr) );
		      break;

		   default:
		      if ( type_stream_id >= 0xC0 )
		      {       /* User Private -- assume data piping */
			  ts_control_ptr->ts_ctrl_type = PIPING_COOKED;
			  ts_control_ptr->ts_orig_type = PIPING_COOKED;
		      }
		      else
		      {
			  ts_control_ptr->ts_ctrl_type = UNDEFINED;
			  ts_control_ptr->ts_orig_type = UNDEFINED;
		      }
		      ts_control_ptr->ts_buf_hdr_ptr = NULL;
		      clear_queue( &(ts_control_ptr->ts_queue_hdr) );
		      ts_control_ptr->ts_ctrl_gen.ts_ctrl_ptr = NULL;
		      break;
		}
	    }
	    c_ptr += (descr_len + 5) ;
	    len_left -= (descr_len + 5) ;
	}
}

/* ---------------------------------------------------------------- ---- */

/*
 * Expand the byte linear DST table currently only for printing summary
 */

build_dst_table(q_hdr)
struct  queue_hdr  *q_hdr;
{
   unsigned int  i, j;
   char          *buf_ptr;
   unsigned int  tbl_sect_len, last_sect, len;
   struct  buf_hdr  *buf_hdr_ptr;

    buf_hdr_ptr = q_hdr->q_head_ptr;
    buf_ptr = buf_hdr_ptr->buf_ptr;

    tbl_sect_len = ((buf_ptr[1]&0x0F)<<8) | (buf_ptr[2]&0xFF);
    last_sect = (buf_ptr[7]&0xFF);

#ifdef DEBUG_SUMMARY
	start_msg("DST table -");
	append_msg(" sect_len=0x%x", tbl_sect_len );
	append_msg(" ver=0x%x",  ((buf_ptr[5]>>1)&0x1F) );
	append_msg(" last_sect=0x%x",  last_sect );
	append_msg(" tbl_len=0x%x",  buf_hdr_ptr->buf_size );
	end_msg('P', 'D', " proto_ver=0x%x", (buf_ptr[8]&0xFF) );
#endif
}

/* ---------------------------------------------------------------- ---- */

/*
 * Expand the byte linear NRT table currently only for printing summary
 */

build_nrt_table(q_hdr)
struct  queue_hdr  *q_hdr;
{
   unsigned int  i, j;
   char          *buf_ptr;
   unsigned int  tbl_sect_len, last_sect, len;
   struct  buf_hdr  *buf_hdr_ptr;

    buf_hdr_ptr = q_hdr->q_head_ptr;
    buf_ptr = buf_hdr_ptr->buf_ptr;

    tbl_sect_len = ((buf_ptr[1]&0x0F)<<8) | (buf_ptr[2]&0xFF);
    last_sect = (buf_ptr[7]&0xFF);

#ifdef DEBUG_SUMMARY
	start_msg("NRT table -");
	append_msg(" sect_len=0x%x", tbl_sect_len );
	append_msg(" ver=0x%x",  ((buf_ptr[5]>>1)&0x1F) );
	append_msg(" last_sect=0x%x",  last_sect );
	end_msg('P', 'D', " tbl_len=0x%x",  buf_hdr_ptr->buf_size );
#endif
}

/* ---------------------------------------------------------------- ---- */

/*
 * Expand the byte linear DCCT table currently only for printing summary
 */

build_dcct_table(q_hdr)
struct  queue_hdr  *q_hdr;
{
   unsigned int  i, j;
   char          *buf_ptr;
   unsigned int  tbl_sect_len, last_sect, s_time, e_time, len;
   unsigned int  f_maj, f_min, t_maj, t_min, sel_cnt, vc_count, sel_type;
   struct  buf_hdr  *buf_hdr_ptr;

    buf_hdr_ptr = q_hdr->q_head_ptr;
    buf_ptr = buf_hdr_ptr->buf_ptr;

    tbl_sect_len = ((buf_ptr[1]&0x0F)<<8) | (buf_ptr[2]&0xFF);
    last_sect = (buf_ptr[7]&0xFF);

#ifdef DEBUG_SUMMARY
	start_msg("DCCT table -");
	append_msg(" sect_len=0x%x", tbl_sect_len );
	append_msg(" ver=0x%x",  ((buf_ptr[5]>>1)&0x1F) );
	append_msg(" last_sect=0x%x",  last_sect );
	append_msg(" tbl_len=0x%x",  buf_hdr_ptr->buf_size );
	append_msg(" proto_ver=0x%x", (buf_ptr[8]&0xFF) );
	vc_count = buf_ptr[9]&0xFF;
	end_msg('P', 'D', " dcc_vc_count=0x%x", vc_count );
	buf_ptr = &buf_ptr[10];
	for ( i=0; i<vc_count; i++ )
	{
		f_maj = ((buf_ptr[0]&0xF)<<6) | ((buf_ptr[1]>>2)&0x3F);
		f_min = ((buf_ptr[1]&0x3)<<8) | (buf_ptr[2]&0xFF);
		t_maj = ((buf_ptr[3]&0xF)<<6) | ((buf_ptr[4]>>2)&0x3F);
		t_min = ((buf_ptr[4]&0x3)<<8) | (buf_ptr[5]&0xFF);
		s_time = ((buf_ptr[6]&0xFF)<<24) | ((buf_ptr[7]&0xFF)<<16)
			 | ((buf_ptr[8]&0xFF)<< 8) | (buf_ptr[9]&0xFF);
		e_time = ((buf_ptr[10]&0xFF)<<24) | ((buf_ptr[11]&0xFF)<<16)
			 | ((buf_ptr[12]&0xFF)<< 8) | (buf_ptr[13]&0xFF);
		write_msg('P', 'D', " %d. from = %d-%d to %d-%d during 0x%x-%x",
			i+1, f_maj, f_min, t_maj, t_min, s_time, e_time );
		sel_cnt  = buf_ptr[14]&0xFF;
		buf_ptr += 15;
		for ( j=0; j<sel_cnt; j++ )
		{
		    sel_type = buf_ptr[0]&0xFF;
		    s_time = ((buf_ptr[1]&0xFF)<<24) | ((buf_ptr[2]&0xFF)<<16)
			     | ((buf_ptr[3]&0xFF)<< 8) | (buf_ptr[4]&0xFF);
		    e_time = ((buf_ptr[5]&0xFF)<<24) | ((buf_ptr[6]&0xFF)<<16)
			     | ((buf_ptr[7]&0xFF)<< 8) | (buf_ptr[8]&0xFF);
		    len = ((buf_ptr[9]&0x3)<<8) | (buf_ptr[10]&0xFF);
		    write_msg('P', 'D', "     sel_type=%d, sel_id=0x%x %x, descr_len=%d",
			    sel_type, s_time, e_time, len );
		    if ( len > 0 )
			prt_all_descr(&buf_ptr[11],len);
		    buf_ptr += (len + 11);
		}
		len = ((buf_ptr[0]&0x3)<<8) | (buf_ptr[1]&0xFF);
		write_msg('P', 'D', "     vc_descr_len=%d", len);
		if ( len > 0 )
		    prt_all_descr(&buf_ptr[2],len);
		buf_ptr += (len + 2);
	}
	len = ((buf_ptr[0]&0x3)<<8) | (buf_ptr[1]&0xFF);
	write_msg('P', 'D', "  addl_descr_len=%d", len);
	if ( len > 0 )
	    prt_all_descr(&buf_ptr[2],len);
#endif
}

/* ---------------------------------------------------------------- ---- */

/*
 * Expand the byte linear DCCSC table currently only for printing summary
 */

build_dccsct_table(q_hdr)
struct  queue_hdr  *q_hdr;
{
   unsigned int  i, j;
   char          *buf_ptr;
   unsigned int  tbl_sect_len, last_sect, sel_cat, code, name_len, len;
   struct  buf_hdr  *buf_hdr_ptr;

    buf_hdr_ptr = q_hdr->q_head_ptr;
    buf_ptr = buf_hdr_ptr->buf_ptr;

    tbl_sect_len = ((buf_ptr[1]&0x0F)<<8) | (buf_ptr[2]&0xFF);
    last_sect = (buf_ptr[7]&0xFF);

#ifdef DEBUG_SUMMARY
	start_msg("DCCSCT table -");
	append_msg(" sect_len=0x%x", tbl_sect_len );
	append_msg(" ver=0x%x",  ((buf_ptr[5]>>1)&0x1F) );
	append_msg(" last_sect=0x%x",  last_sect );
	append_msg(" tbl_len=0x%x",  buf_hdr_ptr->buf_size );
	append_msg(" proto_ver=0x%x", (buf_ptr[8]&0xFF) );
	sel_cat = buf_ptr[9]&0xFF;
	end_msg('P', 'D', " select_cat_defined=0x%x", sel_cat );
	buf_ptr += 10;
	for ( i=0; i<sel_cat; i++ )
	{
		code = buf_ptr[0]&0xFF;
		name_len = buf_ptr[1]&0xFF;
		start_msg(" %d. code = 0x%x, name_len= 0x%x",
			i+1, code, name_len);
		buf_ptr += (name_len + 2);  /* skip over text */
		len = ((buf_ptr[0]&0x3)<<8) | (buf_ptr[1]&0xFF);
		end_msg('P', 'D', "  code_descr_len=%d", len);
		if ( len > 0 )
		    prt_all_descr(&buf_ptr[2],len);
		buf_ptr += (len + 2);
	}
	len = ((buf_ptr[0]&0x3)<<8) | (buf_ptr[1]&0xFF);
	write_msg('P', 'D', "  addl_descr_len=%d", len);
	if ( len > 0 )
	    prt_all_descr(&buf_ptr[2],len);
#endif
}


/*---------------------------------------------------------------------- */

/*
 * Compute CheckSum
 *
 *  Compute the chksum for a data "buf" of byte length "len" and
 *  compare the results to the transmitted chksum "chksum", rtn "0"
 *  if the chksums match, otherwise rtn "ERROR"
 */

check_chksum(buf,len,chksum)
unsigned char *buf;
unsigned int  len;
unsigned int  chksum;
{
   int               j, val_and, val_xor;
   unsigned int      length, val;
   unsigned long int new_sum;

   if ( chksum == 0 )
   {      /* chksum == 0 implies that the chksum should be ignored */
	return( 0 );
   }
   new_sum   = 0x0;     /* init chksum to be computed */
   length = len;

 /* chsum is computed on a 32 bit basis, NOT a byte basis  */
 /* using 1's complement arithmetic NOT std 2's complement */

   for ( j=length; j>3; j -= 4 )
   {
	val = (*buf++)&0xFF;
	val = (val<<8) | ((*buf++)&0xFF);
	val = (val<<8) | ((*buf++)&0xFF);
	val = (val<<8) | ((*buf++)&0xFF);

	val_and = ((val>>31)&0x1) & ((new_sum>>31)&0x1);
	if ( val_and == 1 )
	{
	    new_sum += val;
	    new_sum++;
	}
	else
	{
	    val_xor = ((val>>31)&0x1) ^ ((new_sum>>31)&0x1);
	    new_sum += val;
	    if ( val_xor == 1 )
	    {
		if ( ((new_sum>>31)&0x1) == 0 )
		   new_sum++;
	    }
	}
   }
   if ( j > 0 )
   {
	val = (*buf++)&0xFF;
	if ( j == 1 )
	    val = (val<<24);
	else
	{
	    val = (val<<8) | ((*buf++)&0xFF);
	    if ( j == 2 )
		val = (val<<16);
	    else     /* j == 3 */
		 val = ((val<<16) | (((*buf++)&0xFF)<<8));
	}
	val_and = ((val>>31)&0x1) & ((new_sum>>31)&0x1);
	if ( val_and == 1 )
	{
	    new_sum += val;
	    new_sum++;
	}
	else
	{
	    val_xor = ((val>>31)&0x1) ^ ((new_sum>>31)&0x1);
	    new_sum += val;
	    if ( val_xor == 1 )
	    {
		if ( ((new_sum>>31)&0x1) == 0 )
		   new_sum++;
	    }
	}
   }
   if ( new_sum == 0 )
   {      /* when computed chksum == 0, set to all 1's (alt value of  */
	  /* 1's complement 0) to differentiate between this and the  */
	  /* "to be ignored" chksum case (chksum==0)                  */
	new_sum = 0xFFFFFFFF;
   }
   if ( chksum != new_sum )
   {            /* bad chksum */
#ifdef DEBUG_RCV
	write_msg('P', 'D', "**** ERROR rcvd_chksum =0x%8x, len=0x%x, comp_chksum =0x%8x",
			chksum, len, new_sum);
#endif
	return(ERROR);
    }
      /* good chksum */
    return( 0 );
}

/*---------------------------------------------------------------------- */

/*
 * Compute CRC
 *
 *  Compute the CRC for a data "buf" of byte length "len" and
 *  compare the results to the transmitted CRC "crc, rtn "0"
 *  if CRCs match, otherwise rtn "ERROR"
 */

unsigned long int g_vec = 0x04C11DB6 ;
/*
static unsigned short int g[] =
	{0,1,1,0,1,1,0,1,
	 1,0,1,1,1,0,0,0,
	 1,0,0,0,0,0,1,1,
	 0,0,1,0,0,0,0,0};
*/

check_crc(buf,len,crc)
unsigned char *buf;
unsigned int  len;
unsigned int  crc;
{
   int               j;
   unsigned int      length;
   unsigned char     data_bit, next_byte;
   unsigned long int shift_reg;

   shift_reg = 0xFFFFFFFF;
   length = len;

    while ( length-- > 0 )
    {
	 next_byte = *buf++;
	 for (j=7; j>=0; j--)
	 {
	    data_bit = (next_byte>>j)&0x01;

	    data_bit ^= ((shift_reg>>31)&0x01);
	    shift_reg = (shift_reg<<1);
	    if ( data_bit == 1 )
	    {
		  shift_reg = (shift_reg ^ g_vec) | 1 ;
	    }
	 }
    }
    if ( crc != shift_reg )
    {
#ifdef DEBUG_RCV
	write_msg('P', 'D', "**** ERROR rcvd_crc =0x%8x, len=0x%x, comp_crc =0x%8x",
			crc, len, shift_reg);
#endif
	return(ERROR);
    }
    return( 0 );
}

/* --------------------------------------------------------------------- */

/*
 * print the list of descr info
 */

prt_all_descr(descr_ptr,descr_len)
char *descr_ptr;
unsigned int descr_len;
{
    unsigned int descr_cnt;

    while ( descr_len > 0 )
    {
	descr_cnt = prt_descr(descr_ptr);
	descr_ptr += descr_cnt;
	descr_len -= descr_cnt;
    }
}

/* --------------------------------------------------------------------- */

/*
 * print one descr info
 */

prt_descr(descr_ptr)
char *descr_ptr;
{
    unsigned int tag, len, i, j, k;

    tag  = (descr_ptr[0]&0xFF);
    len  = (descr_ptr[1]&0xFF);
    start_msg("    Tag=0x%x (", tag);
    prt_tag(tag);
    append_msg("), Len=0x%x  ", len);
    for ( i=2; i<len+2; i++)
    {
       append_msg("0x%x ", descr_ptr[i]&0xFF);
       if ( i%18 == 0 )
       {
		end_msg('P', 'D', "");
		start_msg("");
	   }
    }
    end_msg('P', 'D', "");

    switch ( tag )
    {
       case 0xA1:       /* serv_location_descr */
	    write_msg('P', 'D', "            pcr_pid=0x%x",
			(((descr_ptr[2]&0x1F)<<8)|(descr_ptr[3]&0xFF)) );
	    i = descr_ptr[4]&0xFF;
	    descr_ptr += 5;
	    for ( k=0; k<i; k++ )
	    {
		start_msg("         %2d.", k);
		append_msg(" stream=0x%x", (j = descr_ptr[0]&0xFF)  );
		append_msg(" (%s)", ( (j==2) ? "video" :
				       ((j==0x81) ? "AC3-audio" :
				       ((j==0x03) ? "MPEG1-audio" :
				       ((j==0x04) ? "MPEG1-audio" : "") )))
		   );
		append_msg(" pid=0x%x", (((descr_ptr[1]&0x1F)<<8)|(descr_ptr[2]&0xFF)) );
		append_msg(" lang_code=%c", descr_ptr[3]&0xFF );
		append_msg("%c", descr_ptr[4]&0xFF );
		end_msg('P', 'D', "%c", descr_ptr[5]&0xFF );
		descr_ptr += 6;
	    }
	 break;
       default:
	 break;
    }

    return( len+2 );
}

/* --------------------------------------------------------------------- */

/*
 * print descr tag info
 */

prt_tag(tag)
unsigned int tag;
{

    switch ( tag )
    {
       case    2:
	 append_msg("video_stream_descr");
	 break;
       case    3:
	 append_msg("audio_stream_descr");
	 break;
       case    4:
	 append_msg("hierarchy_descr");
	 break;
       case    5:
	 append_msg("registration_descr");
	 break;
       case    6:
	 append_msg("data_stream_align_descr");
	 break;
       case    7:
	 append_msg("target_background_grid_descr");
	 break;
       case    8:
	 append_msg("video_window_descr");
	 break;
       case    9:
	 append_msg("CA_descr");
	 break;
       case   10:    /* 0xA  */
	 append_msg("ISO_639_lang_descr");
	 break;
       case   11:    /* 0xB  */
	 append_msg("syst_clk_descr");
	 break;
       case   12:    /* 0xC  */
	 append_msg("multiplex_buf_utilization_descr");
	 break;
       case   13:    /* 0xD  */
	 append_msg("copyright_descr");
	 break;
       case   14:    /* 0xE  */
	 append_msg("max_bitrate_descr");
	 break;
       case   15:    /* 0xF  */
	 append_msg("private_data_indicator_descr");
	 break;
       case   16:    /* 0x10 */
	 append_msg("smoothing_buf_descr");
	 break;
       case   17:    /* 0x11 */
	 append_msg("STD_descr");
	 break;
       case   18:    /* 0x12 */
	 append_msg("IBP_descr");
	 break;

       case 0x14:
	 append_msg("assoc_tag_descr");
	 break;
       case 0x80:
	 append_msg("stuffing_descr");
	 break;
       case 0x86:
	 append_msg("caption_serv_descr");
	 break;
       case 0x87:
	 append_msg("content_advisory_descr");
	 break;
       case 0xA0:
	 append_msg("extended_ch_name_descr");
	 break;
       case 0xA1:
	 append_msg("serv_location_descr");
	 break;
       case 0xA2:
	 append_msg("time_shifted_serv_descr");
	 break;
       case 0xA3:
	 append_msg("component_name_descr");
	 break;
       case 0xA4:
	 append_msg("data_service_descr");
	 break;
       case 0xA5:
	 append_msg("PID_count_descr");
	 break;
       case 0xA6:
	 append_msg("Download_descr");
	 break;
       case 0xA7:
	 append_msg("multiprotocol_encapsulate_descr");
	 break;
       case 0xA8:
	 append_msg("DCC_departing_event_descr");
	 break;
       case 0xA9:
	 append_msg("DCC_arriving_event_descr");
	 break;
       case 0xAA:
	 append_msg("Content_Protect_Copy_Mngt_descr");
	 break;

       case 0xB4:
	 append_msg("Module_Link_descr");
	 break;
       case 0xB5:
	 append_msg("CRC32_descr");
	 break;
       case 0xB8:
	 append_msg("Group_Link_descr");
	 break;

       default:
	 append_msg("n/a");
	 break;
    }

}

/* --------------------------------------------------------------------- */

/*
 * print info on the RRT table
 */

prt_rrt_table(q_hdr)
struct  queue_hdr  *q_hdr;
{
   int           i, j;
   char          *buf_ptr;
   unsigned int  sect_len, rating_region, ver_num, sect_num;
   unsigned int  last_sect_num, rating_reg_name_len, dimensions;
   unsigned int  dim_name_len;
   unsigned int  descr_len, abbrv_val_len, abbrv_rate_val_len, num_values;
   struct  buf_hdr  *buf_hdr_ptr;

   start_msg("RRT table -");
   buf_hdr_ptr = q_hdr->q_head_ptr;
   buf_ptr = buf_hdr_ptr->buf_ptr;

   sect_len      = ((buf_ptr[1]&0x0F)<<8) | (buf_ptr[2]&0xFF);

/*
	FILE  *output_rrt_stream;         /* handle for output RRT stream */
/*       if ( (output_rrt_stream = fopen("RRT_tbl_binary",  "w")) == NULL )
	 {
	    write_msg('P', 'E', "***ERROR opening Stream <RRT_tbl_binary> ");
	 }
	 else
	    write_msg('P', 'I', "Stream <RRT_tbl_binary> opened successfully");
	 fwrite(buf_ptr,1,sect_len+3, output_rrt_stream);
*/

   rating_region = buf_ptr[4] & 0xFF;
   ver_num =    ((buf_ptr[5]>>1)&0x1F);
   sect_num =      buf_ptr[6] & 0xFF;
   last_sect_num = buf_ptr[7] & 0xFF;
   end_msg('P', 'D', " sect_len=0x%x, rating_reg=0x%x, ver_num=0x%x, sect_num==x%x, last_sect=0x%x",
	    sect_len, rating_region, ver_num, sect_num, last_sect_num);
/*
    start_msg("inp=");
    for (i=1; i<sect_len+3; i++)
    {
	append_msg("%2x ", buf_ptr[i]&0xFF);
	if ( (i%25) == 0 ) {
	    end_msg('P', 'D', "");
	    start_msg("    ");
	    }
    }
    end_msg('P', 'D', "");
*/

   rating_reg_name_len = buf_ptr[9] & 0xFF;
   write_msg('P', 'D', " rating_reg_name_len=0x%x", rating_reg_name_len);
   buf_ptr += (10 + rating_reg_name_len);
   dimensions = buf_ptr[0] & 0xFF;
   write_msg('P', 'D', " dimensions=0x%x", dimensions);
   buf_ptr++;
   for ( i=0; i<dimensions; i++)
   {
	dim_name_len = buf_ptr[0] & 0xFF;
	start_msg("   dim_name_len=0x%x", dim_name_len);
	buf_ptr += (1+dim_name_len);
	num_values = buf_ptr[0] & 0x0F;
	end_msg('P', 'D', " num_values=0x%x", num_values);
	buf_ptr++;
	for ( j=0; j<num_values; j++)
	{
	    abbrv_rate_val_len = buf_ptr[0] & 0xFF;
	    start_msg("     %d. abbrv_rate_val_len=0x%x", j, abbrv_rate_val_len);
	    buf_ptr +=(1+abbrv_rate_val_len);
	    abbrv_val_len = buf_ptr[0] & 0xFF;
	    end_msg('P', 'D', " abbrv_val_len=0x%x", abbrv_val_len);
	    buf_ptr +=(1+abbrv_val_len);
	}
   }
   descr_len =  ((buf_ptr[0]&0x03)<<8) | (buf_ptr[1]&0xFF);
   write_msg('P', 'D', "   descr_len=0x%x", descr_len);
   buf_ptr += 2;
   prt_all_descr(buf_ptr,descr_len);
}
